/*	$Id: i2c.c,v 1.14 2010/01/13 08:05:35 tsubai Exp $	*/

/*-
 * Copyright (c) 2001,2002 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/systm.h>

#include <machine/autoconf.h>
#include <machine/mpc860dev.h>
#include <machine/sysconf.h>

#define I2C_READ	1
#define I2C_WRITE	0

#define TxBD_R	 0x8000		/* Ready */
#define TxBD_W	 0x2000		/* Wrap */
#define TxBD_I	 0x1000		/* Interrupt */
#define TxBD_L	 0x0800		/* Last */
#define TxBD_S	 0x0400		/* Generate start condition */
#define TxBD_NAK 0x0004		/* No acknowledge */
#define TxBD_UN	 0x0002		/* Underrun */
#define TxBD_CL	 0x0001		/* Collision */

#define RxBD_E	 0x8000		/* Empty */
#define RxBD_W	 0x2000		/* Wrap */
#define RxBD_I	 0x1000		/* Interrupt */
#define RxBD_L	 0x0800		/* Last */
#define RxBD_OV	 0x0002		/* Overrun */

static int initted;
static int dpram;

void
i2c_init()
{
	volatile struct mpc860dev *immr = get_immr();
	volatile struct cpm_bdesc *txbd, *rxbd;
	volatile struct i2c_pram *i2c;

	/* Setup port B pin 26 and 27 for I2C. */
	immr->cpm_pbpar |= 0x00000030;
	immr->cpm_pbdir |= 0x00000030;
	immr->cpm_pbodr |= 0x00000030;

	dpram = dpram_alloc(16*4 + 256, 16);
	if (dpram < 0)
		panic("i2c_init: failed to allocate DPRAM");
	rxbd = (void *)&immr->cpm_dpram[dpram];
	rxbd[0].bd_flags = RxBD_W;

	txbd = rxbd + 1;
	txbd[0].bd_flags = 0;
	txbd[1].bd_flags = TxBD_W | TxBD_L;

	i2c = (void *)&immr->cpm_dpram[DPRAM_I2C];
	i2c->rbase = (int)rxbd & 0xfff;
	i2c->tbase = (int)txbd & 0xfff;
	i2c->rfcr = 0x10;
	i2c->tfcr = 0x10;
	i2c->mrblr = 256;

	/* CLK = SYSCLK / PDIV / (2 * (DIV + 3)) */
	immr->cpm_i2mod = I2MOD_PDIV_32;
	immr->cpm_i2brg = 5;

	immr->cpm_i2com = I2COM_MASTER;
	immr->cpm_i2mod = I2MOD_PDIV_32 | I2MOD_EN;
	immr->cpm_i2add = 0;
	immr->cpm_i2cmr = 0;			/* No interrupts */
	immr->cpm_i2cer = 0xff;			/* Clear all events. */

	initted = 1;
}

int
i2c_read(slave, reg, buf, len)
	int slave, reg, len;
	void *buf;
{
	volatile struct mpc860dev *immr = get_immr();
	volatile struct cpm_bdesc *txbd, *rxbd;

	if (initted == 0)
		i2c_init();

	/* Issue an INIT TX AND RX command. */
	immr->cpm_cpcr = 0x0011;
	while (immr->cpm_cpcr & 1);		/* Wait for completion. */

	rxbd = (void *)&immr->cpm_dpram[dpram];
	rxbd[0].bd_flags = RxBD_E | RxBD_W;
	rxbd[0].bd_addr = (void *)(rxbd + 4);

	/*
	 * XXX Master read needs dummy transmit buffer of size n + 1 bytes.
	 * XXX The first byte contains slave address and R/W.  The next n
	 * XXX bytes are used for timing only.
	 */
	txbd = rxbd + 1;
	txbd[0].bd_addr = (void *)(rxbd + 3);
	txbd[0].bd_addr[0] = slave | I2C_WRITE;	/* Slave address | R/W. */
	txbd[0].bd_addr[1] = reg;
	txbd[0].bd_len = 2;
	txbd[0].bd_flags = TxBD_R;

	txbd[1].bd_addr = txbd[0].bd_addr + 2;
	txbd[1].bd_addr[0] = slave | I2C_READ;	/* Slave address | R/W. */
	txbd[1].bd_len = len + 1;
	txbd[1].bd_flags = TxBD_R | TxBD_S | TxBD_W | TxBD_L;

	immr->cpm_i2com |= I2COM_STR;		/* Start transfer. */
	while (rxbd[0].bd_flags & RxBD_E);

	bcopy(rxbd[0].bd_addr, buf, len);

	/* XXX error check */

	return 0;
}

int
i2c_write(slave, reg, buf, len)
	int slave, reg, len;
	void *buf;
{
	volatile struct mpc860dev *immr = get_immr();
	volatile struct cpm_bdesc *txbd, *rxbd;

	if (initted == 0) {
		initted = 1;
		i2c_init();
	}
	if (len > 0x80) {
		printf("i2c_write: len > 256\n");
		return -1;
	}

	/* Issue an INIT TX AND RX command. */
	immr->cpm_cpcr = 0x0011;
	while (immr->cpm_cpcr & 1);		/* Wait for completion. */

	rxbd = (void *)&immr->cpm_dpram[dpram];
	rxbd[0].bd_flags = RxBD_W;
	rxbd[0].bd_addr = 0;

	txbd = rxbd + 1;

	/* Write slave address and data. */
	txbd[0].bd_addr = (void *)(rxbd + 3);
	txbd[0].bd_addr[0] = slave | I2C_WRITE;	/* Slave address | R/W. */
	txbd[0].bd_addr[1] = reg;
	txbd[0].bd_len = len + 2;
	txbd[0].bd_flags = TxBD_R | TxBD_L;
	txbd[1].bd_addr = 0;
	txbd[1].bd_flags = TxBD_W;

	bcopy(buf, txbd[0].bd_addr + 2, len);

	immr->cpm_i2com |= I2COM_STR;		/* Start transfer. */
	while (txbd[0].bd_flags & TxBD_R);

	/* XXX error check */

	return 0;
}
