/*	$Id: scc.c,v 1.67 2012/06/01 13:03:11 tsubai Exp $	*/

/*-
 * Copyright (c) 2001,2002,2004 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/conf.h>
#include <sys/device.h>
#include <sys/fcntl.h>
#include <sys/ioctl.h>
#include <sys/kauth.h>
#include <sys/kernel.h>
#include <sys/systm.h>
#include <sys/tty.h>

#include <uvm/uvm_extern.h>
#include <dev/cons.h>

#include <machine/autoconf.h>
#include <machine/mpc860dev.h>
#include <machine/sysconf.h>

#define SCCUNIT(x)	(minor(x) & 0x7ffff)
#define SCCDIALOUT(x)	(minor(x) & 0x80000)

#define SCC_HW_CONSOLE	0x0001

struct scc_softc {
	struct device sc_dev;

	struct tty *sc_tty;
	int sc_hwflags;
	int sc_swflags;

	struct cpm_bdesc *sc_rxbd;
	struct cpm_bdesc *sc_txbd;

	u_char *sc_tba;
	u_int sc_tbc;
};

int scc_match(struct device *, struct cfdata *, void *);
void scc_attach(struct device *, struct device *, void *);

int scc_set_speed(int);
int scc_get_speed();
void sccstart(struct tty *);
int sccparam(struct tty *, struct termios *);

int sccintr(void *);

CFATTACH_DECL0(scc, sizeof(struct scc_softc), scc_match, scc_attach);

dev_type_open(sccopen);
dev_type_close(sccclose);
dev_type_read(sccread);
dev_type_write(sccwrite);
dev_type_ioctl(sccioctl);
dev_type_stop(sccstop);
dev_type_tty(scctty);
dev_type_poll(sccpoll);

const struct cdevsw scc_cdevsw = {
	sccopen, sccclose, sccread, sccwrite, sccioctl,
	sccstop, scctty, sccpoll, nommap, nokqfilter, D_TTY
};

int sccconsspeed = 9600;
int sccconscflag = TTYDEF_CFLAG|CLOCAL;

extern struct cfdriver scc_cd;

#define TXBUFLEN 256

/* UART events (SCCE/SCCM) */
#define SCCE_TX 0x0002
#define SCCE_RX 0x0001

/* SCC UART RxBD flags */
#define RxBD_E	0x8000	/* Empty */
#define RxBD_W	0x2000	/* Wrap */
#define RxBD_I	0x1000	/* Interrupt */
#define RxBD_C	0x0800	/* Control character */
#define RxBD_A	0x0400	/* Address */
#define RxBD_CM	0x0200	/* Continuous mode */
#define RxBD_ID	0x0100	/* Idle */
#define RxBD_AM	0x0080	/* Address Match */
#define RxBD_BR	0x0020	/* Break */
#define RxBD_FR	0x0010	/* Framing error */
#define RxBD_PR	0x0008	/* Parity error */
#define RxBD_OV	0x0002	/* Overrun */
#define RxBD_CD	0x0001	/* CD lost */

/* SCC UART TxBD flags */
#define TxBD_R	0x8000	/* Ready */
#define TxBD_W	0x2000	/* Wrap */
#define TxBD_I	0x1000	/* Interrupt */
#define TxBD_CR	0x0800	/* CTS report */
#define TxBD_A	0x0400	/* Address */
#define TxBD_CM	0x0200	/* Continuous mode */
#define TxBD_P	0x0100	/* Preamble */
#define TxBD_NS	0x0080	/* No stop bit or shaved stop bit sent */
#define TxBD_CT	0x0001	/* CTS lost */

/* SCC CP commands */
#define INIT_RX_AND_TX_PARAMS 0x0000
#define INIT_RX_PARAMS        0x0100
#define INIT_TX_PARAMS        0x0200
#define ENTER_HUNT_MODE       0x0300
#define STOP_TX               0x0400
#define GRACEFUL_STOP_TX      0x0500
#define RESTART_TX            0x0600
#define CLOSE_RX_BD           0x0700
#define SET_GROUP_ADDRESS     0x0800
#define RESET_BCS             0x0c00

int
scc_match(parent, cf, aux)
	struct device *parent;
	struct cfdata *cf;
	void *aux;
{
	struct confargs *ca = aux;

	if (strcmp(ca->ca_name, "scc") == 0)
		return 1;

	return 0;
}

void
scc_attach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
{
	struct scc_softc *sc = (void *)self;
	struct confargs *ca = aux;
	int vec = ca->ca_vec[0];
	volatile struct mpc860dev *immr = get_immr();
	volatile struct scc_uart *scc = (void *)&immr->cpm_dpram[DPRAM_SCC1];
	struct tty *tp;
	struct cpm_bdesc *txbd, *rxbd;
	int dpram;

	txbd = (void *)&immr->cpm_dpram[scc->tbase];
	rxbd = (void *)&immr->cpm_dpram[scc->rbase];
	sc->sc_rxbd = rxbd;
	sc->sc_txbd = txbd;

	dpram = dpram_alloc(TXBUFLEN, 2);
	if (dpram < 0) {
		printf(": failed to allocate DPRAM\n");
		return;
	}
	txbd->bd_addr = (void *)&immr->cpm_dpram[dpram];

	tp = ttymalloc();
	tp->t_oproc = sccstart;
	tp->t_param = sccparam;
	sc->sc_tty = tp;

	tty_attach(tp);

	if (1) {
		int maj;

		maj = cdevsw_lookup_major(&scc_cdevsw);
		if (maj == -1)
			panic("no scc in cdevsw?");
		cn_tab->cn_dev = makedev(maj, sc->sc_dev.dv_unit);
		sc->sc_hwflags |= SCC_HW_CONSOLE;
		printf(": (console)");
	}

	printf("\n");
	intr_establish(vec, IST_LEVEL, IPL_TTY, sccintr, sc);

	rxbd[0].bd_flags = RxBD_E | RxBD_I;
	rxbd[1].bd_flags = RxBD_E | RxBD_I | RxBD_W;
}

static inline void
eieio()
{
	asm volatile ("eieio; sync");
}

void
scc_uart_init()
{
	volatile struct mpc860dev *immr = get_immr();
	struct cpm_bdesc *txbd, *rxbd;
	struct scc_uart *scc = (void *)&immr->cpm_dpram[DPRAM_SCC1];
	int i, speed;

	speed = scc_get_speed();
	if (speed > 0)
		sccconsspeed = speed;
	scc_set_speed(sccconsspeed);

#define SICR_T1CS 0x00000007  /* Tx BRG[1-4]/CLK[1-4] */
#define SICR_R1CS 0x00000038  /* Rx BRG[1-4]/CLK[1-4] */
#define SICR_SC1  0x00000040
#define SICR_GR1  0x00000080
#define SICR_T2CS 0x00000700  /* Tx BRG[1-4]/CLK[1-4] */
#define SICR_R2CS 0x00003800  /* Rx BRG[1-4]/CLK[1-4] */
#define SICR_SC2  0x00004000
#define SICR_GR2  0x00008000
#define SICR_T3CS 0x00070000  /* Tx BRG[1-4]/CLK[5-8] */
#define SICR_R3CS 0x00380000  /* Rx BRG[1-4]/CLK[5-8] */
#define SICR_SC3  0x00400000
#define SICR_GR3  0x00800000
#define SICR_T4CS 0x07000000  /* Tx BRG[1-4]/CLK[5-8] */
#define SICR_R4CS 0x38000000  /* Rx BRG[1-4]/CLK[5-8] */
#define SICR_SC4  0x40000000
#define SICR_GR4  0x80000000

	/* siu_sdcr = 1; */

	txbd = (void *)&immr->cpm_dpram[scc->tbase];
	rxbd = (void *)&immr->cpm_dpram[scc->rbase];

	/* XXX rxbuf must be EVEN on SCC. */
	rxbd[0].bd_flags = 0;
	rxbd[0].bd_addr = (char *)&immr->cpm_dpram[SCC1_RXBUFADDR];
	rxbd[1].bd_flags = RxBD_W;
	rxbd[1].bd_addr = (char *)&immr->cpm_dpram[SCC1_RXBUFADDR + 2];

	txbd[0].bd_flags = TxBD_W;
	txbd[0].bd_addr = (char *)&immr->cpm_dpram[SCC1_TXBUFADDR];
	eieio();

	/* Initialize SCC1 parameters */
	immr->cpm_cpcr = CPCR_SCC1 | INIT_RX_AND_TX_PARAMS | 1;

	scc->rfcr = 0x10;
	scc->tfcr = 0x10;
	scc->mrblr = 1;		/* maximum receive length */
	scc->max_idl = 0;
	scc->brkcr = 0;
	scc->parec = 0;
	scc->frmec = 0;
	scc->brkec = 0;
	scc->uaddr1 = 0;
	scc->uaddr2 = 0;
	scc->toseq = 0;

	for (i = 0; i < 8; i++)
		scc->character[i] = 0x8000;

	scc->rccm = 0xc0ff;

	rxbd[0].bd_flags = RxBD_E;
	rxbd[1].bd_flags = RxBD_E | RxBD_W;

	/* Clear any previous events. */
	immr->cpm_scce1 = 0xffff;

	/* Enable Tx/Rx interrupts. */
	immr->cpm_sccm1 = SCCE_TX | SCCE_RX;

	/* Enable SCC1 interrupt. */
	immr->cpm_cimr |= 1 << CPM_IVEC_SCC1;

	/* Enable CPM interrupts. */
	immr->cpm_cicr = CICR_SCCPRI_1234|CICR_IRL(4)|CICR_IEN;

	/* XXX Always assert RTS. */
	immr->cpm_gsmr_h1 = GSMR_H_RFW | GSMR_H_RTSM;

	immr->cpm_gsmr_l1 = GSMR_L_TDCR_16 | GSMR_L_RDCR_16 | GSMR_L_MODE_UART;
	immr->cpm_psmr1 = PSMR_FLC | PSMR_CL8;
	eieio();

	/* Enable Tx and Rx */
	immr->cpm_gsmr_l1 |= GSMR_L_ENR | GSMR_L_ENT;
}

#define BRGC_RST   0x00020000	/* Reset BRG */
#define BRGC_EN    0x00010000	/* Enable BRG */
#define BRGC_EXTC  0x0000c000	/* External clock source */
#define   BRGC_BRGCLK 0x00000000
#define   BRGC_CLK2   0x00004000
#define   BRGC_CLK6   0x00008000
#define BRGC_ATB   0x00002000	/* Autobaud */
#define BRGC_CD	   0x00001ffe	/* Clock divider (mask) */
#define BRGC_DIV16 0x00000001	/* 1/16 divider */

int
scc_set_speed(bps)
	int bps;
{
	volatile struct mpc860dev *immr = get_immr();
	int clock, div, div16 = 0;

#define SCCR_DFBRG 0x00001800

	/* Assume using BRGCLK. */
	clock = SYSCLK;
	switch (immr->siu_sccr & SCCR_DFBRG) {
	case 0x00000000:
		clock /= 1;
		break;
	case 0x00000800:
		clock /= 4;
		break;
	case 0x00001000:
		clock /= 16;
		break;
	case 0x00001800:
		clock /= 64;
		break;
	}

	bps *= 16;

#define GSMR_L_TDCR 0x00030000
#define GSMR_L_RDCR 0x0000c000

#if 0
	switch (immr->cpm_gsmr_l1 & GSMR_L_TDCR) {
	case 0x00000000:
		break;
	case 0x00010000:
		bps *= 8;	/* 8x oversampling. */
		break;
	case 0x00020000:
		bps *= 16;	/* 16x oversampling. */
		break;
	case 0x00030000:
		bps *= 32;	/* 32x oversampling. */
		break;
	}
#endif

	/* baud = CLOCK / (1 or DIV16) / (CD + 1) / GSMR_L[TDCR]. */
	div = (clock + bps / 2) / bps - 1;

	if (div > 4095) {
		clock /= 16;
		div16 = BRGC_DIV16;
		div = (clock + bps / 2) / bps - 1;
	}

	if (div < 0 || div > 4095)
		return -1;

	immr->cpm_brgc2 = BRGC_EN | BRGC_BRGCLK | div << 1 | div16;
	return 0;
}

int
scc_get_speed()
{
	volatile struct mpc860dev *immr = get_immr();
	int bps, clock, div, div16, i;
	u_int brg;
	static int speeds[] = {
		115200, 57600, 38400, 19200, 9600, 4800, 2400, 1200, 600
	};

	/* Assume using BRGCLK. */
	clock = SYSCLK;
	switch (immr->siu_sccr & SCCR_DFBRG) {
	case 0x00000000:
		clock /= 1;
		break;
	case 0x00000800:
		clock /= 4;
		break;
	case 0x00001000:
		clock /= 16;
		break;
	case 0x00001800:
		clock /= 64;
		break;
	}

	switch (immr->cpm_sicr & SICR_T1CS) {
	case 0:
		brg = immr->cpm_brgc1;
		break;
	case 1:
		brg = immr->cpm_brgc2;
		break;
	case 2:
		brg = immr->cpm_brgc3;
		break;
	case 3:
		brg = immr->cpm_brgc4;
		break;
	default:
		return 0;
	}

	div16 = brg & BRGC_DIV16;
	div = (brg & BRGC_CD) >> 1;

	if (div16)
		clock /= 16;

	bps = clock / (div + 1);

	switch (immr->cpm_gsmr_l1 & GSMR_L_TDCR) {
	case 0x00000000:
		break;
	case 0x00010000:
		bps /= 8;	/* 8x oversampling. */
		break;
	case 0x00020000:
		bps /= 16;	/* 16x oversampling. */
		break;
	case 0x00030000:
		bps /= 32;	/* 32x oversampling. */
		break;
	}

	for (i = 0; i < 9; i++) {
		if (bps * 100 > speeds[i] * 102)
			continue;
		if (bps * 100 < speeds[i] * 98)
			continue;
		return speeds[i];
	}
	if (bps < 600)
		return bps;
	return 0;
}

void
scccnprobe(cp)
	struct consdev *cp;
{
	cp->cn_dev = makedev(0, 0);
	cp->cn_pri = CN_REMOTE;
}

void
scccninit(cp)
	struct consdev *cp;
{
	scc_uart_init();
}

int
scccngetc(dev)
	dev_t dev;
{
	volatile struct mpc860dev *immr = get_immr();
	volatile struct scc_uart *scc = (void *)&immr->cpm_dpram[DPRAM_SCC1];
	volatile struct cpm_bdesc *rxbd = (void *)&immr->cpm_dpram[scc->rbase];
	int save, c;

	/* XXX Disable/enable should be done in scccnpollc(). */
	save = immr->cpm_sccm1;
	immr->cpm_sccm1 &= ~SCCE_RX;

	for (;;) {
		if ((rxbd[0].bd_flags & RxBD_E) == 0) {
			c = *rxbd[0].bd_addr;
			rxbd[0].bd_flags |= RxBD_E;
			break;
		}
		if ((rxbd[1].bd_flags & RxBD_E) == 0) {
			c = *rxbd[1].bd_addr;
			rxbd[1].bd_flags |= RxBD_E;
			break;
		}
		sswt();
	}
	immr->cpm_scce1 = 0xffff;	/* Clear pending events. */
	immr->cpm_sccm1 = save;		/* Enable Rx interrupts again. */

	return c;
}

void
scccnputc(dev, c)
	dev_t dev;
	int c;
{
	volatile struct mpc860dev *immr = get_immr();
	volatile struct scc_uart *scc = (void *)&immr->cpm_dpram[DPRAM_SCC1];
	volatile struct cpm_bdesc *txbd = (void *)&immr->cpm_dpram[scc->tbase];

	while (txbd->bd_flags & TxBD_R);
	eieio();

	*txbd->bd_addr = c;
	txbd->bd_len = 1;

	eieio();
	txbd->bd_flags = TxBD_R | TxBD_W;

	while (txbd->bd_flags & TxBD_R);
}

struct consdev consdev_scc = {
	scccnprobe,
	scccninit,
	scccngetc,
	scccnputc,
	nullcnpollc,
	NULL
};

int
sccopen(dev, flag, mode, l)
	dev_t dev;
	int flag, mode;
	struct lwp *l;
{
	struct scc_softc *sc = (void *)device_lookup(&scc_cd, SCCUNIT(dev));
	struct tty *tp;
	int s, error;

	if (sc == NULL)
		return ENXIO;

	if (ISSET(sc->sc_dev.dv_flags, DVF_ACTIVE) == 0)
		return ENXIO;

	tp = sc->sc_tty;

	if (kauth_authorize_device_tty(l->l_cred, KAUTH_DEVICE_TTY_OPEN, tp))
		return EBUSY;

	s = spltty();

	/*
	 * Do the following if this is a first open.
	 */
	if (!ISSET(tp->t_state, TS_ISOPEN) && tp->t_wopen == 0) {
		struct termios t;

		tp->t_dev = dev;

		/*
		 * Initialize the termios status to the defaults.  Add in the
		 * sticky bits from TIOCSFLAGS.
		 */
		t.c_ispeed = 0;
		if (ISSET(sc->sc_hwflags, SCC_HW_CONSOLE)) {
			t.c_ospeed = sccconsspeed;
			t.c_cflag = sccconscflag;
		} else {
			t.c_ospeed = TTYDEF_SPEED;
			t.c_cflag = TTYDEF_CFLAG;
		}
		if (ISSET(sc->sc_swflags, TIOCFLAG_CLOCAL))
			SET(t.c_cflag, CLOCAL);
		if (ISSET(sc->sc_swflags, TIOCFLAG_CRTSCTS))
			SET(t.c_cflag, CRTSCTS);
		if (ISSET(sc->sc_swflags, TIOCFLAG_MDMBUF))
			SET(t.c_cflag, MDMBUF);

		/* Make sure comparam() will do something. */
		tp->t_ospeed = 0;
		sccparam(tp, &t);
		tp->t_iflag = TTYDEF_IFLAG;
		tp->t_oflag = TTYDEF_OFLAG;
		tp->t_lflag = TTYDEF_LFLAG;
		ttychars(tp);
		ttsetwater(tp);

#if notyet
		s2 = splserial();

		/*
		 * Turn on DTR.  We must always do this, even if carrier is not
		 * present, because otherwise we'd have to use TIOCSDTR
		 * immediately after setting CLOCAL, which applications do not
		 * expect.  We always assert DTR while the device is open
		 * unless explicitly requested to deassert it.
		 */
		scc_modem(sc, 1);

		splx(s2);
#endif
	}

	splx(s);

	error = ttyopen(tp, SCCDIALOUT(dev), ISSET(flag, O_NONBLOCK));
	if (error)
		goto bad;

	error = (*tp->t_linesw->l_open)(dev, tp);
	if (error)
		goto bad;

	return 0;

bad:
	if (!ISSET(tp->t_state, TS_ISOPEN) && tp->t_wopen == 0) {
		/*
		 * We failed to open the device, and nobody else had it opened.
		 * Clean up the state as appropriate.
		 */
//		scc_shutdown(sc);
	}

	return error;
}

int
sccclose(dev, flag, mode, l)
	dev_t dev;
	int flag, mode;
	struct lwp *l;
{
	struct scc_softc *sc = (void *)device_lookup(&scc_cd, SCCUNIT(dev));
	struct tty *tp = sc->sc_tty;

	(*tp->t_linesw->l_close)(tp, flag);
	ttyclose(tp);

	return 0;
}

int
sccread(dev, uio, flag)
	dev_t dev;
	struct uio *uio;
	int flag;
{
	struct scc_softc *sc = (void *)device_lookup(&scc_cd, SCCUNIT(dev));
	struct tty *tp = sc->sc_tty;

	return (*tp->t_linesw->l_read)(tp, uio, flag);
}

int
sccwrite(dev, uio, flag)
	dev_t dev;
	struct uio *uio;
	int flag;
{
	struct scc_softc *sc = (void *)device_lookup(&scc_cd, SCCUNIT(dev));
	struct tty *tp = sc->sc_tty;

	return (*tp->t_linesw->l_write)(tp, uio, flag);
}

int
sccpoll(dev, events, l)
	dev_t dev;
	int events;
	struct lwp *l;
{
	struct scc_softc *sc = (void *)device_lookup(&scc_cd, SCCUNIT(dev));
	struct tty *tp = sc->sc_tty;

	return (*tp->t_linesw->l_poll)(tp, events, l);
}

struct tty *
scctty(dev)
	dev_t dev;
{
	struct scc_softc *sc = (void *)device_lookup(&scc_cd, SCCUNIT(dev));
	struct tty *tp = sc->sc_tty;

	return tp;
}

int
sccioctl(dev, cmd, data, flag, l)
	dev_t dev;
	u_long cmd;
	void *data;
	int flag;
	struct lwp *l;
{
	struct scc_softc *sc = (void *)device_lookup(&scc_cd, SCCUNIT(dev));
	struct tty *tp = sc->sc_tty;
	int error;

	error = (*tp->t_linesw->l_ioctl)(tp, cmd, data, flag, l);
	if (error >= 0)
		return error;

	error = ttioctl(tp, cmd, data, flag, l);
	if (error >= 0)
		return error;

	return EPASSTHROUGH;
}

void
sccstop(tp, flag)
	struct tty *tp;
	int flag;
{
	struct scc_softc *sc = (void *)device_lookup(&scc_cd, SCCUNIT(tp->t_dev));

	if (ISSET(tp->t_state, TS_BUSY)) {
		/* Stop transmitting at the next chunk. */
		sc->sc_tbc = 0;
		if (!ISSET(tp->t_state, TS_TTSTOP))
			SET(tp->t_state, TS_FLUSH);
	}
}

static void
sccput(sc, cp, len)
	struct scc_softc *sc;
	char *cp;
	int len;
{
	volatile struct cpm_bdesc *txbd = sc->sc_txbd;

#ifdef DIAGNOSTIC
	if (txbd->bd_flags & TxBD_R)
		panic("sccput");
#endif

	bcopy(cp, txbd->bd_addr, len);
	txbd->bd_len = len;

	eieio();
	txbd->bd_flags = TxBD_R | TxBD_W | TxBD_I;
}

void
sccstart(tp)
	struct tty *tp;
{
	struct scc_softc *sc = (void *)device_lookup(&scc_cd, SCCUNIT(tp->t_dev));
	int n, s;

	s = spltty();
	if (ISSET(tp->t_state, TS_BUSY | TS_TIMEOUT | TS_TTSTOP))
		goto out;
	if (!ttypull(tp))
		return;

	sc->sc_tba = tp->t_outq.c_cf;
	sc->sc_tbc = ndqb(&tp->t_outq, 0);

	SET(tp->t_state, TS_BUSY);

	n = min(sc->sc_tbc, TXBUFLEN);
	sccput(sc, sc->sc_tba, n);
	sc->sc_tba += n;
	sc->sc_tbc -= n;

out:
	splx(s);
}

int
sccparam(tp, t)
	struct tty *tp;
	struct termios *t;
{
	struct scc_softc *sc = (void *)device_lookup(&scc_cd, SCCUNIT(tp->t_dev));
	volatile struct mpc860dev *immr = get_immr();

	if (ISSET(sc->sc_swflags, TIOCFLAG_SOFTCAR) ||
	    ISSET(sc->sc_hwflags, SCC_HW_CONSOLE)) {
		    SET(t->c_cflag, CLOCAL);
		    CLR(t->c_cflag, HUPCL);
	}

	tp->t_ispeed = 0;
	tp->t_ospeed = t->c_ospeed;
	tp->t_cflag = t->c_cflag;

	if (ISSET(t->c_cflag, CLOCAL))
		immr->cpm_pcso &= ~0x0020;	/* Ignore CD1 (PC10 = GPIO) */
	else
		immr->cpm_pcso |= 0x0020;	/* PC10 = CD1 */
	if (ISSET(t->c_cflag, CRTSCTS))
		immr->cpm_gsmr_h1 &= ~GSMR_H_RTSM;
	else {
		/* Always assert RTS. */
		immr->cpm_gsmr_h1 |= GSMR_H_RTSM;
	}
	scc_set_speed(t->c_ospeed);

	(*tp->t_linesw->l_modem)(tp, 1);

	return 0;
}

int
sccintr(v)
	void *v;
{
	struct scc_softc *sc = v;
	struct tty *tp = sc->sc_tty;
	int (*rint)(int, struct tty *) = tp->t_linesw->l_rint;
	volatile struct mpc860dev *immr = get_immr();
	volatile struct cpm_bdesc *txbd, *rxbd;
	int event, c;

	/* Check if this interrupt is mine. */
	if (CIVR_VN(immr->cpm_civr) != CPM_IVEC_SCC1)
		return 0;

	rxbd = sc->sc_rxbd;
	txbd = sc->sc_txbd;

	event = immr->cpm_scce1;	/* SCC event */

	/* printf("sccintr event = 0x%x\n", event); */

	if (event & SCCE_RX) {
		if ((rxbd[0].bd_flags & RxBD_E) == 0) {
			c = *rxbd[0].bd_addr;
			rxbd[0].bd_flags |= RxBD_E;
			rint(c, tp);
		}
		if ((rxbd[1].bd_flags & RxBD_E) == 0) {
			c = *rxbd[1].bd_addr;
			rxbd[1].bd_flags |= RxBD_E;
			rint(c, tp);
		}
	}
	if (event & SCCE_TX) {
		if (sc->sc_tbc > 0) {
			int n = min(sc->sc_tbc, TXBUFLEN);
			sccput(sc, sc->sc_tba, n);
			sc->sc_tba += n;
			sc->sc_tbc -= n;
		} else {
			struct tty *tp = sc->sc_tty;
			int n = sc->sc_tba - tp->t_outq.c_cf;

			CLR(tp->t_state, TS_BUSY);
			if (ISSET(tp->t_state, TS_FLUSH))
				CLR(tp->t_state, TS_FLUSH);
			else
				ndflush(&tp->t_outq, n);
			(*tp->t_linesw->l_start)(tp);
		}
	}

	immr->cpm_scce1 = event;

	/* Clear SCC1 interrupt condition. */
	immr->cpm_cisr = 1 << CPM_IVEC_SCC1;

	return 1;
}
