/*	$Id: smc.c,v 1.10 2012/06/01 13:03:11 tsubai Exp $	*/

/*-
 * Copyright (c) 2001,2002,2004 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/conf.h>
#include <sys/device.h>
#include <sys/fcntl.h>
#include <sys/ioctl.h>
#include <sys/kauth.h>
#include <sys/kernel.h>
#include <sys/systm.h>
#include <sys/tty.h>

#include <uvm/uvm_extern.h>
#include <dev/cons.h>

#include <machine/autoconf.h>
#include <machine/mpc860dev.h>
#include <machine/sysconf.h>

#define SMCUNIT(x)	(minor(x) & 0x7ffff)
#define SMCDIALOUT(x)	(minor(x) & 0x80000)

#define SMC_HW_CONSOLE	0x0001

struct smc_softc {
	struct device sc_dev;

	struct tty *sc_tty;
	int sc_hwflags;
	int sc_swflags;

	struct cpm_bdesc *sc_rxbd;
	struct cpm_bdesc *sc_txbd;

	u_char *sc_tba;
	u_int sc_tbc;
};

int smc_match(struct device *, struct cfdata *, void *);
void smc_attach(struct device *, struct device *, void *);

void smcstart(struct tty *);
int smcparam(struct tty *, struct termios *);

int smcintr(void *);
void smc_uart_init(void);

CFATTACH_DECL0(smc, sizeof(struct smc_softc), smc_match, smc_attach);

dev_type_open(smcopen);
dev_type_close(smcclose);
dev_type_read(smcread);
dev_type_write(smcwrite);
dev_type_ioctl(smcioctl);
dev_type_stop(smcstop);
dev_type_tty(smctty);
dev_type_poll(smcpoll);

const struct cdevsw smc_cdevsw = {
	smcopen, smcclose, smcread, smcwrite, smcioctl,
	smcstop, smctty, smcpoll, nommap, nokqfilter, D_TTY
};

int smcconsspeed = 9600;
int smcconscflag = TTYDEF_CFLAG|CLOCAL;

extern struct cfdriver smc_cd;

#define TXBUFLEN 256

/* UART events (SMCE/SMCM) */
#define SMCE_BRKE 0x40	/* Break end */
#define SMCE_BRK  0x10	/* Break character received */
#define SMCE_BSY  0x04	/* Busy condition */
#define SMCE_TX   0x02
#define SMCE_RX   0x01

/* SMC UART RxBD flags */
#define RxBD_E	0x8000	/* Empty */
#define RxBD_W	0x2000	/* Wrap */
#define RxBD_I	0x1000	/* Interrupt */
#define RxBD_CM	0x0200	/* Continuous mode */
#define RxBD_ID	0x0100	/* Idle */
#define RxBD_BR	0x0020	/* Break */
#define RxBD_FR	0x0010	/* Framing error */
#define RxBD_PR	0x0008	/* Parity error */
#define RxBD_OV	0x0002	/* Overrun */

/* SMC UART TxBD flags */
#define TxBD_R	0x8000	/* Ready */
#define TxBD_W	0x2000	/* Wrap */
#define TxBD_I	0x1000	/* Interrupt */
#define TxBD_CM	0x0200	/* Continuous mode */
#define TxBD_P	0x0100	/* Preamble */

/* SMC CP commands */
#define INIT_RX_AND_TX_PARAMS 0x0000
#define INIT_RX_PARAMS        0x0100
#define INIT_TX_PARAMS        0x0200
#define ENTER_HUNT_MODE       0x0300
#define STOP_TX               0x0400
#define GRACEFUL_STOP_TX      0x0500
#define RESTART_TX            0x0600
#define CLOSE_RX_BD           0x0700
#define SET_GROUP_ADDRESS     0x0800
#define RESET_BCS             0x0c00

#define SMCMR_REN 0x0001
#define SMCMR_TEN 0x0002

int
smc_match(parent, cf, aux)
	struct device *parent;
	struct cfdata *cf;
	void *aux;
{
	struct confargs *ca = aux;

	if (strcmp(ca->ca_name, "smc") == 0)
		return 1;

	return 0;
}

void
smc_attach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
{
	struct smc_softc *sc = (void *)self;
	struct cfdata *cf = self->dv_cfdata;
	int vec = cf->cf_vec;
	volatile struct mpc860dev *immr = get_immr();
	volatile struct smc_uart *uart = (void *)&immr->cpm_dpram[DPRAM_SMC1];
	struct tty *tp;
	struct cpm_bdesc *txbd, *rxbd;
	int dpram;

	txbd = (void *)&immr->cpm_dpram[uart->tbase];
	rxbd = (void *)&immr->cpm_dpram[uart->rbase];
	sc->sc_rxbd = rxbd;
	sc->sc_txbd = txbd;

	dpram = dpram_alloc(TXBUFLEN, 2);
	if (dpram < 0) {
		printf(": failed to allocate DPRAM\n");
		return;
	}
	txbd->bd_addr = (void *)&immr->cpm_dpram[dpram];

	tp = ttymalloc();
	tp->t_oproc = smcstart;
	tp->t_param = smcparam;
	sc->sc_tty = tp;

	tty_attach(tp);

	if (1) {
		int maj;

		maj = cdevsw_lookup_major(&smc_cdevsw);
		if (maj == -1)
			panic("no smc in cdevsw?");
		cn_tab->cn_dev = makedev(maj, sc->sc_dev.dv_unit);
		sc->sc_hwflags |= SMC_HW_CONSOLE;
		printf(": (console)");
	}

	printf("\n");
	delay(2000);
	smc_uart_init();

	intr_establish(vec, IST_LEVEL, IPL_TTY, smcintr, sc);
}

static inline void
eieio()
{
	asm volatile ("eieio; sync");
}

void
smc_uart_init()
{
	volatile struct mpc860dev *immr = get_immr();
	volatile struct smc_uart *uart = (void *)&immr->cpm_dpram[DPRAM_SMC1];
	struct cpm_bdesc *txbd, *rxbd;

	txbd = (void *)&immr->cpm_dpram[uart->tbase];
	rxbd = (void *)&immr->cpm_dpram[uart->rbase];

	immr->cpm_smcmr1 &= ~SMCMR_TEN;
	immr->cpm_smcmr1 &= ~SMCMR_REN;
	eieio();

	txbd[0].bd_flags = TxBD_I|TxBD_W;
	txbd[0].bd_addr = (char *)&immr->cpm_dpram[SMC_TXBUFADDR];

	rxbd[0].bd_flags = RxBD_E|RxBD_I;
	rxbd[0].bd_addr = (char *)&immr->cpm_dpram[SMC_RXBUFADDR];
	rxbd[1].bd_flags = RxBD_E|RxBD_I|RxBD_W;
	rxbd[1].bd_addr = (char *)&immr->cpm_dpram[SMC_RXBUFADDR + 2];

	immr->cpm_smce1 = 0xff;
	immr->cpm_smcm1 = SMCE_TX | SMCE_RX | SMCE_BRK;
	immr->cpm_cimr |= 1 << CPM_IVEC_SMC1;
        immr->cpm_cicr = CICR_SCCPRI_1234|CICR_IRL(4)|CICR_IEN;

	uart->mrblr = 1;
	uart->max_idl = 0;

	immr->cpm_cpcr = CPCR_SMC1 | INIT_RX_AND_TX_PARAMS | 1;
	eieio();

	immr->cpm_smcmr1 |= SMCMR_TEN;
	immr->cpm_smcmr1 |= SMCMR_REN;
}

#define BRGC_RST   0x00020000	/* Reset BRG */
#define BRGC_EN    0x00010000	/* Enable BRG */
#define BRGC_EXTC  0x0000c000	/* External clock source */
#define   BRGC_BRGCLK 0x00000000
#define   BRGC_CLK2   0x00004000
#define   BRGC_CLK6   0x00008000
#define BRGC_ATB   0x00002000	/* Autobaud */
#define BRGC_CD	   0x00001ffe	/* Clock divider (mask) */
#define BRGC_DIV16 0x00000001	/* 1/16 divider */

int
smc_set_speed(bps)
	int bps;
{
	return 0;
}

void
smccnprobe(cp)
	struct consdev *cp;
{
	cp->cn_dev = makedev(0, 0);
	cp->cn_pri = CN_REMOTE;
}

void
smccninit(cp)
	struct consdev *cp;
{
	smc_uart_init();
}

int
smccngetc(dev)
	dev_t dev;
{
	volatile struct mpc860dev *immr = get_immr();
	volatile struct smc_uart *uart = (void *)&immr->cpm_dpram[DPRAM_SMC1];
	volatile struct cpm_bdesc *rxbd = (void *)&immr->cpm_dpram[uart->rbase];
	u_int save, c;

	/* XXX Disable/enable should be done in smccnpollc(). */
	save = immr->cpm_smcm1;
	immr->cpm_smcm1 &= ~SMCE_RX;

	for (;;) {
		if ((rxbd[0].bd_flags & RxBD_E) == 0) {
			c = *rxbd[0].bd_addr;
			rxbd[0].bd_flags = RxBD_E | RxBD_I;
			break;
		}
		if ((rxbd[1].bd_flags & RxBD_E) == 0) {
			c = *rxbd[1].bd_addr;
			rxbd[1].bd_flags = RxBD_E | RxBD_I | RxBD_W;
			break;
		}
		sswt();
	}
	immr->cpm_smce1 = 0xff;		/* Clear pending events. */
	immr->cpm_smcm1 = save;		/* Enable Rx interrupts again. */

	return c;
}

void
smccnputc(dev, c)
	dev_t dev;
	int c;
{
	volatile struct mpc860dev *immr = get_immr();
	volatile struct smc_uart *uart = (void *)&immr->cpm_dpram[DPRAM_SMC1];
	volatile struct cpm_bdesc *bd = (void *)&immr->cpm_dpram[uart->tbase];

	while (bd->bd_flags & TxBD_R);
	eieio();

	*bd->bd_addr = c;
	bd->bd_len = 1;

	eieio();
	bd->bd_flags |= TxBD_R;

	while (bd->bd_flags & TxBD_R);
}

struct consdev consdev_smc = {
	smccnprobe,
	smccninit,
	smccngetc,
	smccnputc,
	nullcnpollc,
	NULL
};

int
smcopen(dev, flag, mode, l)
	dev_t dev;
	int flag, mode;
	struct lwp *l;
{
	struct smc_softc *sc = (void *)device_lookup(&smc_cd, SMCUNIT(dev));
	struct tty *tp;
	int s, error;

	if (sc == NULL)
		return ENXIO;

	if (ISSET(sc->sc_dev.dv_flags, DVF_ACTIVE) == 0)
		return ENXIO;

	tp = sc->sc_tty;

	if (kauth_authorize_device_tty(l->l_cred, KAUTH_DEVICE_TTY_OPEN, tp))
		return EBUSY;

	s = spltty();

	/*
	 * Do the following if this is a first open.
	 */
	if (!ISSET(tp->t_state, TS_ISOPEN) && tp->t_wopen == 0) {
		struct termios t;

		tp->t_dev = dev;

		/*
		 * Initialize the termios status to the defaults.  Add in the
		 * sticky bits from TIOCSFLAGS.
		 */
		t.c_ispeed = 0;
		if (ISSET(sc->sc_hwflags, SMC_HW_CONSOLE)) {
			t.c_ospeed = smcconsspeed;
			t.c_cflag = smcconscflag;
		} else {
			t.c_ospeed = TTYDEF_SPEED;
			t.c_cflag = TTYDEF_CFLAG;
		}
		if (ISSET(sc->sc_swflags, TIOCFLAG_CLOCAL))
			SET(t.c_cflag, CLOCAL);
		if (ISSET(sc->sc_swflags, TIOCFLAG_CRTSCTS))
			SET(t.c_cflag, CRTSCTS);
		if (ISSET(sc->sc_swflags, TIOCFLAG_MDMBUF))
			SET(t.c_cflag, MDMBUF);

		/* Make sure comparam() will do something. */
		tp->t_ospeed = 0;
		smcparam(tp, &t);
		tp->t_iflag = TTYDEF_IFLAG;
		tp->t_oflag = TTYDEF_OFLAG;
		tp->t_lflag = TTYDEF_LFLAG;
		ttychars(tp);
		ttsetwater(tp);

#if notyet
		s2 = splserial();

		/*
		 * Turn on DTR.  We must always do this, even if carrier is not
		 * present, because otherwise we'd have to use TIOCSDTR
		 * immediately after setting CLOCAL, which applications do not
		 * expect.  We always assert DTR while the device is open
		 * unless explicitly requested to deassert it.
		 */
		smc_modem(sc, 1);

		splx(s2);
#endif
	}

	splx(s);

	error = ttyopen(tp, SMCDIALOUT(dev), ISSET(flag, O_NONBLOCK));
	if (error)
		goto bad;

	error = (*tp->t_linesw->l_open)(dev, tp);
	if (error)
		goto bad;

	return 0;

bad:
	if (!ISSET(tp->t_state, TS_ISOPEN) && tp->t_wopen == 0) {
		/*
		 * We failed to open the device, and nobody else had it opened.
		 * Clean up the state as appropriate.
		 */
//		smc_shutdown(sc);
	}

	return error;
}

int
smcclose(dev, flag, mode, l)
	dev_t dev;
	int flag, mode;
	struct lwp *l;
{
	struct smc_softc *sc = (void *)device_lookup(&smc_cd, SMCUNIT(dev));
	struct tty *tp = sc->sc_tty;

	(*tp->t_linesw->l_close)(tp, flag);
	ttyclose(tp);

	return 0;
}

int
smcread(dev, uio, flag)
	dev_t dev;
	struct uio *uio;
	int flag;
{
	struct smc_softc *sc = (void *)device_lookup(&smc_cd, SMCUNIT(dev));
	struct tty *tp = sc->sc_tty;

	return (*tp->t_linesw->l_read)(tp, uio, flag);
}

int
smcwrite(dev, uio, flag)
	dev_t dev;
	struct uio *uio;
	int flag;
{
	struct smc_softc *sc = (void *)device_lookup(&smc_cd, SMCUNIT(dev));
	struct tty *tp = sc->sc_tty;

	return (*tp->t_linesw->l_write)(tp, uio, flag);
}

int
smcpoll(dev, events, l)
	dev_t dev;
	int events;
	struct lwp *l;
{
	struct smc_softc *sc = (void *)device_lookup(&smc_cd, SMCUNIT(dev));
	struct tty *tp = sc->sc_tty;

	return (*tp->t_linesw->l_poll)(tp, events, l);
}

struct tty *
smctty(dev)
	dev_t dev;
{
	struct smc_softc *sc = (void *)device_lookup(&smc_cd, SMCUNIT(dev));
	struct tty *tp = sc->sc_tty;

	return tp;
}

int
smcioctl(dev, cmd, data, flag, l)
	dev_t dev;
	u_long cmd;
	void *data;
	int flag;
	struct lwp *l;
{
	struct smc_softc *sc = (void *)device_lookup(&smc_cd, SMCUNIT(dev));
	struct tty *tp = sc->sc_tty;
	int error;

	error = (*tp->t_linesw->l_ioctl)(tp, cmd, data, flag, l);
	if (error >= 0)
		return error;

	error = ttioctl(tp, cmd, data, flag, l);
	if (error >= 0)
		return error;

	return EPASSTHROUGH;
}

void
smcstop(tp, flag)
	struct tty *tp;
	int flag;
{
	struct smc_softc *sc = (void *)device_lookup(&smc_cd, SMCUNIT(tp->t_dev));

	if (ISSET(tp->t_state, TS_BUSY)) {
		/* Stop transmitting at the next chunk. */
		sc->sc_tbc = 0;
		if (!ISSET(tp->t_state, TS_TTSTOP))
			SET(tp->t_state, TS_FLUSH);
	}
}

static void
smcput(sc, cp, len)
	struct smc_softc *sc;
	char *cp;
	int len;
{
	volatile struct cpm_bdesc *txbd = sc->sc_txbd;

#ifdef DIAGNOSTIC
	if (txbd->bd_flags & TxBD_R)
		panic("smcput");
#endif

	bcopy(cp, txbd->bd_addr, len);
	txbd->bd_len = len;

	eieio();
	txbd->bd_flags = TxBD_R | TxBD_W | TxBD_I;
}

void
smcstart(tp)
	struct tty *tp;
{
	struct smc_softc *sc = (void *)device_lookup(&smc_cd, SMCUNIT(tp->t_dev));
	int n, s;

	s = spltty();
	if (ISSET(tp->t_state, TS_BUSY | TS_TIMEOUT | TS_TTSTOP))
		goto out;
	if (!ttypull(tp))
		return;

	sc->sc_tba = tp->t_outq.c_cf;
	sc->sc_tbc = ndqb(&tp->t_outq, 0);

	SET(tp->t_state, TS_BUSY);

	n = min(sc->sc_tbc, TXBUFLEN);
	smcput(sc, sc->sc_tba, n);
	sc->sc_tba += n;
	sc->sc_tbc -= n;

out:
	splx(s);
}

int
smcparam(tp, t)
	struct tty *tp;
	struct termios *t;
{
	struct smc_softc *sc = (void *)device_lookup(&smc_cd, SMCUNIT(tp->t_dev));

	if (ISSET(sc->sc_swflags, TIOCFLAG_SOFTCAR) ||
	    ISSET(sc->sc_hwflags, SMC_HW_CONSOLE)) {
		    SET(t->c_cflag, CLOCAL);
		    CLR(t->c_cflag, HUPCL);
	}

	tp->t_ispeed = 0;
	tp->t_ospeed = t->c_ospeed;
	tp->t_cflag = t->c_cflag;

#if 0	/* XXX SMC does not support RTS, CTS, and CD. */
	if (ISSET(t->c_cflag, CRTSCTS))
		immr->cpm_gsmr_h1 &= ~GSMR_H_RTSM;
	else {
		/* Always assert RTS. */
		immr->cpm_gsmr_h1 |= GSMR_H_RTSM;
	}
#endif
	smc_set_speed(t->c_ospeed);

	(*tp->t_linesw->l_modem)(tp, 1);

	return 0;
}

int
smcintr(v)
	void *v;
{
	struct smc_softc *sc = v;
	struct tty *tp = sc->sc_tty;
	int (*rint)(int, struct tty *) = tp->t_linesw->l_rint;
	volatile struct mpc860dev *immr = get_immr();
	volatile struct cpm_bdesc *txbd, *rxbd;
	int event, c;

	/* Check if this interrupt is mine. */
	if (CIVR_VN(immr->cpm_civr) != CPM_IVEC_SMC1)
		return 0;

	rxbd = sc->sc_rxbd;
	txbd = sc->sc_txbd;

	event = immr->cpm_smce1;	/* SMC event */

	if (event & SMCE_BRK)
		Debugger();

	if (event & SMCE_RX) {
		if ((rxbd[0].bd_flags & RxBD_E) == 0) {
			c = *rxbd[0].bd_addr;
			rxbd[0].bd_flags |= RxBD_E;
			rint(c, tp);
		}
		if ((rxbd[1].bd_flags & RxBD_E) == 0) {
			c = *rxbd[1].bd_addr;
			rxbd[1].bd_flags |= RxBD_E;
			rint(c, tp);
		}
	}
	if (event & SMCE_TX) {
		if (sc->sc_tbc > 0) {
			int n = min(sc->sc_tbc, TXBUFLEN);
			smcput(sc, sc->sc_tba, n);
			sc->sc_tba += n;
			sc->sc_tbc -= n;
		} else {
			struct tty *tp = sc->sc_tty;
			int n = sc->sc_tba - tp->t_outq.c_cf;

			CLR(tp->t_state, TS_BUSY);
			if (ISSET(tp->t_state, TS_FLUSH))
				CLR(tp->t_state, TS_FLUSH);
			else
				ndflush(&tp->t_outq, n);
			(*tp->t_linesw->l_start)(tp);
		}
	}

	immr->cpm_smce1 = event;

	/* Clear SMC1 interrupt condition. */
	immr->cpm_cisr = 1 << CPM_IVEC_SMC1;

	return 1;
}
