/*	$Id: spi.c,v 1.1 2015/06/25 17:40:15 tsubai Exp $	*/

/*-
 * Copyright (c) 2001,2002 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/systm.h>

#include <machine/autoconf.h>
#include <machine/mpc860dev.h>
#include <machine/sysconf.h>

#define TxBD_R	 0x8000		/* Ready */
#define TxBD_W	 0x2000		/* Wrap */
#define TxBD_I	 0x1000		/* Interrupt */
#define TxBD_L	 0x0800		/* Last */
#define TxBD_CM	 0x0200		/* Continuous mode */
#define TxBD_UN	 0x0002		/* Underrun */
#define TxBD_ME	 0x0001		/* Multimaster error */

#define RxBD_E	 0x8000		/* Empty */
#define RxBD_W	 0x2000		/* Wrap */
#define RxBD_I	 0x1000		/* Interrupt */
#define RxBD_L	 0x0800		/* Last */
#define RxBD_CM	 0x0200		/* Continuous mode */
#define RxBD_OV	 0x0002		/* Overrun */
#define RxBD_ME	 0x0001		/* Multimaster error */

/* SPI CP commands */
#define INIT_RX_AND_TX_PARAMS 0x0000
#define INIT_RX_PARAMS        0x0100
#define INIT_TX_PARAMS        0x0200
#define CLOSE_RX_BD           0x0700

#define SPI_BUFLEN 32

static int spi_init(void);
static int initted;

int
spi_init()
{
	volatile struct mpc860dev *immr = get_immr();
	volatile struct spi_pram *spi = (void *)&immr->cpm_dpram[DPRAM_SPI];
	struct cpm_bdesc *txbd, *rxbd;
	int dpram;

	if (initted)
		return 0;

	/*
	 * Configure SPI pins.
	 * Note: SPISEL is *not* connected to SPI.
	 */
	immr->cpm_pbpar |=  0x0000e;
	immr->cpm_pbdir |=  0x0000e;
	immr->cpm_pbodr &= ~0x0000e;

	dpram = dpram_alloc(16*2 + SPI_BUFLEN*2, 8);
	if (dpram < 0)
		panic("spi_init: failed to allocate DPRAM\n");

	spi->tbase = dpram;
	spi->rbase = dpram + 16;
	spi->tfcr = 0x18;
	spi->rfcr = 0x18;
	spi->mrblr = SPI_BUFLEN;
	dpram += 32;

	immr->cpm_cpcr = CPCR_SPI | INIT_RX_AND_TX_PARAMS | 1;
	while (immr->cpm_cpcr & 1);

#if 0
	/* Initialize SDMA. */
	immr->siu_sdcr = 0x0001;
#endif

	txbd = (void *)&immr->cpm_dpram[spi->tbase];
	rxbd = (void *)&immr->cpm_dpram[spi->rbase];
	txbd->bd_addr = (void *)&immr->cpm_dpram[dpram];
	rxbd->bd_addr = (void *)&immr->cpm_dpram[dpram + SPI_BUFLEN];

	/* Clear and mask all events. */
	immr->cpm_spie = 0xff;
	immr->cpm_spim = 0;

	immr->cpm_spmode = SPMODE_CI | SPMODE_CP | SPMODE_DIV16 |
	    SPMODE_REV | SPMODE_MS | 0x70 | 0xf;

	initted = 1;
	return 0;
}

int
spi_xfer(buf, len)
	void *buf;
	int len;
{
	volatile struct mpc860dev *immr = get_immr();
	volatile struct spi_pram *spi = (void *)&immr->cpm_dpram[DPRAM_SPI];
	volatile struct cpm_bdesc *txbd, *rxbd;
	u_int event;

	spi_init();

	if (len >= SPI_BUFLEN)
		panic("spi_xfer: too long");

	txbd = (void *)&immr->cpm_dpram[spi->tbase];
	rxbd = (void *)&immr->cpm_dpram[spi->rbase];

	bcopy(buf, txbd->bd_addr, len);
	bzero(rxbd->bd_addr, len);

	txbd->bd_len = len;
	txbd->bd_flags = TxBD_R | TxBD_W | TxBD_I | TxBD_L;
	rxbd->bd_len = 0;
	rxbd->bd_flags = RxBD_E | RxBD_W | RxBD_I;

	immr->cpm_spmode |= SPMODE_EN;
	immr->cpm_spie = 0xff;
	immr->cpm_spcom = SPCOM_STR;
	while (rxbd->bd_flags & RxBD_E);

	immr->cpm_spmode &= ~SPMODE_EN;
	event = immr->cpm_spie;
	if (event & SPIE_MME)
		printf("spi_xfer: multimaster error\n");
	if (event & SPIE_BSY)
		printf("spi_xfer: busy\n");

	bcopy(rxbd->bd_addr, buf, len);
	return 0;
}
