/*	$Id: clock.c,v 1.28 2015/06/25 23:56:47 tsubai Exp $	*/

/*
 * Copyright (C) 1995, 1996 Wolfgang Solfrank.
 * Copyright (C) 1995, 1996 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/kernel.h>
#include <sys/sysctl.h>
#include <sys/systm.h>
#include <sys/timetc.h>

#include <uvm/uvm_extern.h>

#include <machine/cpufunc.h>
#include <machine/mpc860dev.h>
#include <machine/sysconf.h>

/*
 * Initially we assume TMBCLK is 4 MHz.
 */
static u_long ticks_per_sec = TIMEBASE_FREQ / 4;
static u_long ns_per_tick = 1000000000 / (TIMEBASE_FREQ / 4);
static long ticks_per_intr;

volatile int tickspending;

u_int get_timecount_mftb(struct timecounter *);
void rtc_attach(void);

static struct timecounter tc_mftb = {
	get_timecount_mftb,	/* get_timecount */
	NULL,			/* poll_pps */
	~0,			/* counter_mask */
	0,			/* frequency */
	"timebase",		/* name */
	10			/* quality */
};

u_int
get_timecount_mftb(tc)
	struct timecounter *tc;
{
	u_int tb;

	asm volatile ("mftb %0" : "=r"(tb));
	return tb;
}

static int
sysctl_machdep_timebase(SYSCTLFN_ARGS)
{
	struct sysctlnode node = *rnode;
	int msr, error;

	node.sysctl_idata = ticks_per_sec;
	error = sysctl_lookup(SYSCTLFN_CALL(&node));
	if (error || newp == NULL)
		return error;

	if (node.sysctl_idata < 500*1000)
		return EINVAL;
	if (node.sysctl_idata > 2000*1000)
		return EINVAL;

	msr = mfmsr();
	mtmsr(msr & ~PSL_EE);
	asm volatile ("isync");
	ticks_per_sec = node.sysctl_idata;
	ticks_per_intr = (ticks_per_sec + hz / 2) / hz;
	tc_mftb.tc_frequency = ticks_per_sec;
	asm volatile ("sync");
	mtmsr(msr);
	return 0;
}

void
cpu_initclocks()
{
	rtc_attach();

	ticks_per_intr = (ticks_per_sec + hz / 2) / hz;
	mtdec(ticks_per_intr);

	tc_mftb.tc_frequency = ticks_per_sec;
	tc_init(&tc_mftb);

	/* Create machdep.timebase sysctl. */
	sysctl_createv(NULL, 0, NULL, NULL, CTLFLAG_IMMEDIATE|CTLFLAG_READWRITE,
	    CTLTYPE_INT, "timebase", NULL, sysctl_machdep_timebase, 0, NULL, 0,
	    CTL_MACHDEP, CTL_CREATE, CTL_EOL);
}

void
decr_intr(frame)
	struct clockframe *frame;
{
	int tick, nticks, pri;

	/* Check whether we are initialized. */
	if (!ticks_per_intr)
		return;

	/*
	 * Based on the actual time delay since the last decrementer reload,
	 * we arrange for earlier interrupt next time.
	 */
	tick = mfdec();
	for (nticks = 0; tick < 0; nticks++)
		tick += ticks_per_intr;
	mtdec(tick);

	curcpu()->ci_data.cpu_nintr++;
	intrcnt[CNT_CLOCK]++;

	pri = splclock();
	if (pri & (1 << SPL_CLOCK))
		tickspending += nticks;
	else {
		nticks += tickspending;
		tickspending = 0;

		/*
		 * Reenable interrupts
		 */
		mtmsr(mfmsr() | PSL_EE);

		/*
		 * Do standard timer interrupt stuff.
		 */
		do {
			hardclock(frame);
		} while (--nticks > 0);

		/* Service software watchdog timer. */
		sswt();
	}
	splx(pri);
}

void
timebase_init()
{
	volatile struct mpc860dev *immr = get_immr();

	immr->siu_tbk = 0x55ccaa33;	/* Unlock dec. */
	mtdec(0x7fffffff);

	ticks_per_sec = TIMEBASE_FREQ / 4;
	ns_per_tick = 1000000000 / ticks_per_sec;

	/* Start timebase counter. */
	immr->siu_tbscrk = 0x55ccaa33;
	immr->siu_tbscr = 0x0001;

}

/*
 * Wait for about n microseconds (at least!).
 */
void
delay(n)
	unsigned n;
{
	u_quad_t tb;
	u_long tbh, tbl, scratch;

	/* Service watch dog timer before long wait while boot sequence. */
	if (/* n > 100000 && */ cold)
		sswt();

	tb = mftb();
	tb += (n * 1000 + ns_per_tick - 1) / ns_per_tick;
	tbh = tb >> 32;
	tbl = tb;

	asm volatile ("			\n\
	1:	mftbu	%0		\n\
		sub.	%0,%0,%1	\n\
		blt	1b		\n\
		bgt	2f		\n\
		mftb	%0		\n\
		cmplw	%0,%2		\n\
		blt	1b		\n\
	2:				\n"
		: "=&r"(scratch) : "r"(tbh), "r"(tbl));
}

/*
 * Nothing to do.
 */
void
setstatclockrate(arg)
	int arg;
{
	/* Do nothing */
}
