/*	$Id: extintr.c,v 1.15 2012/06/07 11:34:06 tsubai Exp $	*/

/*-
 * Copyright (c) 1995 Per Fogelstrom
 * Copyright (c) 1993, 1994 Charles M. Hannum.
 * Copyright (c) 1990 The Regents of the University of California.
 * All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * William Jolitz and Don Ahn.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)isa.c	7.2 (Berkeley) 5/12/91
 */

#include <sys/param.h>
#include <sys/malloc.h>
#include <sys/kernel.h>

#include <uvm/uvm_extern.h>

#include <machine/autoconf.h>
#include <machine/cpufunc.h>
#include <machine/intr.h>
#include <machine/mpc860dev.h>
#include <machine/pio.h>
#include <machine/psl.h>

/*
 * Interrupt handler chains.  intr_establish() inserts a handler into
 * the list.  The handler is called with its (single) argument.
 */
struct intrhand {
	int	(*ih_fun)(void *);
	void	*ih_arg;
	u_long	ih_count;
	struct	intrhand *ih_next;
	int	ih_level;
	int	ih_irq;
};

char *intr_typename(int);
int fakeintr(void *);

static void do_pending_int(void);

int imask[NIPL];
int intrtype[ICU_LEN], intrmask[ICU_LEN], intrlevel[ICU_LEN];
struct intrhand *intrhand[ICU_LEN], *soft_intrhand[5];

volatile int cpl = 0xffffffff, ipending;

#define HWMASK	(((1 << ICU_LEN) - 1) << (32 - ICU_LEN))

/*
 * Recalculate the interrupt masks from scratch.
 * We could code special registry and deregistry versions of this function that
 * would be faster, but the code would be nastier, and we don't expect this to
 * happen very much anyway.
 */
void
intr_calculatemasks()
{
	volatile struct mpc860dev *immr = get_immr();
	int irq, level, levels;
	u_int irqs;
	struct intrhand *q;

	/* Determine which IRQs are actually in use. */
	irqs = 0;
	for (irq = 0; irq < ICU_LEN; irq++)
		if (intrhand[irq])
			irqs |= 0x80000000 >> irq;
	for (level = 0; level < NIPL; level++)
		imask[level] = ~irqs & HWMASK;

	/* Figure out which levels each IRQ uses. */
	for (irq = 0; irq < ICU_LEN; irq++) {
		levels = 0;
		for (q = intrhand[irq]; q; q = q->ih_next)
			levels |= 1 << q->ih_level;
		intrlevel[irq] = levels;
	}

	/* Then figure out which IRQs use each level. */
	for (level = 0; level < NIPL; level++) {
		irqs = 0;
		for (irq = 0; irq < ICU_LEN; irq++)
			if (intrlevel[irq] & (1 << level))
				irqs |= 0x80000000 >> irq;
		imask[level] |= irqs;
	}

	/*
	 * IPL_CLOCK should mask clock interrupt even if interrupt handler
	 * is not registered.
	 */
	imask[IPL_CLOCK] |= 1 << SPL_CLOCK;

	/*
	 * Initialize soft interrupt masks to block themselves.
	 */
	imask[IPL_SOFTCLOCK] |= 1 << SIR_CLOCK;
	imask[IPL_SOFTBIO] |= 1 << SIR_BIO;
	imask[IPL_SOFTNET] |= 1 << SIR_NET;
	imask[IPL_SOFTSERIAL] |= 1 << SIR_SERIAL;

	/*
	 * Enforce a hierarchy that gives slow devices a better chance at not
	 * dropping data.
	 */
	imask[IPL_SOFTCLOCK] |= imask[IPL_NONE];
	imask[IPL_SOFTBIO] |= imask[IPL_SOFTCLOCK];
	imask[IPL_SOFTNET] |= imask[IPL_SOFTBIO];
	imask[IPL_BIO] |= imask[IPL_SOFTNET];
	imask[IPL_NET] |= imask[IPL_BIO];
	imask[IPL_SOFTSERIAL] |= imask[IPL_NET];
	imask[IPL_TTY] |= imask[IPL_SOFTSERIAL];

	/*
	 * There are tty, network and disk drivers that use free() at interrupt
	 * time, so vm > (tty | net | bio).
	 */
	imask[IPL_VM] |= imask[IPL_TTY];

	imask[IPL_AUDIO] |= imask[IPL_VM];

	/*
	 * Since run queues may be manipulated by both the statclock and tty,
	 * network, and disk drivers, clock > vm.
	 */
	imask[IPL_CLOCK] |= imask[IPL_AUDIO];

	/*
	 * IPL_HIGH must block everything that can manipulate a run queue.
	 */
	imask[IPL_HIGH] |= imask[IPL_CLOCK];

	/*
	 * We need serial drivers to run at the absolute highest priority to
	 * avoid overruns, so serial > high.
	 */
	imask[IPL_SERIAL] |= imask[IPL_HIGH];

	/* And eventually calculate the complete masks. */
	for (irq = 0; irq < ICU_LEN; irq++) {
		irqs = 0x80000000 >> irq;
		for (q = intrhand[irq]; q; q = q->ih_next)
			irqs |= imask[q->ih_level];
		intrmask[irq] = irqs;
	}

	/* Configure level/edge and wakeup mask. */
	irqs = 0;
	for (irq = 0; irq < ICU_LEN; irq++) {
		/* Skip internal intrs. */
		if (irq & 1)
			continue;

		if (intrtype[irq] == IST_EDGE)
			irqs |= 0x80000000 >> irq;
		if (intrhand[irq])
			irqs |= 0x40000000 >> irq;
	}
	immr->siu_siel = irqs;

#if 0
printf("intr_calculatemask: none %08x, bio %08x, net %08x, high %08x\n",
  imask[IPL_NONE], imask[IPL_BIO], imask[IPL_NET], imask[IPL_HIGH]);
#endif
}

int
fakeintr(arg)
	void *arg;
{

	return 0;
}

#define LEGAL_IRQ(x)	((x) >= 0 && (x) < ICU_LEN)

char *
intr_typename(type)
	int type;
{

	switch (type) {
        case IST_NONE:
		return "none";
        case IST_PULSE:
		return "pulsed";
        case IST_EDGE:
		return "edge-triggered";
        case IST_LEVEL:
		return "level-triggered";
	default:
		panic("intr_typename: invalid type %d", type);
	}
}

/*
 * Register an interrupt handler.
 */
void *
intr_establish(irq, type, level, ih_fun, ih_arg)
	int irq;
	int type;
	int level;
	int (*ih_fun)(void *);
	void *ih_arg;
{
	struct intrhand **p, *q, *ih;
	static struct intrhand fakehand = {fakeintr};

	/* no point in sleeping unless someone can free memory. */
	ih = malloc(sizeof *ih, M_DEVBUF, cold ? M_NOWAIT : M_WAITOK);
	if (ih == NULL)
		panic("intr_establish: can't malloc handler info");

	if (!LEGAL_IRQ(irq) || type == IST_NONE)
		panic("intr_establish: bogus irq or type");

	switch (intrtype[irq]) {
	case IST_NONE:
		intrtype[irq] = type;
		break;
	case IST_EDGE:
	case IST_LEVEL:
		if (type == intrtype[irq])
			break;
	case IST_PULSE:
		if (type != IST_NONE)
			panic("intr_establish: can't share %s with %s",
			    intr_typename(intrtype[irq]),
			    intr_typename(type));
		break;
	}

	/*
	 * Figure out where to put the handler.
	 * This is O(N^2), but we want to preserve the order, and N is
	 * generally small.
	 */
	for (p = &intrhand[irq]; (q = *p) != NULL; p = &q->ih_next);

	/*
	 * Actually install a fake handler momentarily, since we might be doing
	 * this with interrupts enabled and don't want the real routine called
	 * until masking is set up.
	 */
	fakehand.ih_level = level;
	*p = &fakehand;

	intr_calculatemasks();

	/*
	 * Poke the real handler in now.
	 */
	ih->ih_fun = ih_fun;
	ih->ih_arg = ih_arg;
	ih->ih_count = 0;
	ih->ih_next = NULL;
	ih->ih_level = level;
	ih->ih_irq = irq;
	*p = ih;

	return ih;
}

/*
 * Deregister an interrupt handler.
 */
void
intr_disestablish(arg)
	void *arg;
{
	struct intrhand *ih = arg;
	int irq = ih->ih_irq;
	struct intrhand **p, *q;

	if (!LEGAL_IRQ(irq))
		panic("intr_disestablish: bogus irq");

	/*
	 * Remove the handler from the chain.
	 * This is O(n^2), too.
	 */
	for (p = &intrhand[irq]; (q = *p) != NULL && q != ih; p = &q->ih_next);

	if (q)
		*p = q->ih_next;
	else
		panic("intr_disestablish: handler not registered");
	free(ih, M_DEVBUF);

	intr_calculatemasks();

	if (intrhand[irq] == NULL)
		intrtype[irq] = IST_NONE;
}

void *
softintr_establish(level, fun, arg)
	int level;
	void (*fun)(void *);
	void *arg;
{
	struct intrhand *ih;
	int irq, s;

	switch (level) {
	case IPL_SOFTCLOCK:
		irq = SIR_CLOCK;
		break;
	case IPL_SOFTBIO:
		irq = SIR_BIO;
		break;
	case IPL_SOFTNET:
		irq = SIR_NET;
		break;
	case IPL_SOFTSERIAL:
		irq = SIR_SERIAL;
		break;
	default:
		panic("softintr_establish");
	}

	ih = malloc(sizeof *ih, M_DEVBUF, cold ? M_NOWAIT : M_WAITOK);
	if (ih == NULL)
		panic("softintr_establish: can't malloc handler info");
	ih->ih_fun = (void *)fun;
	ih->ih_arg = arg;
	ih->ih_level = level;
	ih->ih_irq = irq;
	s = splhigh();
	ih->ih_next = soft_intrhand[irq];
	soft_intrhand[irq] = ih;
	splx(s);

	return ih;
}

void
softintr_disestablish(arg)
	void *arg;
{
	struct intrhand *ih = arg;
	struct intrhand **p, *q;
	int irq = ih->ih_irq;

	for (p = &soft_intrhand[irq]; (q = *p) != NULL; p = &q->ih_next)
		if (q == ih) {
			*p = q->ih_next;
			free(ih, M_DEVBUF);
			return;
		}

	panic("softintr_disestablish: handler not registered");
}

void
softintr_schedule(arg)
	void *arg;
{
	struct intrhand *ih = arg;
	int irq = ih->ih_irq;

	switch (irq) {
	case SIR_CLOCK:  setsoftclock();  break;
	case SIR_BIO:    setsoftbio();    break;
	case SIR_NET:    setsoftnet();    break;
	case SIR_SERIAL: setsoftserial(); break;
	default:
		panic("softintr_schedule");
	}
}

/*
 * external interrupt handler
 */
void
ext_intr()
{
	volatile struct mpc860dev *immr = get_immr();
	struct intrhand *ih;
	int pcpl, vec, msr;

	pcpl = cpl;
	msr = mfmsr();

	vec = immr->siu_sivec >> 26;
	if (vec == 0xf) {
		printf("spurious interrupt\n");
		printf("pcpl   = 0x%x\n", pcpl);
		printf("simask = 0x%x\n", immr->siu_simask);
		printf("sipend = 0x%x\n", immr->siu_sipend);

		return;
		panic("ext_intr");
	}

redo:
	/*
	 * XXX Handle CPM interrupts.
	 */
	if (vec == CPMIRQ * 2 + 1) {
		int cpmvec;

		/* Acknowledge interrupt. */
		immr->cpm_civr |= 1;
		asm volatile ("eieio");

		cpmvec = immr->cpm_civr >> 11;
		if (cpmvec == 0)
			goto out;
	}

	/* XXX ack if edge-triggered */
#define idepth (curcpu()->ci_idepth)

	splraise(intrmask[vec]);
	mtmsr(msr|PSL_EE);
	ih = intrhand[vec];
	idepth++;
	while (ih) {
		if (ih->ih_level == IPL_VM) {
			KERNEL_LOCK(1, NULL);
			(*ih->ih_fun)(ih->ih_arg);
			KERNEL_UNLOCK_ONE(NULL);
		} else
			(*ih->ih_fun)(ih->ih_arg);
		ih = ih->ih_next;
	}
	idepth--;
	mtmsr(msr);
	splx(pcpl);

out:
	curcpu()->ci_data.cpu_nintr++;
	intrcnt[vec]++;

	vec = immr->siu_sivec >> 26;
	if (vec != 0xf)
		goto redo;
}

void
do_pending_int()
{
	volatile struct mpc860dev *immr = get_immr();
	struct intrhand *ih;
	int pcpl;
	int emsr, dmsr;
	static int processing;

	if (processing)
		return;

	processing = 1;
	emsr = mfmsr();
	dmsr = emsr & ~PSL_EE;
	mtmsr(dmsr);

	pcpl = cpl;

again:
	if ((ipending & ~pcpl) & (1 << SIR_SERIAL)) {
		ipending &= ~(1 << SIR_SERIAL);
		splsoftserial();
		mtmsr(emsr);
		ih = soft_intrhand[SIR_SERIAL];
		while (ih) {
			(*ih->ih_fun)(ih->ih_arg);
			ih = ih->ih_next;
		}
		mtmsr(dmsr);
		cpl = pcpl;
		immr->siu_simask = ~pcpl & HWMASK;
		goto again;
	}
	if ((ipending & ~pcpl) & (1 << SIR_BIO)) {
		ipending &= ~(1 << SIR_BIO);
		splsoftbio();
		mtmsr(emsr);
		ih = soft_intrhand[SIR_BIO];
		while (ih) {
			(*ih->ih_fun)(ih->ih_arg);
			ih = ih->ih_next;
		}
		mtmsr(dmsr);
		cpl = pcpl;
		immr->siu_simask = ~pcpl & HWMASK;
		goto again;
	}
	if ((ipending & ~pcpl) & (1 << SIR_NET)) {
		ipending &= ~(1 << SIR_NET);
		splsoftnet();
		mtmsr(emsr);
		ih = soft_intrhand[SIR_NET];
		while (ih) {
			(*ih->ih_fun)(ih->ih_arg);
			ih = ih->ih_next;
		}
		mtmsr(dmsr);
		cpl = pcpl;
		immr->siu_simask = ~pcpl & HWMASK;
		goto again;
	}
	if ((ipending & ~pcpl) & (1 << SIR_CLOCK)) {
		ipending &= ~(1 << SIR_CLOCK);
		splsoftclock();
		mtmsr(emsr);
		ih = soft_intrhand[SIR_CLOCK];
		while (ih) {
			(*ih->ih_fun)(ih->ih_arg);
			ih = ih->ih_next;
		}
		mtmsr(dmsr);
		cpl = pcpl;
		immr->siu_simask = ~pcpl & HWMASK;
		goto again;
	}

	if (ipending & ~pcpl) panic("do_pending_int(A)");
	if (cpl != pcpl) panic("do_pending_int(B)");
	//immr->siu_simask = ~pcpl & HWMASK;

	processing = 0;
	mtmsr(emsr);
}

int
splraise(ncpl)
	int ncpl;
{
	volatile struct mpc860dev *immr = get_immr();
	int ocpl, msr;

	msr = mfmsr();
	mtmsr(msr & ~PSL_EE);
	asm volatile ("sync");

	ocpl = cpl;
	ncpl |= ocpl;
	cpl = ncpl;
	immr->siu_simask = ~ncpl & HWMASK;

	asm volatile ("sync");
	mtmsr(msr);
	return ocpl;
}

void
splx(ncpl)
	int ncpl;
{
	volatile struct mpc860dev *immr = get_immr();
	int msr;

	msr = mfmsr();
	mtmsr(msr & ~PSL_EE);
	asm volatile ("sync");

	cpl = ncpl;
	immr->siu_simask = ~ncpl & HWMASK;

	asm volatile ("sync");
	mtmsr(msr);
	if (ipending & ~ncpl)
		do_pending_int();
}

int
spllower(ncpl)
	int ncpl;
{
	volatile struct mpc860dev *immr = get_immr();
	int ocpl, msr;

	msr = mfmsr();
	mtmsr(msr & ~PSL_EE);
	asm volatile ("sync");

	ocpl = cpl;
	cpl = ncpl;
	immr->siu_simask = ~ncpl & HWMASK;

	asm volatile ("sync");
	mtmsr(msr);
	if (ipending & ~ncpl)
		do_pending_int();
	return ocpl;
}

void
softintr(ipl)
	int ipl;
{
	int msr;

	msr = mfmsr();
	mtmsr(msr & ~PSL_EE);
	asm volatile ("sync");
	ipending |= 1 << ipl;
	mtmsr(msr);
}
