/*	$Id: pmap.c,v 1.101 2015/06/26 13:38:52 tsubai Exp $	*/

/*
 * Copyright (C) 1995, 1996 Wolfgang Solfrank.
 * Copyright (C) 1995, 1996 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/pool.h>
#include <sys/systm.h>

#include <uvm/uvm.h>

#include <machine/cpufunc.h>
#include <machine/pcb.h>
#include <machine/spr.h>

#ifndef NKPTP
#define NKPTP	4
#endif

#ifndef PMAP_PV_LOWAT
#define PMAP_PV_LOWAT 16
#endif

#define L1INDEX(va) ((va) >> 22)
#define L2INDEX(va) (((va) >> 12) & 0x3ff)

struct pv_entry {
	struct pv_entry *pv_next;	/* Linked list of mappings */
	u_int pv_exec;			/* Page has EXEC mapping */
	vaddr_t pv_va;			/* Virtual address of mapping */
	struct pmap *pv_pmap;
};

int physmem;
int nkptp = NKPTP;
struct pmap kernel_pmap_store;

static int pmap_initialized;
static int asidmap;
static krwlock_t pmap_main_lock;

int pmap_pv_lowat = PMAP_PV_LOWAT;
static struct pool pv_pool;
static struct pool pmap_pool;

#define pmap_alloc_pv()  pool_get(&pv_pool, PR_NOWAIT)
#define pmap_free_pv(pv) pool_put(&pv_pool, pv)
#define syncicache(va, len) __syncicache((void *)(va), len)

static void tlbie(vaddr_t);
static void tlbkill(vaddr_t);
static void tlbsync(void);
static void tlbia(void);
static struct pv_entry *pa_to_pv(paddr_t);
static char *pa_to_attr(paddr_t);
static int pmap_enter_pv(struct pmap *, vaddr_t, paddr_t, vm_prot_t);
static void pmap_remove_pv(struct pmap *, vaddr_t, paddr_t, u_int);
static int pmap_get_ptp(struct pmap *, vaddr_t);
int pmap_alloc_asid(void);
void pmap_free_asid(int);
void savetlb(void), dumptlb(void);

/* XXX tlbie does not work in some case. */
#define tlbieXXX(ea) tlbkill(ea)

static inline void
tlbie(ea)
	vaddr_t ea;
{
	asm volatile ("tlbie %0" :: "r"(ea) : "memory");
}

static inline void
tlbsync()
{
	/* tlbsync is no-operation on mpc860. */
	/* asm volatile ("sync; tlbsync; sync" ::: "memory"); */
}

static void
tlbia()
{
	asm volatile ("sync; tlbia" ::: "memory");
	tlbsync();
}

static inline struct pv_entry *
pa_to_pv(pa)
	paddr_t pa;
{
	int bank, pg;

	bank = vm_physseg_find(atop(pa), &pg);
	if (bank == -1)
		return NULL;
	return &vm_physmem[bank].pmseg.pvent[pg];
}

static inline char *
pa_to_attr(pa)
	paddr_t pa;
{
	int bank, pg;

	bank = vm_physseg_find(atop(pa), &pg);
	if (bank == -1)
		return NULL;
	return &vm_physmem[bank].pmseg.attrs[pg];
}

void
pmap_bootstrap()
{
	u_int *l1pt, *pte;
	paddr_t pa;
	int i;

	curpcb->pcb_pm = pmap_kernel();

	uvm_page_physget(&pa);
	l1pt = (void *)pa;
	bzero(l1pt, NBPG);
	mtspr(M_TWB, l1pt);

	pmap_kernel()->pm_pdir = l1pt;
	pmap_kernel()->pm_pdirpa = (paddr_t)l1pt;

	pte = &l1pt[L1INDEX(KERNBASE)];
	for (i = 0; i < nkptp; i++) {
		uvm_page_physget(&pa);
		bzero((void *)pa, NBPG);
		*pte++ = pa | PG_V;
	}

	rw_init(&pmap_main_lock);

	pool_init(&pmap_pool, sizeof(struct pmap), 0, 0, 0, "pmpool",
	    &pool_allocator_nointr, IPL_NONE);
	pool_init(&pv_pool, sizeof(struct pv_entry), 0, 0, 0, "pvpool",
	    &pool_allocator_nointr, IPL_NONE);

	tlbia();
}

/*
 * Restrict given range to physical memory
 */
void
pmap_real_memory(start, size)
	paddr_t *start;
	psize_t *size;
{
	/* XXX */
}

void
pmap_init()
{
	int npages, bank;
	vaddr_t addr;
	vsize_t sz;
	struct pv_entry *pv;
	char *attr;

	/*
	 * Compute the number of pages we have and then allocate RAM
	 * for each pages' pv_entry and saved attributes.
	 */
	npages = 0;
	for (bank = 0; bank < vm_nphysseg; bank++)
		npages += (vm_physmem[bank].end - vm_physmem[bank].start);
	sz = sizeof(struct pv_entry) * npages + npages;
	sz = round_page(sz);
	addr = uvm_km_alloc(kernel_map, sz, 0, UVM_KMF_WIRED|UVM_KMF_ZERO);
	if (addr == 0)
		panic("pmap_init: unable to allocate pv_table");

	pv   = (void *)addr;
	attr = (char *)(pv + npages);
	for (bank = 0; bank < vm_nphysseg; bank++) {
		sz = vm_physmem[bank].end - vm_physmem[bank].start;
		vm_physmem[bank].pmseg.pvent = pv;
		vm_physmem[bank].pmseg.attrs = attr;
		pv += sz;
		attr += sz;
	}

	pool_setlowat(&pv_pool, pmap_pv_lowat);

	/* Reserve ASID 0. */
	pmap_alloc_asid();

	pmap_initialized = 1;
}

void
pmap_virtual_space(startp, endp)
	vaddr_t *startp;
	vaddr_t *endp;
{
	*startp = VM_MIN_KERNEL_ADDRESS;
	*endp   = VM_MAX_KERNEL_ADDRESS;
}

struct pmap *
pmap_create()
{
	struct pmap *pmap;
	struct vm_page *pg;

	pmap = pool_get(&pmap_pool, PR_WAITOK);
	bzero(pmap, sizeof *pmap);
	pg = uvm_pagealloc(NULL, 0, NULL, UVM_PGA_USERESERVE|UVM_PGA_ZERO);
	if (pg == NULL)
		panic("pmap_create");
	pmap->pm_pdirpa = VM_PAGE_TO_PHYS(pg);
	pmap->pm_pdir = (void *)pmap->pm_pdirpa;
	pmap->pm_refs = 1;
	pmap->pm_asid = pmap_alloc_asid();
	pmap->pm_stats.wired_count = 0;
	pmap->pm_stats.resident_count = 1;	/* Count PTP as resident. */

	return pmap;
}

void
pmap_reference(pmap)
	struct pmap *pmap;
{
	pmap->pm_refs++;
}

void
pmap_destroy(pm)
	struct pmap *pm;
{
	struct vm_page *pg;
	u_int *l1pt;
	int i;

	if (--pm->pm_refs == 0) {
		l1pt = pm->pm_pdir;
		for (i = 0; i < 1024; i++) {
			if (l1pt[i]) {
				pg = PHYS_TO_VM_PAGE(l1pt[i]);
				if (pg == NULL)
					panic("pmap_destroy");
				uvm_pagefree(pg);
			}
		}
		pg = PHYS_TO_VM_PAGE(pm->pm_pdirpa);
		if (pg == NULL)
			panic("pmap_destroy");
		uvm_pagefree(pg);
		pmap_free_asid(pm->pm_asid);
		pool_put(&pmap_pool, pm);
	}
}

void
pmap_zero_page(pa)
	paddr_t pa;
{
#if 0
	bzero(pa, NBPG);
#else
	int i;

	for (i = NBPG/CACHELINESIZE; i > 0; i--) {
		asm volatile ("dcbz 0,%0" :: "r"(pa));
		pa += CACHELINESIZE;
	}
#endif
}

void
pmap_copy_page(src, dst)
	paddr_t src, dst;
{
#if 0
	bcopy((caddr_t)src, (caddr_t)dst, NBPG);
#else
	int *s = (int *)src;
	int *d = (int *)dst;
	int i, j;

	for (i = NBPG/CACHELINESIZE; i > 0; i--) {
		asm volatile ("dcbz 0,%0" :: "r"(d) : "memory");
		for (j = CACHELINESIZE/sizeof *s; j > 0; j--)
			*d++ = *s++;
	}
#endif
}

static int
pmap_get_ptp(pmap, va)
	struct pmap *pmap;
	vaddr_t va;
{
	struct vm_page *pg;

	if (pmap->pm_pdir[L1INDEX(va)])
		return 0;

	/* Allocate a new PTP. */
	pg = uvm_pagealloc(NULL, 0, NULL, UVM_PGA_USERESERVE|UVM_PGA_ZERO);
	if (pg == NULL)
		return ENOMEM;

	pmap->pm_pdir[L1INDEX(va)] = VM_PAGE_TO_PHYS(pg) | PG_V;
	pmap->pm_stats.resident_count++;	/* Count PTP as resident. */
	return 0;
}

/*
 * This returns whether this is the first (EXEC) mapping of a page.
 */
static inline int
pmap_enter_pv(pm, va, pa, prot)
	struct pmap *pm;
	vaddr_t va;
	paddr_t pa;
	vm_prot_t prot;
{
	struct pv_entry *pv, *npv;
	int s, exec;

	if (!pmap_initialized)
		return 0;

	s = splvm();

	pv = pa_to_pv(pa);
	if (pv->pv_pmap == NULL) {
		/*
		 * No entries yet, use header as the first entry.
		 */
		exec = 0;	/* No previous exec mapping. */
		pv->pv_va = va;
		pv->pv_pmap = pm;
		pv->pv_exec = (prot & VM_PROT_EXECUTE) ? 1 : 0;
		pv->pv_next = NULL;
	} else {
		/*
		 * There is at least one other VA mapping this page.
		 * Place this entry after the header.
		 */
		/* Is there at least one exec mapping of this page? */
		for (npv = pv; npv; npv = npv->pv_next) {
			exec = npv->pv_exec;
			if (exec)
				break;
		}

		npv = pmap_alloc_pv();
		if (npv == NULL)
			return ENOMEM;
		npv->pv_va = va;
		npv->pv_pmap = pm;
		npv->pv_exec = (prot & VM_PROT_EXECUTE) ? 1 : 0;
		npv->pv_next = pv->pv_next;
		pv->pv_next = npv;
	}
	splx(s);
	return exec;
}

void
pmap_remove_pv(pm, va, pa, pte)
	struct pmap *pm;
	vaddr_t va;
	paddr_t pa;
	u_int pte;
{
	struct pv_entry *pv, *npv;
	char *attr;

	/*
	 * First transfer reference/change bits to cache.
	 */
	attr = pa_to_attr(pa);
	if (attr == NULL)
		return;
	*attr |= (pte & (PG_REF|PG_CHG)) >> 4;

	/*
	 * Remove from the PV table.
	 */
	pv = pa_to_pv(pa);

	/*
	 * If it is the first entry on the list, it is actually
	 * in the header and we must copy the following entry up
	 * to the header.  Otherwise we must search the list for
	 * the entry.  In either case we free the now unused entry.
	 */
	if (pm == pv->pv_pmap && va == pv->pv_va) {
		npv = pv->pv_next;
		if (npv) {
			*pv = *npv;
			pmap_free_pv(npv);
		} else
			pv->pv_pmap = NULL;
	} else {
		for (; (npv = pv->pv_next) != NULL; pv = npv)
			if (pm == npv->pv_pmap && va == npv->pv_va)
				break;
		if (npv) {
			pv->pv_next = npv->pv_next;
			pmap_free_pv(npv);
		}
#ifdef DIAGNOSTIC
		else
			panic("pmap_remove_pv: not on list");
#endif
	}
}

int
pmap_enter(pmap, va, pa, prot, flags)
	struct pmap *pmap;
	vaddr_t va;
	paddr_t pa;
	vm_prot_t prot;
	u_int flags;
{
	int bank, off, error;
	int wireddelta = 0, resdelta = 0;
	int wired = flags & PMAP_WIRED;
	u_int wimg = pa & PG_CI;
	u_int *l1pt, *l2pt, opte, npte;

	KASSERT(pmap_initialized);

#ifdef DIAGNOSTIC
	if (va >= VM_MAX_KERNEL_ADDRESS)
		panic("pmap_enter: out of range");
#endif

	rw_enter(&pmap_main_lock, RW_READER);

	pa &= PG_FRAME;
	if (pmap_get_ptp(pmap, va)) {
		if (flags & PMAP_CANFAIL) {
			error = ENOMEM;
			goto out;
		}
		panic("pmap_enter: get ptp failed");
	}
	l1pt = pmap->pm_pdir;
	l2pt = (void *)(l1pt[L1INDEX(va)] & PG_FRAME);
	opte = l2pt[L2INDEX(va)];

	if (opte & PG_V) {
		if (wired && (opte & PG_W) == 0)
			wireddelta = 1;
		else if (!wired && (opte & PG_W))
			wireddelta = -1;

		if ((opte & PG_FRAME) == pa) {
			/* Sync R/M bit. */
			bank = vm_physseg_find(atop(pa), &off);
			if (bank != -1)
				vm_physmem[bank].pmseg.attrs[off] |= opte >> 4;
			goto enter_now;
		}
		pmap_remove_pv(pmap, va, opte & PG_FRAME, opte);
	} else {
		resdelta = 1;
		if (wired)
			wireddelta = 1;
	}

	if (uvm_pageismanaged(pa)) {
		int rv;
		rv = pmap_enter_pv(pmap, va, pa, prot);
		if (rv == ENOMEM) {
			if (flags & PMAP_CANFAIL) {
				error = ENOMEM;
				goto out;
			}
			panic("pmap_enter: out of memory");
		}
		if (rv == 0) {
			if (prot & VM_PROT_EXECUTE)
				syncicache(pa, NBPG);
		}
	}

enter_now:
	npte = pa | wimg | PG_V;
	if (prot & VM_PROT_WRITE)
		npte |= PG_RW;
	if (wired)
		npte |= PG_W;
	if (pmap == pmap_kernel())
		npte |= PG_SH;

	l2pt[L2INDEX(va)] = npte;
	asm volatile ("sync");
	tlbieXXX(va);

	pmap->pm_stats.resident_count += resdelta;
	pmap->pm_stats.wired_count += wireddelta;
	error = 0;

out:
	rw_exit(&pmap_main_lock);

	return error;
}

void
pmap_remove(pmap, sva, eva)
	struct pmap *pmap;
	vaddr_t sva, eva;
{
	u_int *l1pt, *l2pt, *pte, opte;

	rw_enter(&pmap_main_lock, RW_READER);

	l1pt = pmap->pm_pdir;
	for (; sva < eva; sva += NBPG) {
		l2pt = (void *)(l1pt[L1INDEX(sva)] & PG_FRAME);
		if (l2pt == NULL)
			continue;
		pte = &l2pt[L2INDEX(sva)];

		if ((*pte & PG_V) == 0)
			continue;

		opte = *pte;
		*pte = 0;
		asm volatile ("sync");
		tlbieXXX(sva);

		pmap_remove_pv(pmap, sva, opte & PG_FRAME, opte);

		if (opte & PG_W)
			pmap->pm_stats.wired_count--;
		pmap->pm_stats.resident_count--;
	}
	rw_exit(&pmap_main_lock);
}

void
pmap_kenter_pa(va, pa, prot, flags)
	vaddr_t va;
	paddr_t pa;
	vm_prot_t prot;
	u_int flags;
{
	u_int *l1pt, *l2pt, pte;

	pmap_get_ptp(pmap_kernel(), va);
	l1pt = pmap_kernel()->pm_pdir;
	l2pt = (void *)(l1pt[L1INDEX(va)] & PG_FRAME);
	if (l2pt == NULL)
		panic("pmap_kenter_pa");

	pte = pa | PG_SH | PG_V;
	if (prot & VM_PROT_WRITE)
		pte |= PG_RW | PG_CHG;
	if (prot & PMAP_CI)
		pte |= PG_CI;
	l2pt[L2INDEX(va)] = pte;
	asm volatile ("sync");
	tlbie(va);
}

void
pmap_kremove(va, len)
	vaddr_t va;
	vsize_t len;
{
	u_int *l1pt, *l2pt;

	len >>= PAGE_SHIFT;
	l1pt = pmap_kernel()->pm_pdir;
	while (len > 0) {
		l2pt = (void *)(l1pt[L1INDEX(va)] & PG_FRAME);
		if (l2pt) {
			l2pt[L2INDEX(va)] = 0;
			asm volatile ("sync");
			tlbie(va);
		}
		va += NBPG;
		len--;
	}
}

bool
pmap_extract(pmap, va, pap)
	struct pmap *pmap;
	vaddr_t va;
	paddr_t *pap;
{
	u_int *l1pt, *l2pt, pte;

	l1pt = pmap->pm_pdir;
	l2pt = (void *)(l1pt[L1INDEX(va)] & PG_FRAME);
	if (l2pt == NULL)
		return 0;
	pte = l2pt[L2INDEX(va)];
	if ((pte & PG_V) == 0)
		return 0;
	if (pap)
		*pap = (pte & PG_FRAME) | (va & ~PG_FRAME);
	return 1;
}

void
pmap_page_protect(pg, prot)
	struct vm_page *pg;
	vm_prot_t prot;
{
	int bank, off;
	struct pv_entry *pv;
	char *myattrs;
	u_int *l1pt, *l2pt, npte;

	if (prot & VM_PROT_WRITE)
		return;

	if (prot == VM_PROT_NONE) {
		pmap_page_remove(pg);
		return;
	}

	bank = vm_physseg_find(atop(VM_PAGE_TO_PHYS(pg)), &off);
	if (bank == -1) {
		printf("pmap_page_protect: unmanaged page?\n");
		return;
	}

	pv = &vm_physmem[bank].pmseg.pvent[off];
	if (pv->pv_pmap == NULL)
		return;

	rw_enter(&pmap_main_lock, RW_WRITER);
	myattrs = &vm_physmem[bank].pmseg.attrs[off];

	for (; pv; pv = pv->pv_next) {
		l1pt = pv->pv_pmap->pm_pdir;
		l2pt = (void *)(l1pt[L1INDEX(pv->pv_va)] & PG_FRAME);
		if (l2pt == NULL)
			panic("pmap_page_protect: no L2 PTE?");

		npte = l2pt[L2INDEX(pv->pv_va)];
		*myattrs |= (npte & PG_CHG) >> 4;
		npte &= ~(PG_RW|PG_CHG);
		if (l2pt[L2INDEX(pv->pv_va)] != npte) {
			l2pt[L2INDEX(pv->pv_va)] = npte;
			asm volatile ("sync");
			tlbieXXX(pv->pv_va);
		}
	}

	rw_exit(&pmap_main_lock);
}

void
pmap_page_remove(pg)
	struct vm_page *pg;
{
	paddr_t pa = VM_PAGE_TO_PHYS(pg);
	int bank, off;
	struct pv_entry *pv;
	vaddr_t va;
	struct pmap *pm;
	u_int *l1pt, *l2pt, opte;

	bank = vm_physseg_find(atop(pa), &off);
	if (bank == -1) {
		printf("pmap_page_remove: unmanaged page?\n");
		return;
	}

	pv = &vm_physmem[bank].pmseg.pvent[off];
	if (pv->pv_pmap == NULL)
		return;

	rw_enter(&pmap_main_lock, RW_WRITER);

	while (pv->pv_pmap) {
		va = pv->pv_va;
		pm = pv->pv_pmap;
		l1pt = pm->pm_pdir;
		if (l1pt == NULL)
			panic("pmap_page_remove1");
		l2pt = (void *)(l1pt[L1INDEX(va)] & PG_FRAME);
		if (l2pt == NULL)
			panic("pmap_page_remove2");
		opte = l2pt[L2INDEX(va)];
		if ((opte & PG_V) == 0)
			panic("pmap_page_remove: !PG_V");

		l2pt[L2INDEX(va)] = 0;
		asm volatile ("sync");
		tlbieXXX(va);
		tlbsync();
		pmap_remove_pv(pm, va, pa, opte);
		if (opte & PG_W)
			pm->pm_stats.wired_count--;
		pm->pm_stats.resident_count--;
	}

	rw_exit(&pmap_main_lock);
}

int
pmap_change_attrs(pg, setbits, clearbits)
	struct vm_page *pg;
	int setbits, clearbits;
{
	u_int result;
	int bank, off;
	struct pv_entry *pv;
	char *myattrs;
	u_int *l1pt, *l2pt, npte;

	setbits >>= 4;
	clearbits >>= 4;

	bank = vm_physseg_find(atop(VM_PAGE_TO_PHYS(pg)), &off);
	if (bank == -1) {
		printf("pmap_change_attrs: unmanaged page?\n");
		return 0;
	}

	rw_enter(&pmap_main_lock, RW_WRITER);
	pv = &vm_physmem[bank].pmseg.pvent[off];

	myattrs = &vm_physmem[bank].pmseg.attrs[off];
	result = *myattrs & clearbits;
	*myattrs = (*myattrs | setbits) & ~clearbits;

	result <<= 4;
	setbits <<= 4;
	clearbits <<= 4;

	if (pv->pv_pmap == NULL)
		goto out;
	for (; pv; pv = pv->pv_next) {
		l1pt = pv->pv_pmap->pm_pdir;
		l2pt = (void *)(l1pt[L1INDEX(pv->pv_va)] & PG_FRAME);
		if (l2pt == NULL)
			panic("pmap_change_attrs: mapping w/o PTP detected");
		npte = l2pt[L2INDEX(pv->pv_va)];
		result |= npte & clearbits;
		npte = (npte | setbits) & ~clearbits;
		if (l2pt[L2INDEX(pv->pv_va)] != npte) {
			l2pt[L2INDEX(pv->pv_va)] = npte;
			asm volatile ("sync");
			tlbieXXX(pv->pv_va);
		}
	}

out:
	rw_exit(&pmap_main_lock);

	return result;
}

int
pmap_test_attrs(pg, testbits)
	struct vm_page *pg;
	int testbits;
{
	int bank, off;
	char *myattrs;
	struct pv_entry *pv;
	u_int *l1pt, *l2pt, pte;

	bank = vm_physseg_find(atop(VM_PAGE_TO_PHYS(pg)), &off);
	if (bank == -1) {
		printf("pmap_test_attrs: unmanaged page?\n");
		return 0;
	}

	testbits >>= 4;

	myattrs = &vm_physmem[bank].pmseg.attrs[off];
	if (*myattrs & testbits)
		return 1;

	pv = &vm_physmem[bank].pmseg.pvent[off];
	if (pv->pv_pmap == NULL)
		return 0;

	rw_enter(&pmap_main_lock, RW_WRITER);

	for (; pv; pv = pv->pv_next) {
		l1pt = pv->pv_pmap->pm_pdir;
		l2pt = (void *)(l1pt[L1INDEX(pv->pv_va)] & PG_FRAME);
		if (l2pt == NULL)
			panic("pmap_test_attrs");
		pte = l2pt[L2INDEX(pv->pv_va)];
		pte >>= 4;
		*myattrs |= pte;
		if (*myattrs & testbits)
			break;
	}

	rw_exit(&pmap_main_lock);
	return *myattrs & testbits;
}

void
pmap_write_protect(pmap, sva, eva, prot)
	struct pmap *pmap;
	vaddr_t sva, eva;
	vm_prot_t prot;
{
	u_int *l1pt, *l2pt, npte;
	u_int md_prot;

	if (prot & VM_PROT_WRITE)
		md_prot = PG_RW;
	else
		md_prot = 0;
	for (; sva < eva; sva += NBPG) {
		l1pt = pmap->pm_pdir;
		l2pt = (void *)(l1pt[L1INDEX(sva)] & PG_FRAME);
		if (l2pt == NULL)
			continue;

		npte = (l2pt[L2INDEX(sva)] & ~PG_PROT) | md_prot;
		if (npte != l2pt[L2INDEX(sva)]) {
			l2pt[L2INDEX(sva)] = npte;
			asm volatile ("sync");
			tlbieXXX(sva);
		}
	}
}

void
pmap_unwire(pmap, va)
	struct pmap *pmap;
	vaddr_t va;
{
	u_int *l1pt, *l2pt;

	l1pt = pmap->pm_pdir;
	l2pt = (void *)(l1pt[L1INDEX(va)] & PG_FRAME);

	if (l2pt == NULL)
		panic("pmap_unwire: invalid L1 pte");

	if (l2pt[L2INDEX(va)] & PG_W) {
		l2pt[L2INDEX(va)] &= ~PG_W;
		pmap->pm_stats.wired_count--;
	} else {
		printf("pmap_unwire: pmap %p va 0x%lx was not wired\n",
		    pmap, va);
	}
}

void
pmap_activate(l)
	struct lwp *l;
{
	struct pcb *pcb = l->l_addr;
	struct pmap *pmap = l->l_proc->p_vmspace->vm_map.pmap;
	int asid;

	pcb->pcb_pm = pmap;
	if (l == curlwp) {
		asm volatile ("sync");
		mtspr(M_TWB, pmap->pm_pdirpa);
		asid = pmap->pm_asid;
		if (asid > 0) {
			/* Valid (unique) ASID. */
			mtspr(M_CASID, asid);
		} else {
			/* Shared ASID... (XXX dynamic allocation?) */
			mtspr(M_CASID, 0);
			if (pmap != pmap_kernel())
				tlbia();
		}
	}
}

void
pmap_deactivate(l)
	struct lwp *l;
{
}

paddr_t
vtophys(va)
	vaddr_t va;
{
	paddr_t pa;
	u_int pte;

	pte = pmap_kernel()->pm_pdir[L1INDEX(va)];
	if ((pte & TWC_PS) == TWC_8M)
		return (pte & 0xff800000) | (va & 0x007fffff);

	if (pmap_extract(pmap_kernel(), va, &pa))
		return pa;
	return 0;
}

void
pmap_collect(pmap)
	struct pmap *pmap;
{
}

/*
 * Synchronize caches in proc p.
 */
void
pmap_procwr(p, va, len)
	struct proc *p;
	vaddr_t va;
	size_t len;
{
	paddr_t pa;

	if (pmap_extract(p->p_vmspace->vm_map.pmap, va, &pa) == 0)
		return;
	syncicache(pa, len);
}

int
pmap_emul_modify(pmap, va)
	struct pmap *pmap;
	vaddr_t va;
{
	u_int *l1pt, *l2pt, pte;

	l1pt = pmap->pm_pdir;
	l2pt = (void *)(l1pt[L1INDEX(va)] & PG_FRAME);
	if (l2pt == NULL) {
		printf("pmap_emul_modify: invalid pdir va = 0x%lx\n", va);
		return 0;
	}

	pte = l2pt[L2INDEX(va)];
	if ((pte & (PG_RW|PG_V)) != (PG_RW|PG_V))
		return 0;

	if (pte & PG_CHG)
		panic("pmap_emul_modify: already PG_CHG, va = 0x%lx", va);
	pte |= PG_CHG;
	l2pt[L2INDEX(va)] = pte;
	asm volatile ("sync" ::: "memory");
	tlbieXXX(va);
	return 1;
}

/*
 * XXX tlbie workaround.
 * XXX tlbie doesn't work correctly if CTR_PPCS is used.
 */
void
tlbkill(va)
	vaddr_t va;
{
	u_int msr, ictr, dctr;

	msr = mfmsr();
	mtmsr(msr & ~(PSL_EE|PSL_IR|PSL_DR));

	dctr = mfspr(MD_CTR);
	ictr = mfspr(MI_CTR);
	mtspr(MD_CTR, CTR_TWAM);
	mtspr(MI_CTR, 0);

	va &= PG_FRAME;
	tlbie(va);
	tlbie(va + 1024);

	mtspr(MD_CTR, dctr);
	mtspr(MI_CTR, ictr);
	mtmsr(msr);
}

/*
 * Allocate an ASID.  MPC860 has (only) 16 ASIDs.
 */
int
pmap_alloc_asid()
{
	int i;

	if (asidmap == 0xffff)
		return -1;

	i = ffs(~asidmap) - 1;
	asidmap |= 1 << i;
	return i;
}

void
pmap_free_asid(i)
	int i;
{
	if (i > 0)
		asidmap &= ~(1 << i);
}

void
kdtlbload(va)
	vaddr_t va;
{
	u_int *l1pt, *l2pt;
	u_int rpn;

	/* printf("kdtlbload va = 0x%lx\n", va); */
	l1pt = (void *)pmap_kernel()->pm_pdirpa;
	if ((l1pt[L1INDEX(va)] & TWC_PS) == TWC_8M) {
		/* Use 8MB TLB. */
		mtspr(MD_TWC, TWC_8M | TWC_V);
		rpn = l1pt[L1INDEX(va)] & (PG_FRAME | PG_CI);
		rpn |= RPN_8M | RPN_C | RPN_SH | RPN_V;
		rpn |= RPN_KHIT;	/* Hit only supervisor, kernel R/W. */
		mtspr(MD_RPN, rpn);
		return;
	}
	l2pt = (void *)(l1pt[L1INDEX(va)] & PG_FRAME);
	if (l2pt == NULL) {
		mtspr(MD_TWC, TWC_V);
		mtspr(MD_RPN, RPN_KHIT);
		return;
	}
	rpn = l2pt[L2INDEX(va)];
	rpn &= ~0xef0;
	rpn |= RPN_KHIT;	/* Hit only supervisor, kernel R/W. */
	mtspr(MD_TWC, TWC_V);
	mtspr(MD_RPN, rpn);
}

void
kitlbload(va)
	vaddr_t va;
{
	u_int *l1pt, *l2pt;
	u_int rpn;

	/* printf("kitlbload va = 0x%lx\n", va); */

	l1pt = (void *)pmap_kernel()->pm_pdirpa;
	if ((l1pt[L1INDEX(va)] & TWC_PS) == TWC_8M) {
		/* Use 8MB TLB. */
		mtspr(MI_TWC, TWC_8M | TWC_V);
		rpn = l1pt[L1INDEX(va)] & PG_FRAME;
		rpn |= RPN_8M | RPN_SH | RPN_V;
		rpn |= RPN_KHIT;	/* Hit only supervisor, kernel R/W. */
		mtspr(MI_RPN, rpn);
		return;
	}
	l2pt = (void *)(l1pt[L1INDEX(va)] & PG_FRAME);
	if (l2pt == NULL) {
		mtspr(MI_TWC, TWC_V);
		mtspr(MI_RPN, RPN_KHIT);
		return;
	}
	rpn = l2pt[L2INDEX(va)];
	rpn &= ~0xef0;
	rpn |= RPN_KHIT;	/* Hit only supervisor, kernel R/W. */
	mtspr(MI_TWC, TWC_V);
	mtspr(MI_RPN, rpn);
}

#if 0
struct dtlbdump {
	u_int epn:20;
	u_int spv:4;
	u_int ps:3;
	u_int sh:1;
	u_int asid:4;

	u_int rpn:20;
	u_int ps2:3;
	u_int apg:4;
	u_int g:1;
	u_int wt:1;
	u_int ci:1;
	u_int pad1:2;

	u_int pad2:17;
	u_int c:1;
	u_int v:1;
	u_int sa:4;	/* Supervisor access */
	u_int sat:1;	/* Supervisor access type (0:ro / 1:rw) */
	u_int urp0:1;
	u_int uwp0:1;
	u_int urp1:1;
	u_int uwp1:1;
	u_int urp2:1;
	u_int uwp2:1;
	u_int urp3:1;
	u_int uwp3:1;
};

struct itlbdump {
	u_int epn:20;
	u_int ps:3;
	u_int asid:4;
	u_int sh:1;
	u_int spv:4;

	u_int rpn:20;
	u_int ps2:3;
	u_int ci:1;
	u_int apg:4;
	u_int sfp:4;	/* Supervisor fetch permission */

	u_int pad:26;
	u_int ufp:4;	/* User fetch permission */
	u_int v:1;
	u_int g:1;
};

static struct dtlbdump dtlb[32];
static struct itlbdump itlb[32];

void
savetlb()
{
	u_int ictr, dctr;
	u_int tlb, msr, *p;

	asm volatile ("sync");
	msr = mfmsr();
	mtmsr(msr & ~(PSL_EE|PSL_IR|PSL_DR));
	asm volatile ("isync");

	dctr = mfspr(MD_CTR);
	ictr = mfspr(MI_CTR);

	p = (void *)dtlb;
	for (tlb = 0; tlb < 32; tlb++) {
		mtspr(MD_CTR, CTR_TLB_INDX(tlb));
		mtspr(MD_CAM, 0);
		*p++ = mfspr(MD_CAM);
		*p++ = mfspr(MD_RAM0);
		*p++ = mfspr(MD_RAM1);
	}
	p = (void *)itlb;
	for (tlb = 0; tlb < 32; tlb++) {
		mtspr(MI_CTR, CTR_TLB_INDX(tlb));
		mtspr(MI_CAM, 0);
		*p++ = mfspr(MI_CAM);
		*p++ = mfspr(MI_RAM0);
		*p++ = mfspr(MI_RAM1);
	}

	mtspr(MD_CTR, dctr);
	mtspr(MI_CTR, ictr);
	mtmsr(msr);
}

void
dumptlb()
{
	struct dtlbdump *dp = dtlb;
	struct itlbdump *ip = itlb;
	u_int ur, uw;
	int tlb;

	printf("DTLB\n");
	for (tlb = 0; tlb < 32; tlb++) {
		printf("%2d", tlb);
		printf(" EPN=0x%08x", dp->epn << 12);
		printf(" RPN=0x%08x", dp->rpn << 12);
		if (dp->ps == 0x0) printf(" 4K");
		if (dp->ps == 0x1) printf(" 16K");
		if (dp->ps == 0x3) printf(" 512K");
		if (dp->ps == 0x7) printf(" 8M");
		if (dp->v)  printf(" V");
		if (dp->c)  printf(" C");
		if (dp->sh) printf(" SH");
		if (dp->g)  printf(" G");
		if (dp->wt) printf(" WT");
		if (dp->ci) printf(" CI");
		printf(" ASID=%d", dp->asid);
		printf(" SPV=0x%x", dp->spv);
		printf(" SA=0x%x", dp->sa);
		printf(dp->sat ? "(rw)" : "(ro)");
		ur = dp->urp0 << 3 | dp->urp1 << 2 | dp->urp2 << 1 | dp->urp3;
		uw = dp->uwp0 << 3 | dp->uwp1 << 2 | dp->uwp2 << 1 | dp->uwp3;
		if (ur) printf(" UR=0x%x", ur);
		if (uw) printf(" UW=0x%x", uw);
		printf("\n");
		sswt();
		dp++;
	}

	printf("ITLB\n");
	for (tlb = 0; tlb < 32; tlb++) {
		printf("%2d", tlb);
		printf(" EPN=0x%08x", ip->epn << 12);
		printf(" RPN=0x%08x", ip->rpn << 12);
		if (ip->ps == 0x0) printf(" 4K");
		if (ip->ps == 0x1) printf(" 16K");
		if (ip->ps == 0x3) printf(" 512K");
		if (ip->ps == 0x7) printf(" 8M");
		printf(" ASID=%d", ip->asid);
		if (ip->v)  printf(" V");
		if (ip->sh) printf(" SH");
		if (ip->ci) printf(" CI");
		if (ip->g)  printf(" G");
		printf(" SPV=0x%x", ip->spv);
		printf(" SFP=0x%x", ip->sfp);
		printf(" UFP=0x%x", ip->ufp);
		printf("\n");
		sswt();
		ip++;
	}
}
#endif
