/*	$Id: trap.c,v 1.42 2012/06/20 01:40:51 tsubai Exp $	*/

/*
 * Copyright (C) 1995, 1996 Wolfgang Solfrank.
 * Copyright (C) 1995, 1996 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "opt_ddb.h"

#include <sys/param.h>
#include <sys/kauth.h>
#include <sys/kmem.h>
#include <sys/pool.h>
#include <sys/proc.h>
#include <sys/systm.h>

#include <uvm/uvm_extern.h>

#include <machine/cpu.h>
#include <machine/db_machdep.h>
#include <machine/frame.h>
#include <machine/pcb.h>
#include <machine/pmap.h>
#include <machine/psl.h>
#include <machine/trap.h>

#include <powerpc/fpu/fpu_extern.h>

void
trap(frame)
	struct trapframe *frame;
{
	struct lwp *l = curlwp;
	struct proc *p = l ? l->l_proc : NULL;
	struct pcb *pcb = curpcb;
	vaddr_t va = frame->dar;
	int dsisr = frame->dsisr;
	int type = frame->exc;
	int ftype, rv;
	ksiginfo_t ksi;
	faultbuf *fb;

	curcpu()->ci_data.cpu_ntrap++;

	if (frame->srr1 & PSL_PR) {
		type |= EXC_USER;
		LWP_CACHE_CREDS(l, p);
	}

	switch (type) {
	case EXC_TRC|EXC_USER:
		frame->srr1 &= ~PSL_SE;
		KSI_INIT_TRAP(&ksi);
		ksi.ksi_signo = SIGTRAP;
		ksi.ksi_addr = (void *)frame->srr0;
		ksi.ksi_trap = EXC_TRC;
		ksi.ksi_code = TRAP_TRACE;
		trapsignal(l, &ksi);
		break;

	case EXC_DTLBERR:
		/* XXX CHECK intr_depth. */

		if (dsisr & DSISR_STORE)
			ftype = VM_PROT_WRITE;
		else
			ftype = VM_PROT_READ;

		/* XXX */
		if (dsisr & 0x08000000)
			printf("DSISR(k) 0x%08x va 0x%lx\n", dsisr, va);
		/* PG_CHG emulation */
		if ((dsisr & DSISR_ALL) == (DSISR_PROT|DSISR_STORE) &&
		    pmap_emul_modify(pmap_kernel(), va))
			return;

		rv = uvm_fault(kernel_map, trunc_page(va), ftype);
		if (rv == 0)
			return;
		if (rv == EACCES)
			rv = EFAULT;
		if ((fb = pcb->pcb_onfault) != NULL) {
			frame->srr0 = (*fb)[0];
			frame->fixreg[1] = (*fb)[1];
			frame->fixreg[2] = (*fb)[2];
			frame->fixreg[3] = rv;
			frame->cr = (*fb)[3];
			bcopy(&(*fb)[4], &frame->fixreg[13],
			    19 * sizeof(register_t));
			return;
		}
		goto brain_damage;

	case EXC_DTLBERR|EXC_USER:
		/* XXX */
		if (dsisr & 0x08000000)
			printf("DSISR(u) 0x%08x va 0x%lx\n", dsisr, va);
		/* PG_CHG emulation */
		if ((dsisr & DSISR_ALL) == (DSISR_PROT|DSISR_STORE) &&
		    pmap_emul_modify(p->p_vmspace->vm_map.pmap, va))
			return;

		if (dsisr & DSISR_STORE)
			ftype = VM_PROT_WRITE;
		else
			ftype = VM_PROT_READ;
		rv = uvm_fault(&p->p_vmspace->vm_map, trunc_page(va), ftype);
		if (rv == 0) {
			uvm_grow(p, va);
			break;
		}
		KSI_INIT_TRAP(&ksi);
		ksi.ksi_addr = (void *)va;
		ksi.ksi_trap = EXC_DTLBERR;
		if (rv == EACCES)
			ksi.ksi_code = SEGV_ACCERR;
		else
			ksi.ksi_code = SEGV_MAPERR;
		if (rv == ENOMEM) {
			printf("UVM: pid %d (%s) lid %d, uid %d killed: "
			       "out of swap\n",
			       p->p_pid, p->p_comm, l->l_lid,
			       l->l_cred ? kauth_cred_geteuid(l->l_cred) : -1);
			ksi.ksi_signo = SIGKILL;
			trapsignal(l, &ksi);
		} else {
			ksi.ksi_signo = SIGSEGV;
			trapsignal(l, &ksi);
		}
		break;

	case EXC_ITLBERR|EXC_USER:
		ftype = VM_PROT_READ | VM_PROT_EXECUTE;
		rv = uvm_fault(&p->p_vmspace->vm_map, trunc_page(frame->srr0),
		    ftype);
		if (rv == 0)
			break;
		KSI_INIT_TRAP(&ksi);
		ksi.ksi_signo = SIGSEGV;
		ksi.ksi_addr = (void *)frame->srr0;
		ksi.ksi_trap = EXC_ITLBERR;
		ksi.ksi_code = SEGV_MAPERR;
		trapsignal(l, &ksi);
		break;

	case EXC_AST|EXC_USER:
		astpending = 0;		/* we are about to do it */
		curcpu()->ci_data.cpu_nsoft++;
		if (l->l_pflag & LP_OWEUPC) {
			l->l_pflag &= ~LP_OWEUPC;
			ADDUPROF(l);
		}
		/* Check whether we are being preempted. */
		if (want_resched)
			preempt();
		break;

	case EXC_ALI|EXC_USER:
		KSI_INIT_TRAP(&ksi);
		ksi.ksi_signo = SIGBUS;
		ksi.ksi_addr = (void *)frame->srr0;
		ksi.ksi_trap = EXC_ALI;
		ksi.ksi_code = BUS_ADRALN;
		trapsignal(l, &ksi);
		break;

	case EXC_PGM|EXC_USER:
/* XXX temporarily */
		KSI_INIT_TRAP(&ksi);
		ksi.ksi_addr = (void *)frame->srr0;
		ksi.ksi_trap = EXC_PGM;
		if (frame->srr1 & 0x00020000) {
			ksi.ksi_signo = SIGTRAP;
			ksi.ksi_code = TRAP_BRKPT;
			trapsignal(l, &ksi);
		} else {
			ksi.ksi_signo = SIGILL;
			ksi.ksi_code = ILL_ILLOPC;
			trapsignal(l, &ksi);
		}
		break;

	case EXC_MCHK:
		if (pcb->pcb_onfault) {
			fb = pcb->pcb_onfault;
			frame->srr0 = (*fb)[0];
			frame->fixreg[1] = (*fb)[1];
			frame->fixreg[2] = (*fb)[2];
			frame->fixreg[3] = EFAULT;
			frame->cr = (*fb)[3];
			bcopy(&(*fb)[4], &frame->fixreg[13],
			      19 * sizeof(register_t));
			return;
		}
		goto brain_damage;

	case EXC_SOFTEMU|EXC_USER:
		if (!(pcb->pcb_flags & PCB_FPU)) {
			bzero(&pcb->pcb_fpu, sizeof pcb->pcb_fpu);
			pcb->pcb_flags |= PCB_FPU;
		}

		rv = fpu_emulate(frame, (struct fpreg *)&pcb->pcb_fpu);
		if (rv) {
			KSI_INIT_TRAP(&ksi);
			ksi.ksi_signo = rv;
			ksi.ksi_addr = (void *)frame->srr0;
			ksi.ksi_trap = EXC_SOFTEMU;
			switch (rv) {
			case SIGSEGV:
				ksi.ksi_code = SEGV_MAPERR;
				break;
			case SIGTRAP:
				ksi.ksi_code = TRAP_TRACE;
				break;
			case SIGILL:
				ksi.ksi_code = ILL_ILLOPC;
				break;
			case SIGFPE:
			default:
				ksi.ksi_code = FPE_FLTINV;	/* XXX */
				break;
			}
			trapsignal(l, &ksi);
		}
		break;

	default:
brain_damage:
		printf("trap type %x at %lx\n", type, frame->srr0);
#ifdef DDB
		if (kdb_trap(type, frame))
			return;
#endif
#ifdef TRAP_PANICWAIT
		printf("Press a key to panic.\n");
		cngetc();
#endif
		panic("trap");
	}

	userret(l, frame);
}

int
copyin(udaddr, kaddr, len)
	const void *udaddr;
	void *kaddr;
	size_t len;
{
	vaddr_t up = (vaddr_t)udaddr;
	char *kp = kaddr;
	faultbuf env;
	int rv, n;
	paddr_t pa;
	struct vm_map *map;

	if ((rv = setfault(env)) != 0) {
		curpcb->pcb_onfault = 0;
		return rv;
	}

	while (len > 0) {
		n = MIN(NBPG - (up & PGOFSET), len);
		if (pmap_extract(curpcb->pcb_pm, up, &pa) == 0) {
			struct proc *p = curlwp->l_proc;
			map = &p->p_vmspace->vm_map;
			rv = uvm_fault(map, trunc_page(up), VM_PROT_READ);
			if (rv) {
				curpcb->pcb_onfault = 0;
				return rv;
			}
			if (pmap_extract(curpcb->pcb_pm, up, &pa) == 0)
				panic("copyin");
		}
		bcopy((void *)pa, kp, n);
		up += n;
		kp += n;
		len -= n;
	}

	asm volatile ("sync");
	curpcb->pcb_onfault = 0;
	return 0;
}

int
copyout(kaddr, udaddr, len)
	const void *kaddr;
	void *udaddr;
	size_t len;
{
	const char *kp = kaddr;
	vaddr_t up = (vaddr_t)udaddr;
	faultbuf env;
	int rv, n;
	paddr_t pa;
	struct vm_map *map;

	if ((rv = setfault(env)) != 0) {
		curpcb->pcb_onfault = 0;
		return rv;
	}

	while (len > 0) {
		n = MIN(NBPG - (up & PGOFSET), len);
		if (pmap_extract(curpcb->pcb_pm, up, &pa) == 0) {
			struct proc *p = curlwp->l_proc;
			map = &p->p_vmspace->vm_map;
			rv = uvm_fault(map, trunc_page(up), VM_PROT_WRITE);
			if (rv) {
				curpcb->pcb_onfault = 0;
				return rv;
			}
			if (pmap_extract(curpcb->pcb_pm, up, &pa) == 0)
				panic("copyout");
		}
		bcopy(kp, (void *)pa, n);
		up += n;
		kp += n;
		len -= n;
	}
	/* XXX modified attribute */

	asm volatile ("sync");
	curpcb->pcb_onfault = 0;
	return 0;
}

/*
 * kcopy(const void *src, void *dst, size_t len);
 *
 * Copy len bytes from src to dst, aborting if we encounter a fatal
 * page fault.
 *
 * kcopy() _must_ save and restore the old fault handler since it is
 * called by uiomove(), which may be in the path of servicing a non-fatal
 * page fault.
 */
int
kcopy(src, dst, len)
	const void *src;
	void *dst;
	size_t len;
{
	faultbuf env, *oldfault;
	int rv;

	oldfault = curpcb->pcb_onfault;
	if ((rv = setfault(env)) != 0) {
		curpcb->pcb_onfault = oldfault;
		return rv;
	}

	bcopy(src, dst, len);

	asm volatile ("sync");
	curpcb->pcb_onfault = oldfault;
	return 0;
}

int
badaddr(addr, size)
	void *addr;
	u_int size;
{
	faultbuf env;
	int rv;

	if ((rv = setfault(env)) != 0) {
		curpcb->pcb_onfault = 0;
		return rv;
	}

	switch (size) {
	case 1:
		(void)*(volatile int8_t *)addr;
		break;
	case 2:
		(void)*(volatile int16_t *)addr;
		break;
	default:
		(void)*(volatile int32_t *)addr;
	}
	asm volatile ("sync" ::: "memory");
	curpcb->pcb_onfault = 0;
	return 0;
}

/*
 * Start a new LWP.
 */
void
startlwp(arg)
	void *arg;
{
	int err;
	ucontext_t *uc = arg;
	struct lwp *l = curlwp;
	struct trapframe *frame = trapframe(l);

	err = cpu_setmcontext(l, &uc->uc_mcontext, uc->uc_flags);
#ifdef DIAGNOSTIC
	if (err)
		printf("Error %d from cpu_setmcontext.", err);
#endif
	kmem_free(uc, sizeof *uc);
	userret(l, frame);
}

#include <sys/userret.h>

void
userret(l, frame)
	struct lwp *l;
	struct trapframe *frame;
{
	mi_userret(l);
}
