/*	$Id: flashutil.c,v 1.35 2011/12/14 18:05:13 tsubai Exp $	*/

/*-
 * Copyright (c) 2009, 2010 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/mman.h>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <unistd.h>

#include <machine/pio.h>

#define FLASH_VENDOR_AMD 0x01
#define FLASH_DEVICE_S29GL128N  0x7e2101
#define FLASH_DEVICE_S29GL256N  0x7e2201
#define FLASH_DEVICE_S29GL512N  0x7e2301
#define FLASH_DEVICE_AM29LV160B 0x49
#define FLASH_DEVICE_AM29LV160T 0xc4
#define FLASH_DEVICE_AM29LV640  0xd7

#define FLASH_VENDOR_ATMEL 0x1f
#define FLASH_DEVICE_AT49BV160  0xc0
#define FLASH_DEVICE_AT49BV160T 0xc2

#define FLASH_VENDOR_TOSHIBA 0x98
#define FLASH_DEVICE_TC58FVT160 0xc2

#define FLASH_VENDOR_MICRON 0x2c
#define FLASH_VENDOR_INTEL  0x89
#define FLASH_DEVICE_28F008	0x99
#define FLASH_DEVICE_28F320	0x16	/* 32Mbit */
#define FLASH_DEVICE_28F640	0x17	/* 64Mbit */
#define FLASH_DEVICE_28F128	0x18	/* 128MBit */

#define FLASH_VENDOR_SHARP 0xb0
#define FLASH_DEVICE_LHF12F16	0x10	/* LH28F128BFHT-PTTL75A */
#define FLASH_DEVICE_LHF12F17	0x11	/* LH28F128BFHT-PBTL75A */

#define FLASH_VENDOR_SST 0xbf
#define FLASH_DEVICE_SST39xF512 0xd4	/* SST39LF512/SST39VF512 */
#define FLASH_DEVICE_SST39xF010 0xd5	/* SST39LF010/SST39VF010 */
#define FLASH_DEVICE_SST39xF020 0xd6	/* SST39LF020/SST39VF020 */
#define FLASH_DEVICE_SST39xF040 0xd7	/* SST39LF040/SST39VF040 */

#define ns_per_tick 1000

int probe(void);
void save_rom(u_int, u_int);
int noentry();
int amd_erase_sector(u_int);
int amd_write_word(u_int, u_int);
int amd_lock_status(void);
int atmel_lock_sector(u_int);
int atmel_lock_status(void);
int sst_erase_sector(u_int);
int sst_write_word(u_int, u_int);
int sst_lock_status(void);
int micron_erase_sector(u_int);
int micron_write_word(u_int, u_int);
int micron_lock_status(void);
void write_sector(u_int, char *);
void read_sector(u_int, char *);
u_int inXX(u_int);
void outXX(u_int, u_int);
void cmdXX(u_int, u_int);
void *map(u_int, u_int);
void delay(u_int);
void dump(void *, int);
void usage(void);
#define unmap munmap

int flash_fd;
u_int manu_id, dev_id;
u_int total_size, sector_size, word_size, bus_size;
char *chipname;

char *savefile = NULL;
char *romfile = NULL;
char *flashdev = "/dev/flash0";

char *flash_base;

int (*erase_sector)() = noentry;
int (*write_word)() = noentry;
int (*lock_sector)() = noentry;
int (*lock_status)() = noentry;

char rombuf[0x200000];

main(argc, argv)
	int argc;
	char **argv;
{
	FILE *fp;
	int rflag = 0, wflag = 0, lflag = 0, sflag = 0, uflag = 0, eflag = 0;
	int c, fd;
	u_int addr = -1;

	while ((c = getopt(argc, argv, "e:f:l:o:rsuw:")) != -1) {
		switch (c) {
		case 'e':
			eflag = 1;
			addr = strtoul(optarg, NULL, 16);
			break;
		case 'f':
			flashdev = optarg;
			break;
		case 'l':
			lflag = 1;
			addr = strtoul(optarg, NULL, 16);
			break;
		case 'o':
			savefile = optarg;
			break;
		case 'r':
			rflag = 1;
			if (optind < argc && argv[optind][0] != '-')
				addr = strtoul(argv[optind++], NULL, 16);
			break;
		case 's':
			sflag = 1;
			break;
		case 'u':
			uflag = 1;
			break;
		case 'w':
			wflag = 1;
			addr = strtoul(optarg, NULL, 16);
			break;
		default:
			usage();
		}
	}
	argc -= optind;
	argv += optind;

	if (wflag && argc > 0)
		romfile = *argv--;
	if (wflag) {
		if (romfile == NULL)
			usage();
		fp = fopen(romfile, "r");
		if (fp == NULL)
			errx(0, "failed to open romfile");

		memset(rombuf, 0xff, sizeof rombuf);
		fread(rombuf, sizeof rombuf, 1, fp);
	}

	if (rflag && savefile == NULL)
		usage();

	fd = open(flashdev, 2);
	if (fd < 0)
		errx(0, "failed to open device");

	flash_fd = fd;
	flash_base = map(0, 0x10000);

	word_size = 1;
	bus_size = 1;
	if (probe())
		goto matched;
	word_size = 2;
	bus_size = 2;
	if (probe())
		goto matched;
	word_size = 2;
	bus_size = 1;
	if (probe())
		goto matched;
	errx(0, "unknown flash chip");

matched:
	printf("ManuID 0x%04x\n", manu_id);
	printf("DevID  0x%04x\n", dev_id);
	printf("name: %s\n", chipname);
	printf("sector size: %d KB\n", sector_size / 1024);
	printf(" total size: %d KB\n", total_size / 1024);

	if (addr != -1) {
		if (addr % sector_size)
			errx(0, "wrong sector address");
	}

	if (lflag) {
		if (addr == -1)
			errx(0, "need lock address");
		lock_sector(addr);
	}

	if (sflag)
		lock_status();

	if (rflag) {
		if (addr == -1)
			save_rom(0, total_size);
		else
			save_rom(addr, sector_size);
	}

	if (eflag) {
		if (addr == -1)
			errx(0, "missing erase address");
		erase_sector(addr);
	}

	if (wflag) {
		if (addr == -1)
			errx(0, "missing write address");
		write_sector(addr, rombuf);
	}
}

#define CASE(vendor, name) \
    case __CONCAT(FLASH_DEVICE_, name) << 8 | __CONCAT(FLASH_VENDOR_, vendor)

int
probe()
{
	u_int manu, dev1, dev2, dev3;

	/* Reset */
	cmdXX(0, 0xf0);

	/* Read Identifier codes */
	cmdXX(0x5555, 0xaa);	/* SST, Winbond */
	cmdXX(0x2aaa, 0x55);
	cmdXX(0x5555, 0x90);

	cmdXX(0x555, 0xaa);	/* AMD */
	cmdXX(0x2aa, 0x55);
	cmdXX(0x555, 0x90);

	manu = inXX(0);
	dev1 = inXX(1  * word_size);
	dev2 = inXX(14 * word_size);
	dev3 = inXX(15 * word_size);

	/* Reset */
	cmdXX(0, 0xf0);

	dev1 &= 0xff;
	dev2 &= 0xff;
	dev3 &= 0xff;

	if (dev1 == 0x7e)
		dev1 = dev1 << 16 | dev2 << 8 | dev3;

	manu_id = manu;
	dev_id = dev1;

	switch (dev1 << 8 | manu) {

	CASE (TOSHIBA, TC58FVT160):
		/* 64KB * 32 = 2MB */
		chipname = "TC58FVT160";
		sector_size = 64*1024;
		total_size = sector_size * 32;
		erase_sector = amd_erase_sector;
		write_word = amd_write_word;
		lock_status = amd_lock_status;
		return 1;

		/*
		 * SA0  0x000000 - 0x00ffff (64KB)
		 * SA1  0x010000 - 0x01ffff (64KB)
		 * ...
		 * SA30 0x1e0000 - 0x1effff (64KB)
		 * SA31 0x1f0000 - 0x1f7fff (32KB)
		 * SA32 0x1f8000 - 0x1f9fff (8KB)
		 * SA33 0x1fa000 - 0x1fbfff (8KB)
		 * SA34 0x1fc000 - 0x1fffff (16KB)
		 */

	CASE (AMD, AM29LV160B):
		/* 64KB * 32 = 2MB */
		chipname = "AM29LV160B";
		sector_size = 64*1024;
		total_size = sector_size * 32;
		erase_sector = amd_erase_sector;
		write_word = amd_write_word;
		return 1;

	CASE (AMD, AM29LV640):
		/* 64KB * 128 = 8MB */
		chipname = "Am29LV640";
		sector_size = 64*1024;
		total_size = sector_size * 128;
		erase_sector = amd_erase_sector;
		write_word = amd_write_word;
		lock_status = amd_lock_status;
		return 1;

	CASE (AMD, S29GL512N):
		/* 128KB * 512 = 64MB */
		chipname = "S29GL512N";
		sector_size = 128*1024;
		total_size = sector_size * 512;
		erase_sector = amd_erase_sector;
		write_word = amd_write_word;
		lock_status = amd_lock_status;
		return 1;

	CASE (ATMEL, AT49BV160):
	CASE (ATMEL, AT49BV160T):
		/* 64KB * 32 = 2MB, 8/16bit */
		chipname = "AT49BV160";
		sector_size = 64*1024;
		total_size = sector_size * 32;
		/* XXX ǡȤˤ 555/2aa ưʤ */
		erase_sector = sst_erase_sector;
		write_word = sst_write_word;
		lock_sector = atmel_lock_sector;
		lock_status = atmel_lock_status;
		/* XXX ʤ󤫥åǤʤ */
		return 1;
		/*
		 * 160:		160T:
		 * SA0 8KB	SA0 64KB
		 * SA1 8KB	SA1 64KB
		 * ...		...
		 * SA7 8KB	SA30 64KB
		 * SA8 64KB	SA31 8KB
		 * SA9 64KB	SA32 8KB
		 * ...		...
		 * SA38 64KB	SA38 8KB
		 */

	CASE (SST, SST39xF020):
		/* 4KB * 64 = 256KB */
		chipname = "SST39xF020";
		sector_size = 4096;
		total_size = sector_size * 64;
		erase_sector = sst_erase_sector;
		write_word = sst_write_word;
		lock_status = sst_lock_status;
		return 1;

	CASE (INTEL, 28F008):
		/* XXX no datasheet */
		chipname = "28F008";
		sector_size = 4096;
		total_size = sector_size * 256;
		lock_status = micron_lock_status;
		/* erase_sector = micron_erase_sector;	XXX not yet */
		/* write_word = micron_write_word;	XXX not yet */
		return 1;

		/*
		 * From UniFlash: 28F008-B
		 *  16KB
		 *   8KB
		 *   8KB
		 *  96KB
		 * 128KB * 7
		 */

	CASE (INTEL, 28F640):
		chipname = "28F640";
		sector_size = 128*1024;
		total_size = sector_size * 64;
		erase_sector = micron_erase_sector;
		write_word = micron_write_word;
		lock_status = micron_lock_status;
		return 1;

	CASE (INTEL, 28F128):
		chipname = "28F128";
		sector_size = 128*1024;
		total_size = sector_size * 128;
		erase_sector = micron_erase_sector;
		write_word = micron_write_word;
		lock_status = micron_lock_status;
		return 1;
	}

	return 0;
}

void
save_rom(addr, len)
	u_int addr, len;
{
	FILE *fp;
	char *buf;

	buf = malloc(sector_size);
	if (buf == NULL)
		err(0, "malloc");

	fp = fopen(savefile, "w");

	len /= sector_size;

	while (len-- > 0) {
		read_sector(addr, buf);
		fwrite(buf, sector_size, 1, fp);
		addr += sector_size;
	}
	fclose(fp);
}

int
noentry(addr)
	u_int addr;
{
	return -1;
}

int
amd_erase_sector(addr)
	u_int addr;		/* Sector Address */
{
	int i;
	u_int a, b;

	printf("Erasing offset 0x%x\n", addr);

	/* Sector Erase */
	cmdXX(0x555, 0xaa);
	cmdXX(0x2aa, 0x55);
	cmdXX(0x555, 0x80);
	cmdXX(0x555, 0xaa);
	cmdXX(0x2aa, 0x55);
	outXX(addr,  0x30);

	/* Typ. 1sec */
	for (i = 0; i < 1000; i++) {
		usleep(10000);
		a = inXX(addr);
		b = inXX(addr);
		if (a == b)
			return 0;
	}
	printf("Erase timeout!\n");
	return -1;
}

int
sst_erase_sector(addr)
	u_int addr;		/* Sector Address */
{
	int i;
	u_int a, b;

	printf("Erasing offset 0x%x\n", addr);

	/* Sector Erase */
	cmdXX(0x5555, 0xaa);
	cmdXX(0x2aaa, 0x55);
	cmdXX(0x5555, 0x80);
	cmdXX(0x5555, 0xaa);
	cmdXX(0x2aaa, 0x55);
	outXX(addr,   0x30);

	/* Typ. 1sec */
	for (i = 0; i < 1000; i++) {
		usleep(10000);
		a = inXX(addr);
		b = inXX(addr);
		if (a == b)
			return 0;
	}
	printf("Erase timeout!\n");
	return -1;
}

int
micron_erase_sector(addr)
	u_int addr;		/* Sector Address */
{
	int i;

	printf("Erasing offset 0x%x\n", addr);

	outXX(addr, 0x20);
	outXX(addr, 0xd0);

	/* Typ. 0.75 sec, max 5 sec */
	for (i = 0; i < 1000; i++) {
		usleep(10000);
		if (inXX(addr) & 0x80)
			return 0;
	}
	printf("Erase timeout!\n");
	return -1;
}

int
amd_write_word(addr, val)
	u_int addr, val;
{
	u_int i, a, b;

	if (inXX(addr) == val)
		return 0;

	cmdXX(0x555, 0xaa);
	cmdXX(0x2aa, 0x55);
	cmdXX(0x555, 0xa0);
	outXX(addr,  val);

	/* Typ. 10us */
	for (i = 0; i < 1000; i++) {
		delay(1);
		a = inXX(addr);
		b = inXX(addr);
		if (a == b)
			return 0;
	}
	return -1;	/* Write timeout! */
}

int
sst_write_word(addr, val)
	u_int addr, val;
{
	u_int i, a, b;

	if (inXX(addr) == val)
		return 0;

	cmdXX(0x5555, 0xaa);
	cmdXX(0x2aaa, 0x55);
	cmdXX(0x5555, 0xa0);
	outXX(addr,   val);

	/* Typ. 10us */
	for (i = 0; i < 1000; i++) {
		delay(1);
		a = inXX(addr);
		b = inXX(addr);
		if (a == b)
			return 0;
	}
	return -1;	/* Write timeout! */
}

int
micron_write_word(addr, val)
	u_int addr, val;
{
	u_int i;

	cmdXX(0, 0xff);		/* Read Array */

	if (inXX(addr) == val)
		return 0;

	cmdXX(0, 0x40);
	outXX(addr, val);

	/* Typ. 10us, max 600us */
	for (i = 0; i < 1000; i++) {
		delay(1);
		if (inXX(addr) & 0x80) {
			cmdXX(0, 0xff);
			return 0;
		}
	}
	return -1;	/* Write timeout! */
}

int
atmel_lock_sector(addr)
	u_int addr;
{
	cmdXX(0x5555, 0xaa);
	cmdXX(0x2aaa, 0x55);
	cmdXX(0x5555, 0x80);
	cmdXX(0x5555, 0xaa);
	cmdXX(0x2aaa, 0x55);
	outXX(addr,   0x60);
	return 0;
}

int
amd_lock_status()
{
	return -1;
}

int
atmel_lock_status()
{
	u_int addr, x, i = 1;

	printf("\nBlock Lock status\n");
	cmdXX(0x5555, 0xaa);
	cmdXX(0x2aaa, 0x55);
	cmdXX(0x5555, 0x90);

	/* XXX !uniform sector... */
	for (addr = 0; addr < total_size; addr += sector_size) {
		x = inXX(addr + 2 * word_size);
		printf("0x%06x: %s    ", addr, (x & 1) ? "locked" : "  -   ");
		if (i++ % 4 == 0)
			printf("\n");
	}
	cmdXX(0, 0xf0);
}

int
sst_lock_status()
{
	return -1;
}

int
micron_lock_status()
{
	u_int addr, i = 1;

	printf("\nBlock Lock status\n");
	cmdXX(0, 0x90);

	for (addr = 0; addr < total_size; addr += sector_size) {
		printf("0x%06x: %s    ", addr,
		    (inXX(addr + 2 * word_size) & 1) ? "locked" : "  -   ");
		if (i++ % 4 == 0)
			printf("\n");
	}
	cmdXX(0, 0xff);
	return 0;
}

void
write_sector(addr, data)
	u_int addr;
	char *data;
{
	int n = sector_size / bus_size;
	int error_count = 0;
	int i;
	u_int x;

	printf("Programming offset 0x%x\n", addr);
	for (i = 0; i < n; i++) {
		switch (bus_size) {
		case 1: x = *(u_int8_t *)data; break;
		case 2: x = *(u_int16_t *)data; break;
		default: errx(0, "strange word size");
		}

		if (write_word(addr, x)) {
			printf("X"); fflush(stdout);	/* Write timeout! */
			if (error_count++ > 10) {
				printf("\n");
				printf("excessive write timeout; ");
				printf("stop programming\n");
				errx(0, "write failed");
			}
		}
		if ((i & 0xfff) == 0) {
			printf("."); fflush(stdout);
		}

		addr += bus_size;
		data += bus_size;
	}
	printf("\n");
}

void
read_sector(addr, buf)
	u_int addr;
	char *buf;
{
	char *p;

	p = map(addr, sector_size);
	bcopy(p, buf, sector_size);
	unmap(p, sector_size);
}

char *current_base;
u_int current_sector = -1;

u_int
inXX(addr)
	u_int addr;
{
	u_int sec, off;

	if (addr < 0x10000)
		switch (bus_size) {
		case 1:
			return in8(flash_base + addr);
		case 2:
			return in16(flash_base + addr);
		default:
			errx(0, "strange word size");
		}

	off = addr & (sector_size - 1);
	sec = addr - off;
	if (sec != current_sector) {
		unmap(current_base, sector_size);
		current_base = map(sec, sector_size);
		current_sector = sec;
	}
	switch (bus_size) {
	case 1:
		return in8(current_base + off);
	case 2:
		return in16(current_base + off);
	default:
		errx(0, "strange word size");
	}
}

void
outXX(addr, val)
	u_int addr, val;
{
	u_int sec, off;

	if (addr < 0x10000)
		switch (bus_size) {
		case 1:
			return out8(flash_base + addr, val);
		case 2:
			return out16(flash_base + addr, val);
		default:
			errx(0, "strange word size");
		}

	off = addr & (sector_size - 1);
	sec = addr - off;
	if (sec != current_sector) {
		unmap(current_base, sector_size);
		current_base = map(sec, sector_size);
		current_sector = sec;
	}
	switch (bus_size) {
	case 1:
		return out8(current_base + off, val);
	case 2:
		return out16(current_base + off, val);
	default:
		errx(0, "strange word size");
	}
}

void
cmdXX(addr, val)
	u_int addr, val;
{
	u_int sec, off;

	addr *= word_size;
	if (addr > 0x10000)
		errx(0, "strange cmd address");

	switch (bus_size) {
	case 1:
		return out8(flash_base + addr, val);
	case 2:
		return out16(flash_base + addr, val);
	default:
		errx(0, "strange word size");
	}
}

void *
map(off, len)
	u_int off, len;
{
	void *p;

	p = mmap(NULL, len, PROT_READ|PROT_WRITE, MAP_SHARED, flash_fd, off);
	if (p == MAP_FAILED)
		err(0, "mmap");
	return p;
}

static inline u_int64_t
mftb()
{
	u_int64_t tb;
	int tmp;

	asm volatile ("1: mftbu %0; mftb %0+1; mftbu %1; cmplw %0,%1; bne- 1b"
	    : "=r"(tb), "=r"(tmp));
	return tb;
}

void
delay(n)
	unsigned int n;
{
	u_int64_t tb;
	u_long tbh, tbl, scratch;

	tb = mftb();
	tb += (n * 1000 + ns_per_tick - 1) / ns_per_tick;
	tbh = tb >> 32;
	tbl = tb;

	asm volatile ("			\n\
	1:	mftbu	%0		\n\
		sub.	%0,%0,%1	\n\
		blt	1b		\n\
		bgt	2f		\n\
		mftb	%0		\n\
		cmplw	%0,%2		\n\
		blt	1b		\n\
	2:				\n"
		: "=&r"(scratch) : "r"(tbh), "r"(tbl));

	asm volatile ("sync; isync");
}

void
dump(addr, len)
	void *addr;
	int len;
{
	char *p = addr;
	int i;

	while (len > 0) {
		printf("%08x", (int)p);
		for (i = 0; i < 16; i++)
			printf(" %02x", *p++);
		printf("\n");
		len -= 16;
	}
}

void
usage()
{
	printf("usage: flashutil [-r [offset] -o savefile] [-e offset]\n");
	printf("       flashutil -w offset romfile\n");
	exit(0);
}

/*
 * OpenBlockS  flash ROM 
 *   0xffc00000 - 0xffdfffff
 *   0xffe00000 - 0xffffffff
 *  2ġ
 * Ⱦ 0x030000 ֡ȥϿƤ롣
 */

/*
 * TS400
 *   0x20000 config
 *   0x40000 entry point
 */
