/*	$Id: cpu_7400.c,v 1.8 2009/01/01 10:11:59 tsubai Exp $	*/

/*-
 * Copyright (c) 2004 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/systm.h>

#include <machine/cpufunc.h>
#include <machine/spr.h>

/* Hardware Implementation Dependent registers for the 7400 and 7410 */

#define HID0_EMCP   0x80000000  /* Enable MCP */
#define HID0_EBA    0x20000000  /* Enable 60x bus address parity checking */
#define HID0_EBD    0x10000000  /* Enable 60x bus data parity checking */
#define HID0_BCLK   0x08000000  /* CLK_OUT clock type selection */
#define HID0_ECLK   0x02000000  /* CLK_OUT clock type selection */
#define HID0_PAR    0x01000000  /* Disable precharge of ARTRY */
#define HID0_DOZE   0x00800000  /* Enable doze mode */
#define HID0_NAP    0x00400000  /* Enable nap mode */
#define HID0_SLEEP  0x00200000  /* Enable sleep mode */
#define HID0_DPM    0x00100000  /* Enable Dynamic power management */
#define HID0_RISEG  0x00080000  /* Read I-SEG */
#define HID0_EIEC   0x00040000  /* Enable internal error checking */
#define HID0_NHR    0x00010000  /* Not hard reset */
#define HID0_ICE    0x00008000  /* Enable i-cache */
#define HID0_DCE    0x00004000  /* Enable d-cache */
#define HID0_ILOCK  0x00002000  /* i-cache lock */
#define HID0_DLOCK  0x00001000  /* d-cache lock */
#define HID0_ICFI   0x00000800  /* i-cache flush invalidate */
#define HID0_DCFI   0x00000400  /* d-cache flush invalidate */
#define HID0_SPD    0x00000200  /* Disable speculative cache access */
#define HID0_IFTT   0x00000100  /* I-fetch TTx encoding differentiation */
#define HID0_SGE    0x00000080  /* Enable store gathering */
#define HID0_DCFA   0x00000040  /* Data cache flush assist */
#define HID0_BTIC   0x00000020  /* Enable BTIC */
#define HID0_BHT    0x00000004  /* Enable branch history table */
#define HID0_NOPDST 0x00000002  /* No-op dst, dstt, dstst, and dststt */
#define HID0_NOPTI  0x00000001  /* No-op the dcbt(st) */

#define HID0_BITMASK "\020" \
    "\040EMCP\036EBA\035EBD\034BCLK\032ECLK\031PAR" \
    "\030DOZE\027NAP\026SLEEP\025DPM\024RISEG\023EIEC\021NHR" \
    "\020ICE\017DCE\016ILOCK\015DLOCK\014ICFI\013DCFI\012SPD\011IFTT" \
    "\010SGE\007DCFA\006BTIC\003BHT\002NOPDST\001NOPTI"

#define HID1_PC0    0x80000000  /* PLL configuration bit 0 (RO) */
#define HID1_PC1    0x40000000  /* PLL configuration bit 1 (RO) */
#define HID1_PC2    0x20000000  /* PLL configuration bit 2 (RO) */
#define HID1_PC3    0x10000000  /* PLL configuration bit 3 (RO) */

#define HID1_BITMASK "\020" \
    "\040PC0\037PC1\036PC2\035PC3"

/*
 * Configure HID0 for 7400 and 7410.
 */
void
mpc7400_config_hid()
{
	u_int hid0, msscr1, rev;

	asm ("mfspr %0,HID0" : "=r"(hid0));

	hid0 &= HID0_EBA | HID0_EBD | HID0_BCLK | HID0_ECLK | HID0_PAR |
	    HID0_ICE | HID0_DCE;
	hid0 |= HID0_EMCP | HID0_DOZE | HID0_DPM;
	hid0 |= HID0_SGE | HID0_BTIC | HID0_BHT;

	switch (mfpvr() >> 16) {
	case MPC7400:
		hid0 |= HID0_EIEC;
	}

	asm ("mfspr %0,MSSCR1" : "=r"(msscr1));

	rev = mfpvr() & 0xffff;
	switch (mfpvr() >> 16) {
	case MPC7400:
		if (rev < 0x0201) {
			/* Errata 4 */
			/* Force the Data Reload Table size to 4 entries. */
			msscr1 |= 0x18000000;

			/* Errata 7 */
			hid0 &= ~HID0_SGE;

			/* Errata 10 */
			/* Set DL1 bypass disable. */
			msscr1 |= 0x00000004;
		}
		if (rev < 0x0202) {
			/* Errata 12 */
			/* Set LFQ_SIZE to 0b10. */
			msscr1 &= ~0x06000000;
			msscr1 |=  0x04000000;
		}
		if (rev < 0x0208) {
			/* Errata 13 */
			/* Set L1OPQ_SIZE to 0b10. */
			msscr1 &= ~0x01800000;
			msscr1 |=  0x01000000;

			/* Errata 14 */
			/* Set L2MQ_SIZE; limit the L2 miss queue to 1 entry. */
			msscr1 |= 0x00020000;

			/* Errata 15 */
			hid0 |= HID0_NOPDST;
		}
		break;

	case MPC7410:
		if (rev < 0x1101) {
			/* Errata 3 */
			/* Set L1OPQ_SIZE to 0b10. */
			msscr1 &= ~0x01800000;
			msscr1 |=  0x01000000;

			/* Errata 4 */
			/* Set L2MQ_SIZE. */
			msscr1 |= 0x00020000;

			/* Errata 5 */
			hid0 |= HID0_NOPDST;
		}

		if (rev >= 0x1103) {
			/* Errata 8 */
			msscr1 &= ~0x00003800;
			msscr1 |=  0x00000800;
		}
		break;

		/* XXX Errata 6 stwcx */
	}
	asm volatile ("sync; isync");
	asm volatile ("mtspr MSSCR1,%0" :: "r"(msscr1));
	asm volatile ("mtspr HID0,%0" :: "r"(hid0));
}

void
mpc7400_print_hid(name)
	const char *name;
{
	u_int hid0;
	char buf[160];

	asm ("mfspr %0,HID0" : "=r"(hid0));
	snprintb(buf, sizeof buf, HID0_BITMASK, hid0);
	printf("%s: HID0 %s\n", name, buf);
}

u_int
mpc7400_corefreq(busfreq)
	u_int busfreq;
{
	u_int hid1, pll;

	/* Bus to Core Multiplier * 2 */
	static const char pll_mul2[] = {
		18, 15, 14,  2,  4, 13,  5,  9,
		 6, 11,  8, 10, 16, 12,  7,  0
	};

	asm ("mfspr %0,HID1" : "=r"(hid1));
	pll = hid1 >> 28;

	return busfreq * pll_mul2[pll] / 2;
}
