/*	$Id: cpu_7450.c,v 1.60 2010/12/30 22:34:17 tsubai Exp $	*/

/*-
 * Copyright (c) 2002-2004 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "opt_ictc_config.h"
#include "opt_l3cr_config.h"
#include "opt_multiprocessor.h"

#include <sys/param.h>
#include <sys/callout.h>
#include <sys/kernel.h>
#include <sys/reboot.h>
#include <sys/systm.h>

#include <uvm/uvm_extern.h>

#include <machine/cpufunc.h>
#include <machine/spr.h>

/*
 * 7450/7455 specific part.
 */
#define HID0_TBEN    0x04000000	/* Timebase enable */
#define HID0_STEN    0x01000000	/* Software table search enable */
#define HID0_HBATEN  0x00800000	/* Enable BAT[4-7] (7455) */
#define HID0_NAP     0x00400000	/* Enable nap mode */
#define HID0_SLEEP   0x00200000	/* Enable sleep mode */
#define HID0_DPM     0x00100000	/* Enable Dynamic power management */
#define HID0_BHTCLR  0x00040000	/* Clear branch history table */
#define HID0_XAEN    0x00020000	/* Enable extended Addressing */
#define HID0_NHR     0x00010000	/* Not hard reset */
#define HID0_ICE     0x00008000	/* Enable i-cache */
#define HID0_DCE     0x00004000	/* Enable d-cache */
#define HID0_ILOCK   0x00002000	/* i-cache lock */
#define HID0_DLOCK   0x00001000	/* d-cache lock */
#define HID0_ICFI    0x00000800	/* i-cache flush invalidate */
#define HID0_DCFI    0x00000400	/* d-cache flush invalidate */
#define HID0_SPD     0x00000200	/* Disable speculative cache access */
#define HID0_XBSEN   0x00000100	/* Enable extended BAT size (7455) */
#define HID0_SGE     0x00000080	/* Enable store gathering */
#define HID0_BTIC    0x00000020	/* Enable BTIC */
#define HID0_LRSTK   0x00000010	/* Enable link register stack */
#define HID0_FOLD    0x00000008	/* Enable branch folding */
#define HID0_BHT     0x00000004	/* Enable branch history table */
#define HID0_NOPDST  0x00000002	/* No-op dst, dstt, dstst, and dststt */
#define HID0_NOPTI   0x00000001	/* No-op the dcbt(st) */

#define HID0_BITMASK "\020" \
    "\033TBEN\031STEN\030HBATEN" \
    "\027NAP\026SLEEP\025DPM\023BHTCLR\022XAEN\021NHR" \
    "\020ICE\017DCE\016ILOCK\015DLOCK\014ICFI\013DCFI\012SPD\011XBSEN" \
    "\010SGE\006BTIC\005LRSTK\004FOLD\003BHT\002NOPDST\001NOPTI"

#define HID1_EMCP    0x80000000	/* Enable MCP */
#define HID1_EBA     0x20000000	/* Enable bus addr parity checking */
#define HID1_EBD     0x10000000	/* Enable bus data parity checking */
#define HID1_BCLK    0x08000000	/* CLK_OUT clock type selection */
#define HID1_ECLK    0x02000000	/* CLK_OUT clock type selection */
#define HID1_PAR     0x01000000	/* Disable precharge of ARTRY */
#define HID1_DFS4    0x00800000	/* DFS divide-by-four mode (7448) */
#define HID1_DFS2    0x00400000	/* DFS divide-by-two mode (7447A) */
#define HID1_PC5     0x00020000	/* PLL configuration bit 5 (RO) (7448) */
#define HID1_PC0     0x00010000	/* PLL configuration bit 0 (RO) */
#define HID1_PC1     0x00008000	/* PLL configuration bit 1 (RO) */
#define HID1_PC2     0x00004000	/* PLL configuration bit 2 (RO) */
#define HID1_PC3     0x00002000	/* PLL configuration bit 3 (RO) */
#define HID1_PC4     0x00001000	/* PLL configuration bit 4 (RO) */
#define HID1_SYNCBE  0x00000800	/* Enable address broadcast for sync, eieio */
#define HID1_ABE     0x00000400	/* Enable address broadcast for dcbf, dcbst,
                                   dcbi, icbi, tlbie, and tlbsync */

#define HID1_BITMASK "\020" \
    "\040EMCP\036EBA\035EBD\034BCLK\032ECLK\031PAR" \
    "\030DFS4\027DFS2" \
    "\022PC5\021PC0\020PC1" \
    "\017PC2\016PC3\015PC4\014SYNCBE\013ABE"

#define L2CR_L2E     0x80000000	/* 0: L2 enable */
#define L2CR_L2PE    0x40000000	/* 1: L2 data parity checking enable */
				/* 2-3: reserved (must be 0b00) */
#define L2CR_L2I     0x00200000	/* 10: L2 global invalidate */
#define L2CR_L2IO    0x00100000	/* 11: L2 instruction only */
#define L2CR_L3OH0   0x00080000	/* 12: L3 output hold 0 (7455) */
#define L2CR_L2DO    0x00010000	/* 15: L2 data only */
#define L2CR_L2REP   0x00001000	/* 19: L2 replacement algorithm */
#define L2CR_L2HWF   0x00000800	/* 20: L2 hardware flush */
#define L2CR_LVRAME  0x00000080	/* 24: LVRAM enable (7448) */
#define L2CR_LVRAMM  0x00000070	/* 25-27: LVRAM mode (7448) */

#define L2CR_BITMASK "\020" \
    "\040L2E\037L2PE\026L2I\025L2IO\024L3OH0\021L2DO\015L2REP\014L2HWF" \
    "\010LVRAME"

#define L3CR_L3E     0x80000000	/* 0: L3 enable */
#define L3CR_L3PE    0x40000000	/* 1: L3 data parity checking enable */
#define L3CR_L3APE   0x20000000	/* 2: L3 address parity checking enable */
#define L3CR_L3SIZ   0x10000000	/* 3: L3 size (0:1MB, 1:2MB) */
#define L3CR_L3CLKEN 0x08000000	/* 4: enable L3_CLK signals */
				/* 5: reserved (must be 1) */
#define L3CR_L3CLK   0x03800000	/* 6-8: L3 clock ratio */
#define  L3CLK_6	0x00000000	/* 1 / 6   (!7457) */
#define  L3CLK_2	0x01000000	/* 1 / 2   (!7457) */
#define  L3CLK_25	0x01800000	/* 1 / 2.5 (!7457) */
#define  L3CLK_3	0x02000000	/* 1 / 3   (!7457) */
#define  L3CLK_35	0x02800000	/* 1 / 3.5 (!7457) */
#define  L3CLK_4	0x03000000	/* 1 / 4   (!7457) */
#define  L3CLK_5	0x03800000	/* 1 / 5   (!7457) */
#define  L3CLK_7	0x00000000	/* 1 / 7   (7457) */
#define  L3CLK_8	0x00800000	/* 1 / 8   (7457) */
#define  L3CLK_45	0x01000000	/* 1 / 4.5 (7457) */
#define  L3CLK_55	0x01800000	/* 1 / 5.5 (7457) */
#define  L3CLK_65	0x02000000	/* 1 / 6.5 (7457) */
#define  L3CLK_75	0x02800000	/* 1 / 7.5 (7457) */
#define L3CR_L3IO    0x00400000	/* 9: L3 instruction only */
#define L3CR_L3CLKEXT 0x00200000 /* 10: L3 clock ratio extension (7457) */
#define L3CR_L3CKSPEXT 0x00100000 /* 11: L3 clock sample point ext. (7457) */
#define L3CR_L3OH1   0x00080000	/* 12: L3 output hold 1 (7455) */
#define L3CR_L3SPO   0x00040000	/* 13: L3 sample point override */
#define L3CR_L3CKSP  0x00030000	/* 14-15: L3 clock sample point */
#define  L3CKSP0	0x00000000
#define  L3CKSP1	0x00010000
#define  L3CKSP2	0x00020000
#define  L3CKSP3	0x00030000
#define L3CR_L3PSP   0x0000e000	/* 16-18: L3 P-clock sample point */
#define L3CR_L3REP   0x00001000	/* 19: L3 replacement algorithm */
#define L3CR_L3HWF   0x00000800	/* 20: L3 hardware flush */
#define L3CR_L3I     0x00000400	/* 21: L3 global invalidate */
#define L3CR_L3RT    0x00000300	/* 22-23: L3 SRAM type */
#define  L3RT_DDR	0x00000000	/* MSUG2 DDR */
#define  L3RT_LATE	0x00000100	/* pipelined synchronous late-write */
#define  L3RT_PB2	0x00000300	/* PB2 */
#define L3CR_L3NIRCA 0x00000080	/* 24: L3 SRAM clock control */
#define L3CR_L3DO    0x00000040	/* 25: L3 data only */
#define L3CR_PMEN    0x00000004	/* 29: private memory enable */
#define L3CR_PMSIZ   0x00000003	/* 30-31: private memory size */
#define  L3CR_PMSIZ_1M  0x00000000	/* 1MB */
#define  L3CR_PMSIZ_2M  0x00000001	/* 2MB */
#define  L3CR_PMSIZ_4M  0x00000002	/* 4MB (7457) */

#define L3CR_BITMASK "\020" \
    "\040L3E\037L3PE\036L3APE\035L3SIZ\034L3CLKEN" \
    "\027L3IO\026L3CLKEXT\025L3CKSPEXT\024L3OH1\023L3SPO" \
    "\015L3REP\014L3HWF\013L3I" \
    "\010L3NIRCA\007L3DO\003PMEN"

#define ICTRL_CIRQ 0x80000000	/* CPU interrupt request */
#define ICTRL_EIEC 0x08000000	/* Instruction cache parity error enable */
#define ICTRL_EDCE 0x04000000	/* Data cache parity error enable */
#define ICTRL_EICP 0x00000100	/* Enable instruction cache parity checking */
#define ICTRL_ICWL 0x000000ff	/* Instruction cache way lock */

#define ICTRL_BITMASK "\020\040CIRQ\034EIEC\033EDCE\011EICP"

#ifdef L3CR_CONFIG
u_int l3cr_config = L3CR_CONFIG;
#else
u_int l3cr_config = 0;
#endif

#ifdef ICTC_CONFIG
u_int ictc_config = ICTC_CONFIG;
#else
u_int ictc_config = 0;
#endif

asm (".machine altivec");

void
mpc7450_config_hid()
{
	u_int hid0, hid1, msscr0, rev;

	asm ("mfspr %0,HID0" : "=r"(hid0));
	asm ("mfspr %0,HID1" : "=r"(hid1));

	switch (mfpvr() >> 16) {
	case MPC7450:
		hid0 &= HID0_TBEN | HID0_ICE | HID0_DCE;
		break;
	default:
		hid0 &= HID0_TBEN | HID0_ICE | HID0_DCE | HID0_XBSEN;
		break;
	}
	hid0 |= HID0_DPM;
	hid0 |= HID0_SGE | HID0_BTIC | HID0_LRSTK | HID0_FOLD | HID0_BHT;
#if 0 /* Can we really use nap mode? */
	hid0 |= HID0_NAP;
#endif

	rev = mfpvr() & 0xffff;
	switch (mfpvr() >> 16) {
	case MPC7450:
		if (rev < 0x0200) {
			/* 7450 Errata 14 */
			hid0 &= ~HID0_NOPTI;

			/* 7450 Errata 31 */
			hid0 &= ~HID0_BTIC;
		}
		if (rev < 0x0203) {
			/* XXX 7450 Errata 39 */
		}
		break;

	case MPC7457:
		if (rev < 0x0102) {
			/* 7457 Errata 20 */
			hid0 &= ~HID0_BTIC;
		}
		break;
	}

#define MSSCR0_L2PFE 0x00000003 /* L2 prefetching enable */
	/* Disable hardware prefetch. */
	asm ("mfspr %0,MSSCR0" : "=r"(msscr0));
	msscr0 &= ~MSSCR0_L2PFE;
	asm volatile ("sync; mtspr MSSCR0,%0" :: "r"(msscr0));

	hid1 |= HID1_EMCP | HID1_SYNCBE | HID1_ABE;
	asm volatile ("mtspr HID0,%0" :: "r"(hid0));
	asm volatile ("mtspr HID1,%0" :: "r"(hid1));
	asm volatile ("isync");

#ifdef ICTC_CONFIG
	/* Instruction Cache Throttling Control */
	/* 3  1/2 餤®٤ */
	ictc_config &= 0xff;
	asm volatile ("mtspr 1019,%0" :: "r"(ictc_config << 1 | 1));
#endif
}

/*
 * Enable extended BAT size to use BAT_BL_1G to map physical memory.
 */
void
mpc7455_xbsen()
{
	u_int hid0;

	asm ("mfspr %0,HID0" : "=r"(hid0));
	hid0 |= HID0_XBSEN;
	asm volatile ("mtspr HID0,%0; isync" :: "r"(hid0));
}

/*
 * Print HID[01] configuration.
 */
void
mpc7450_print_hid(name)
	const char *name;
{
	u_int hid0, hid1;
	char buf[128];

	asm ("mfspr %0,HID0" : "=r"(hid0));
	asm ("mfspr %0,HID1" : "=r"(hid1));

	snprintb(buf, sizeof buf, HID0_BITMASK, hid0);
	printf("%s: HID0 %s\n", name, buf);
	snprintb(buf, sizeof buf, HID1_BITMASK, hid1);
	printf("%s: HID1 %s\n", name, buf);
}

void
mpc7450_enable_nap()
{
	u_int hid0;

	asm ("mfspr %0,HID0" : "=r"(hid0));

	hid0 &= ~(HID0_NAP | HID0_SLEEP);
	hid0 |= HID0_NAP;

	asm volatile ("mtspr HID0,%0" :: "r"(hid0));
}

void
mpc7450_enable_sleep()
{
	u_int hid0;

	asm ("mfspr %0,HID0" : "=r"(hid0));

	hid0 &= ~(HID0_NAP | HID0_SLEEP);
	hid0 |= HID0_SLEEP;

	asm volatile ("mtspr HID0,%0" :: "r"(hid0));
}

void
mpc7450_enable_L1cache()
{
	u_int hid0;

	asm volatile ("mtspr ICTRL,%0"
	    :: "r"(ICTRL_EIEC|ICTRL_EDCE|ICTRL_EICP));

	/* Enable L1 caches. */
	asm volatile ("mfspr %0,HID0" : "=r"(hid0));
	if ((hid0 & HID0_ICE) == 0) {
		hid0 |= HID0_ICFI|HID0_ICE;
		asm volatile ("isync");
		asm volatile ("mtspr HID0,%0" :: "r"(hid0));
		asm volatile ("isync");
	}

	asm volatile ("mfspr %0,HID0" : "=r"(hid0));
	if ((hid0 & HID0_DCE) == 0) {
		hid0 |= HID0_DCFI|HID0_DCE;
		asm volatile ("sync");
		asm volatile ("mtspr HID0,%0" :: "r"(hid0));
	}
}

/*
 * Invalidate and enable the L2 cache.
 */
void
mpc7450_enable_L2cache()
{
	u_int l2cr;

	asm volatile ("mfspr %0,L2CR" : "=r"(l2cr));

	/* Do nothing if already enabled. */
	if (l2cr & L2CR_L2E)
		return;

	asm volatile ("dssall; sync");

	/* Invalidate the entire L2 cache. */
	l2cr |= L2CR_L2I;
	asm volatile ("mtspr L2CR,%0" :: "r"(l2cr));

	/* Wait for the invalidation done. */
	while (l2cr & L2CR_L2I)
		asm volatile ("mfspr %0,L2CR" : "=r"(l2cr));

	/* Enable L2 cache. */
	l2cr |= L2CR_L2PE;
	asm volatile ("mtspr L2CR,%0" :: "r"(l2cr));
	l2cr |= L2CR_L2E;
	asm volatile ("mtspr L2CR,%0" :: "r"(l2cr));
}

/*
 * Invalidate and enable the L3 cache.
 */
void
mpc7450_enable_L3cache(l3param)
	u_int l3param;
{
	u_int l3cr;

	asm ("mfspr %0,L3CR" : "=r"(l3cr));

	/* Do nothing if already enabled. */
	if (l3cr & L3CR_L3E)
		return;

	l3cr = l3param & L3CR_L3CLK;
	asm volatile ("mtspr L3CR,%0" :: "r"(l3cr));

	l3cr = l3param & ~(L3CR_L3E|L3CR_L3I|L3CR_L3PE|L3CR_L3CLKEN);
	asm volatile ("mtspr L3CR,%0" :: "r"(l3cr));

	/* Make sure L3CR[5] == 1. */
	l3cr |= 0x04000000;

	l3cr |= L3CR_L3CLKEN;
	asm volatile ("mtspr L3CR,%0" :: "r"(l3cr));
	delay(2);

	/* Invalidate the entire L3 cache. */
	asm volatile ("dssall; sync");
	l3cr |= L3CR_L3I;
	asm volatile ("mtspr L3CR,%0" :: "r"(l3cr));

	/* Wait for the invalidation done. */
	while (l3cr & L3CR_L3I)
		asm volatile ("mfspr %0,L3CR" : "=r"(l3cr));

	l3cr &= ~L3CR_L3CLKEN;
	asm volatile ("mtspr L3CR,%0" :: "r"(l3cr));
	asm volatile ("sync");
	delay(2);

	l3cr |= L3CR_L3E | L3CR_L3CLKEN;
	asm volatile ("mtspr L3CR,%0" :: "r"(l3cr));
	asm volatile ("sync");
	delay(2);
}

u_int
mpc7450_corefreq(busfreq)
	u_int busfreq;
{
	u_int hid1, pll;
	int m;

	/* Bus to Core Multiplier * 2 */
	static const char pll_mul2[] = {
		23, 34, 15, 30, 14, 36,  2, 40,
		 4, 42, 13, 26, 17, 48, 19, 18,
		 6, 21, 11, 22,  8, 20, 10, 24,
		16, 28, 12, 32, 27, 56,  0, 25
	};
	static const char pll_mul2_dfs2[] = {
		23, 34, 15, 30, 14, 36,  2, 40,
		 4, 42, 13, 26,  5, 48,  9, 18,
		 6, 21, 11, 22,  8, 20, 10, 24,
		16, 28, 12, 32,  7, 56,  0, 25
	};

	asm ("mfspr %0,HID1" : "=r"(hid1));

	/* XXX PC5 on 7458 */
	pll = hid1;
	pll &= HID1_PC0|HID1_PC1|HID1_PC2|HID1_PC3|HID1_PC4;
	pll >>= 12;

	m = pll_mul2[pll];

	/*
	 * The ratios 2.5:1, 3.5:1, and 4.5:1 share PLL configuration with
	 * 8.5:1, 13.5:1, and 9.5:1 respectively when DFS2 is set.
	 */
	if (hid1 & HID1_DFS2)
		m = pll_mul2_dfs2[pll];

	/* XXX DFS4 */

	return busfreq * m / 2;
}

void
mpc7450_set_dfs2(on)
	int on;
{
	u_int hid1;

	asm ("mfspr %0,HID1" : "=r"(hid1));
	if (on)
		hid1 |= HID1_DFS2;
	else
		hid1 &= ~HID1_DFS2;
	asm volatile ("sync; mtspr HID1,%0; sync; isync" :: "r"(hid1));
}

void
mpc7450_config_l3cr(name)
	const char *name;
{
	u_int l2cr, l3cr;
	char *size;
	static const char clktab[][4] = {
		"6", "?", "2", "2.5", "3", "3.5", "4", "5"
	};
	static const char ramtab[][11] = {
		"MSUG2 DDR", "late-write", "unknown", "PB2"
	};

	asm ("mfspr %0,L2CR" : "=r"(l2cr));
	asm ("mfspr %0,L3CR" : "=r"(l3cr));

	/* XXX Should we disable L2 cache first? */
	l2cr |= L2CR_L2PE;
	asm volatile ("mtspr L2CR,%0; sync; isync" :: "r"(l2cr));

	/*
	 * Configure L2 and L3 cache if not enabled.
	 */
	mpc7450_enable_L2cache();
	asm volatile ("mfspr %0,L2CR" : "=r"(l2cr));

	if (l3cr_config) {
		mpc7450_enable_L3cache(l3cr_config);
		asm volatile ("mfspr %0,L3CR" : "=r"(l3cr));
	}

	if (name == NULL)
		return;

	if (bootverbose) {
		char buf[128];
		u_int ictrl;

		snprintb(buf, sizeof buf, L2CR_BITMASK, l2cr);
		printf("%s: L2CR %s\n", name, buf);
		snprintb(buf, sizeof buf, L3CR_BITMASK, l3cr);
		printf("%s: L3CR %s\n", name, buf);

		asm ("mfspr %0,ICTRL" : "=r"(ictrl));
		snprintb(buf, sizeof buf, ICTRL_BITMASK, ictrl);
		printf("%s: ICTRL %s\n", name, buf);
	}

	switch (mfpvr() >> 16) {
	case MPC7448:
		size = "1MB";
		break;
	case MPC7447A:
	case MPC7457:
		size = "512KB";
		break;
	default:
		size = "256KB";
		break;
	}
	printf("%s: %s on-chip L2 cache", name, size);

	if ((l3cr & L3CR_L3E) == 0) {
		printf("\n");
		return;
	}

	printf(", %dMB %s:1 %s SRAM L3 cache\n",
	    (l3cr & L3CR_L3SIZ) ? 2 : 1,
	    clktab[(l3cr & L3CR_L3CLK) >> 23],
	    ramtab[(l3cr & L3CR_L3RT) >> 8]);
}

#ifdef AUTONAP
static void autonap(void *);
static struct callout autonap_ch;

void
cpu_autonap_init()
{
	switch (mfpvr() >> 16) {
	case MPC7450:
	case MPC7455:
	case MPC7457:
	case MPC7447A:
	case MPC7448:
		callout_init(&autonap_ch);
		callout_reset(&autonap_ch, 30*hz, autonap, NULL);
	}
}

void
autonap(v)
	void *v;
{
	u_int hid0;
	static int lastintr, timo;

	asm ("mfspr %0,HID0" : "=r"(hid0));

	if ((int)curcpu()->ci_data.cpu_nintr - lastintr - hz > 20) {
		/* HIGH-POWER mode */
		timo = 5;
		hid0 &= ~(HID0_NAP | HID0_SLEEP);
	} else {
		/* LOW-POWER (NAP) mode */

		/* ö H ˤʤääʤ褦ˤ */
		if (timo > 0)
			goto out;
		timo = 0;
		hid0 &= ~(HID0_NAP | HID0_SLEEP);
		hid0 |= HID0_NAP;
	}
	asm volatile ("mtspr HID0,%0" :: "r"(hid0));
out:
	timo--;
	lastintr = curcpu()->ci_data.cpu_nintr;
	callout_reset(&autonap_ch, hz, autonap, NULL);
}
#endif

#ifdef MULTIPROCESSOR

void
mpc7450_initialize(h)
	volatile struct cpu_hatch_data *h;
{
	int hid0;

	/* Enable timebase */
	asm volatile ("mfspr %0,HID0" : "=r"(hid0));
	hid0 |= HID0_TBEN;
	asm volatile ("mtspr HID0,%0" :: "r"(hid0));

	asm volatile ("mtspr HID1,%0" :: "r"(h->hid1));
	asm volatile ("mtspr MSSCR0,%0" :: "r"(h->msscr0));

	/* Enable the L1 caches. */
	mpc7450_enable_L1cache();

	/* Enable the L2 cache. */
	mpc7450_enable_L2cache();

	/* Enable the L3 cache. */
	if (h->l3cr & L3CR_L3E)
		mpc7450_enable_L3cache(h->l3cr);
}

#endif
