/*	$Id: cpu_750.c,v 1.19 2012/06/01 13:07:38 tsubai Exp $	*/

/*-
 * Copyright (c) 2002-2004 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "opt_l2cr_config.h"

#include <sys/param.h>
#include <sys/reboot.h>
#include <sys/systm.h>

#include <machine/cpufunc.h>
#include <machine/pio.h>
#include <machine/spr.h>

/* Hardware Implementation Dependent registers for the 750 */

#define HID0_EMCP   0x80000000  /* Enable MCP */
#define HID0_DBP    0x40000000  /* Disable 60x bus parity generation */
#define HID0_EBA    0x20000000  /* Enable 60x bus address parity checking */
#define HID0_EBD    0x10000000  /* Enable 60x bus data parity checking */
#define HID0_BCLK   0x08000000  /* CLK_OUT clock type selection */
#define HID0_ECLK   0x02000000  /* CLK_OUT clock type selection */
#define HID0_PAR    0x01000000  /* Disable precharge of ARTRY */
#define HID0_DOZE   0x00800000  /* Enable doze mode */
#define HID0_NAP    0x00400000  /* Enable nap mode */
#define HID0_SLEEP  0x00200000  /* Enable sleep mode */
#define HID0_DPM    0x00100000  /* Enable Dynamic power management */
#define HID0_NHR    0x00010000  /* Not hard reset */
#define HID0_ICE    0x00008000  /* Enable i-cache */
#define HID0_DCE    0x00004000  /* Enable d-cache */
#define HID0_ILOCK  0x00002000  /* i-cache lock */
#define HID0_DLOCK  0x00001000  /* d-cache lock */
#define HID0_ICFI   0x00000800  /* i-cache flush invalidate */
#define HID0_DCFI   0x00000400  /* d-cache flush invalidate */
#define HID0_SPD    0x00000200  /* Disable speculative cache access */
#define HID0_IFEM   0x00000100  /* Enable M bit for instruction fetches */
#define HID0_SGE    0x00000080  /* Enable store gathering */
#define HID0_DCFA   0x00000040  /* Data cache flush assist */
#define HID0_BTIC   0x00000020  /* Enable BTIC */
#define HID0_ABE    0x00000008  /* Address broadcast enable */
#define HID0_BHT    0x00000004  /* Enable branch history table */
#define HID0_NOOPTI 0x00000001  /* No-op the dcbt(st) */

#define HID0_BITMASK "\020" \
    "\040EMCP\037DBP\036EBA\035EBD\034BCLK\032ECLK\031PAR" \
    "\030DOZE\027NAP\026SLEEP\025DPM\021NHR" \
    "\020ICE\017DCE\016ILOCK\015DLOCK\014ICFI\013DCFI\012SPD\011IFEM" \
    "\010SGE\007DCFA\006BTIC\004ABE\003BHT\001NOOPTI"

#define HID1_PC0    0x80000000  /* PLL configuration bit 0 (RO) */
#define HID1_PC1    0x40000000  /* PLL configuration bit 1 (RO) */
#define HID1_PC2    0x20000000  /* PLL configuration bit 2 (RO) */
#define HID1_PC3    0x10000000  /* PLL configuration bit 3 (RO) */

#define HID1_BITMASK "\020" \
    "\040PC0\037PC1\036PC2\035PC3"

#define L2CR_L2E	0x80000000   /* 0: L2 enable */
#define L2CR_L2PE	0x40000000   /* 1: L2 data parity enable */
#define L2CR_L2SIZ	0x30000000   /* 2-3: L2 size */
#define   L2SIZ_2M	  0x00000000 /* (!750) */
#define   L2SIZ_256K	  0x10000000
#define   L2SIZ_512K	  0x20000000
#define   L2SIZ_1M	  0x30000000
#define L2CR_L2CLK	0x0e000000   /* 4-6: L2 clock ratio */
#define   L2CLK_DIS	  0x00000000 /* disable L2 clock */
#define   L2CLK_10	  0x02000000 /* core clock / 1   */
#define   L2CLK_15	  0x04000000 /*            / 1.5 */
#define   L2CLK_35	  0x06000000 /*            / 3.5 (!750) */
#define   L2CLK_20	  0x08000000 /*            / 2   */
#define   L2CLK_25	  0x0a000000 /*            / 2.5 */
#define   L2CLK_30	  0x0c000000 /*            / 3   */
#define   L2CLK_40	  0x0e000000 /*            / 4   (!750) */
#define L2CR_L2RAM	0x01800000   /* 7-8: L2 RAM type */
#define   L2RAM_FLOWTHRU  0x00000000 /* (!7410) */
#define   L2RAM_PB3	  0x00800000 /* (7410) */
#define   L2RAM_PB2	  0x01000000
#define   L2RAM_LATE	  0x01800000
#define L2CR_L2DO	0x00400000   /* 9: L2 data-only. */
#define L2CR_L2I	0x00200000   /* 10: L2 global invalidate. */
#define L2CR_L2CTL	0x00100000   /* 11: L2 RAM control (ZZ enable). */
#define L2CR_L2WT	0x00080000   /* 12: L2 write-through. */
#define L2CR_L2TS	0x00040000   /* 13: L2 test support. */
#define L2CR_L2OH	0x00030000   /* 14-15: L2 output hold. */
#define   L2OH_05	  0x00000000 /*     0.5ns (750) */
#define   L2OH_10	  0x00010000 /*     1.0ns (750) */
#define L2CR_L2SL	0x00008000   /* 16: L2 DLL slow. */
#define L2CR_L2DF	0x00004000   /* 17: L2 differential clock. */
#define L2CR_L2BYP	0x00002000   /* 18: L2 DLL bypass. */
#define L2CR_L2FA	0x00001000   /* 19: L2 flush assist. (!750) */
#define L2CR_L2HWF	0x00000800   /* 20: L2 hardware flush. (!750) */
#define L2CR_L2IO	0x00000400   /* 21: L2 instruction-only. (!750) */
#define L2CR_L2CLKSTP	0x00000200   /* 22: L2 clock stop. (!750) */
#define L2CR_DRO	0x00000100   /* 23: L2DLL rollover checkstop (7400) */
#define L2CR_L2IP	0x00000001   /* 31: L2 global invalidate in progress
					    (read only). */

#define L2CR_BITMASK "\020" \
    "\040E\037PE\027DO\026I\025CTL\024WT\023TS" \
    "\020SL\017DF\016BYP\015FA\014HWF\013IO\012CLKSTP\011DRO\001IP"

#if 0
/* 750CX: */
#define L2CR_L2E	0x80000000   /* 0: L2 enable */
#define L2CR_L2CE	0x40000000   /* 1: L2 checkstop enable */
#define L2CR_L2DO	0x00400000   /* 9: L2 data-only. */
#define L2CR_L2I	0x00200000   /* 10: L2 global invalidate. */
#define L2CR_L2WT	0x00080000   /* 12: L2 write-through. */
#define L2CR_L2TS	0x00040000   /* 13: L2 test support. */
#define L2CR_L2IP	0x00000001   /* 31: L2 global invalidate in progress */
#endif

void probe_l2cache(void);

#ifdef L2CR_CONFIG
u_int l2cr_config = L2CR_CONFIG;
#else
u_int l2cr_config = 0;
#endif

/*
 * Configure HID0 for 750.
 */
void
mpc750_config_hid()
{
	u_int hid0;

	asm ("mfspr %0,HID0" : "=r"(hid0));

	hid0 &= HID0_EBA | HID0_EBD | HID0_BCLK | HID0_ECLK | HID0_PAR |
	    HID0_ICE | HID0_DCE;
	hid0 |= HID0_EMCP | HID0_DOZE | HID0_DPM;
	hid0 |= HID0_SGE | HID0_BTIC | HID0_ABE | HID0_BHT;

	asm volatile ("mtspr HID0,%0" :: "r"(hid0));
}

void
mpc750_print_hid(name)
	const char *name;
{
	u_int hid0, hid1;
	char buf[160];
	static const char pllcfg[][4] = {
		"?", "7.5", "7", "1", "2", "6.5", "10",  "4.5",
		"3", "5.5", "4", "5", "8", "6",   "3.5", "0"
	};

	asm ("mfspr %0,HID0" : "=r"(hid0));
	asm ("mfspr %0,HID1" : "=r"(hid1));
	snprintb(buf, sizeof buf, HID0_BITMASK, hid0);
	printf("%s: HID0 %s\n", name, buf);

	if (bootverbose)
		printf("%s: %s:1 (core:bus) ratio\n", name, pllcfg[hid1 >> 28]);
}

void
mpc750_enable_nap()
{
	u_int hid0;

	asm ("mfspr %0,HID0" : "=r"(hid0));

	hid0 &= ~(HID0_DOZE | HID0_NAP | HID0_SLEEP);
	hid0 |= HID0_NAP;

	asm volatile ("mtspr HID0,%0" :: "r"(hid0));
}

void
mpc750_enable_sleep()
{
	u_int hid0;

	asm ("mfspr %0,HID0" : "=r"(hid0));

	hid0 &= ~(HID0_DOZE | HID0_NAP | HID0_SLEEP);
	hid0 |= HID0_SLEEP;

	asm volatile ("mtspr HID0,%0" :: "r"(hid0));
}

void
mpc750_disable_clkout()
{
	u_int hid0;

	asm ("mfspr %0,HID0" : "=r"(hid0));
	hid0 &= ~(HID0_BCLK | HID0_ECLK);
	asm volatile ("mtspr HID0,%0" :: "r"(hid0));
}

/*
 * Print L2 cache configuration on 750, 7400, and 7410.
 */
void
mpc750_config_l2cr(name)
	const char *name;
{
	u_int l2cr;
	static const char siztab[][5] = {	/* L2 size */
		"2M", "256K", "512K", "1M"
	};
	static const char clktab[][4] = {	/* L2 PLL */
		"0",   "1", "1.5", "3.5",
		"2", "2.5",   "3",   "4"
	};
	static const char ramtab[][13] = {	/* L2 RAM type */
		"flow-through", "PB3", "PB2", "late-write"
	};

	/*
	 * Configure on-chip L2 cache on PPC750CX/CXe.
	 */
	if (mfpvr() >> 16 == MPC750) {
		switch (mfpvr() & 0xfff0) {
		case 0x2200: /* CX  2.x */
		case 0x2210: /* CXe 2.x */
		case 0x3210: /* CXe 2.4 */
		case 0x2310: /* CXe 3.x */
		case 0x3310: /* CXe 3.1 */
			ppc750cx_config_l2cr(name);
			return;
		}
	}

	asm ("mfspr %0,L2CR" : "=r"(l2cr));

#ifndef L2CR_WRITETHRU
	if (l2cr & L2CR_L2WT) {
		l2cr &= ~L2CR_L2WT;
		asm volatile ("sync; isync; mtspr L2CR,%0" :: "r"(l2cr));
	}
#endif

	/*
	 * Configure L2 cache if not enabled.
	 */
	if ((l2cr & L2CR_L2E) == 0 && l2cr_config != 0) {
		probe_l2cache();
		asm volatile ("mfspr %0,L2CR" : "=r"(l2cr));
	}

	if (name == NULL)
		return;

	if ((l2cr & L2CR_L2E) == 0 || (l2cr & L2CR_L2CLK) == L2CLK_DIS) {
		printf("%s: L2 cache not enabled\n", name);
		return;
	}

	printf("%s: %sB, %s:1, %s SRAM", name,
	    siztab[(l2cr & L2CR_L2SIZ) >> 28],
	    clktab[(l2cr & L2CR_L2CLK) >> 25],
	    ramtab[(l2cr & L2CR_L2RAM) >> 23]);
	if (l2cr & L2CR_L2WT)
		printf(", write-thru");
	if (l2cr & L2CR_L2PE)
		printf(", with parity");

	printf(" backside L2 cache\n");

	if (bootverbose) {
		char buf[128];

		snprintb(buf, sizeof buf, L2CR_BITMASK, l2cr);
		printf("%s: L2CR %s, L2OH %d\n", name, buf,
		    (l2cr & L2CR_L2OH) >> 16);
	}
}

/*
 * Print L2 cache configuration on 750CX/CXe.
 */
void
ppc750cx_config_l2cr(name)
	const char *name;
{
	u_int l2cr, hid0, msr, x;

	asm ("mfspr %0,L2CR" : "=r"(l2cr));

	/*
	 * Configure L2 cache if not enabled.
	 */
	if ((l2cr & L2CR_L2E) == 0) {

		/* Disable interrupts and DPM. */
		msr = mfmsr();
		mtmsr(msr & ~PSL_EE);
		asm volatile ("mfspr %0,HID0" : "=r"(hid0));
		asm volatile ("mtspr HID0,%0" :: "r"(hid0 & ~HID0_DPM));

		asm volatile ("sync");

		/* L2 global invalidation. */
		asm volatile ("mtspr L2CR,%0" :: "r"(l2cr | L2CR_L2I));

		do {
			asm volatile ("mfspr %0,L2CR" : "=r"(x));
		} while (x & L2CR_L2IP);

		asm volatile ("mtspr L2CR,%0" :: "r"(l2cr));

#define L2CR_L2CE	0x40000000   /* 1: L2 checkstop enable */
		/* Set L2 parameters. */
		l2cr |= L2CR_L2CE;
		asm volatile ("mtspr L2CR,%0" :: "r"(l2cr));
		asm volatile ("sync; isync");

		/* Enable L2 cache. */
		l2cr |= L2CR_L2E;
		asm volatile ("mtspr L2CR,%0" :: "r"(l2cr));

		asm volatile ("isync");
		asm volatile ("mtspr HID0,%0" :: "r"(hid0));
		mtmsr(msr);
	}

	l2cr |= L2CR_L2CE;
	asm volatile ("sync; isync; mtspr L2CR,%0" :: "r"(l2cr));

	if (name == NULL)
		return;

	printf("%s: 256KB on-chip L2 cache\n", name);
	if (bootverbose) {
		char buf[128];

		snprintb(buf, sizeof buf, L2CR_BITMASK, l2cr);
		printf("%s: L2CR %s\n", name, buf);
	}
}

void
probe_l2cache()
{
	int i, x, size, maxsize;
	u_int msr, l2cr, hid0;
	const int vers = mfpvr() >> 16;
	volatile char *addr = (char *)0x80000000; /* XXX scratch address */

#ifdef DIAGNOSTIC
	asm ("mfspr %0,L2CR" : "=r"(l2cr));
	if (l2cr & L2CR_L2E)
		panic("L2 cache is already enabled");
#endif

	l2cr = l2cr_config & ~(L2CR_L2SIZ | L2CR_L2E);

	switch (vers) {
	case MPC750:
		maxsize = 1024*1024;
		l2cr |= L2SIZ_1M;
		break;
	case MPC7400:
	case MPC7410:
		maxsize = 2048*1024;
		l2cr |= L2SIZ_2M;
		break;
	default:
		return;
	}

	/* XXX Flush L1 cache. */
	for (i = 0; i < 1024*2*4; i++)
		in32(i * 32);

	/* Start L2 clock. */
	asm volatile ("mtspr L2CR,%0" :: "r"(l2cr));

	/* Wait for L2 clock to be stable (640 L2 clocks). */
	delay(1000);

	/* Invalidate all L2 contents. */
	/* XXX isync is needed because of silicon bug. */
	asm volatile ("mtspr L2CR,%0; isync" :: "r"(l2cr | L2CR_L2I));
	do {
		asm volatile ("mfspr %0,L2CR" : "=r"(x));
	} while (x & L2CR_L2IP);

	/*
	 * Probe L2 cache size.
	 */
	/* Disable data address translation. */
	asm volatile ("sync; isync");
	msr = mfmsr();
	mtmsr(msr & ~PSL_DR);

	/* Enable L2 test mode. */
	asm volatile ("mtspr L2CR,%0"
	    :: "r"(l2cr | L2CR_L2E | L2CR_L2DO | L2CR_L2TS));
	asm volatile ("sync; isync");

	/* Zero-fill the entire L2 cache. */
	for (i = 0; i < maxsize; i += CACHELINESIZE)
		asm volatile ("dcbz 0,%0; dcbf 0,%0" :: "r"(addr + i));

	/* Invalidate and lock the L1 d-cache. */
	asm ("mfspr %0,HID0" : "=r"(hid0));
#if 1 /* XXX */
	asm volatile ("mtspr HID0,%0" :: "r"(hid0 | HID0_DLOCK));
#else
	asm volatile ("mtspr HID0,%0" :: "r"(hid0 | HID0_DLOCK | HID0_DCFI));
	delay(1000);
#endif

	/* Test cache RAM. */
	size = 0;
	for (i = 0; i < maxsize; i += CACHELINESIZE) {
		if (addr[i] == 0)
			size += CACHELINESIZE;
		addr[i] = 0xff;
	}
	asm volatile ("sync");

	/* Restore original state. */
	asm volatile ("mtspr L2CR,%0" :: "r"(l2cr));
	asm volatile ("mtspr HID0,%0" :: "r"(hid0));
	mtmsr(msr);

	/*
	 * Initialize L2 cache.
	 */
	l2cr &= ~L2CR_L2SIZ;

	switch (size) {
	case 256*1024:
		l2cr |= L2SIZ_256K; break;
	case 512*1024:
		l2cr |= L2SIZ_512K; break;
	case 1024*1024:
		l2cr |= L2SIZ_1M; break;
	case 2048*1024:
		l2cr |= L2SIZ_2M; break;
	default:
		return;
	}

	mpc750_enable_L2cache(l2cr);
}

void
mpc750_enable_L2cache(l2cr)
	u_int l2cr;
{
	int i;
	u_int x;

	l2cr &= ~L2CR_L2E;

	/* XXX Flush L1 cache. */
	for (i = 0; i < 1024*2*4; i++)
		in32(i * 32);

	/* Start L2 clock. */
	asm volatile ("mtspr L2CR,%0" :: "r"(l2cr));

	/* Wait for L2 clock to be stable (640 L2 clocks). */
	delay(1000);

	/* Invalidate the entire L2 cache. */
	/* XXX isync is needed because of silicon bug. */
	asm volatile ("mtspr L2CR,%0; isync" :: "r"(l2cr | L2CR_L2I));

	/* Wait for the invalidation done. */
	do {
		asm volatile ("mfspr %0,L2CR" : "=r"(x));
	} while (x & L2CR_L2IP);

	l2cr |= L2CR_L2E;

	/* Enable L2 cache. */
	asm volatile ("mtspr L2CR,%0" :: "r"(l2cr));
}

u_int
mpc750_corefreq(busfreq)
	u_int busfreq;
{
	u_int hid1, pll;

	/* Bus to Core Multiplier * 2 */
	static const char pll_mul2[] = {
		 0, 15, 14,  2,  4, 13, 20,  9,
		 6, 11,  8, 10, 16, 12,  7,  0
	};

	asm ("mfspr %0,HID1" : "=r"(hid1));
	pll = hid1 >> 28;

	return busfreq * pll_mul2[pll] / 2;
}

/*
 * PLL_CFG configuration
 * [0:3]
 * 0000  0  Reserved
 * 0001  1  7.5x
 * 0010  2  7x
 * 0011  3  PLL Bypass
 * 0100  4  2x   (750L)
 * 0101  5  6.5x
 * 0110  6  10x  (750L, 750CX)
 * 0111  7  4.5x
 * 1000  8  3x
 * 1001  9  5.5x
 * 1010 10  4x
 * 1011 11  5x
 * 1100 12  8x
 * 1101 13  6x
 * 1110 14  3.5x
 * 1111 15  Off
 */
