/*	$Id: cpu_common.c,v 1.50 2008/09/17 10:06:20 tsubai Exp $	*/

/*-
 * Copyright (c) 2002-2005 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "opt_multiprocessor.h"

#include <sys/param.h>
#include <sys/systm.h>

#include <uvm/uvm_extern.h>

#include <machine/bat.h>
#include <machine/cpufunc.h>
#include <machine/powerpc.h>
#include <machine/spr.h>

struct cpu_vector {
	void (*config_hid)(void);
	void (*print_hid)(const char *);
	void (*config_cache)(const char *);
	void (*enable_nap)(void);
	void (*enable_sleep)(void);
	u_int (*corefreq)(u_int);
	u_int flags;
};
static struct cpu_vector *ppc_vector;

static struct cpu_vector e300 = {
	e300_config_hid,
	e300_print_hid,
	NULL,
	NULL,
	NULL,
	NULL,
	0
};

static struct cpu_vector mpc603 = {
	mpc603_config_hid,
	mpc60x_print_hid,
	NULL,
	NULL,
	NULL,
	mpc603_corefreq,
	0
};

static struct cpu_vector mpc604 = {
	NULL,			/* XXX Do nothing for now. */
	mpc60x_print_hid,
	NULL,
	NULL,
	NULL,
	mpc604_corefreq,
	0
};

static struct cpu_vector mpc750 = {
	mpc750_config_hid,
	mpc750_print_hid,
	mpc750_config_l2cr,
	mpc750_enable_nap,
	mpc750_enable_sleep,
	mpc750_corefreq,
	0
};

static struct cpu_vector ppc750fx = {
	ppc750fx_config_hid,
	ppc750fx_print_hid,
	ppc750fx_config_l2cr,
	mpc750_enable_nap,
	mpc750_enable_sleep,
	ppc750fx_corefreq,
	0
};

static struct cpu_vector mpc7400 = {
	mpc7400_config_hid,
	mpc7400_print_hid,
	mpc750_config_l2cr,
	mpc750_enable_nap,
	mpc750_enable_sleep,
	mpc7400_corefreq,
	0
};

static struct cpu_vector mpc7450 = {
	mpc7450_config_hid,
	mpc7450_print_hid,
	mpc7450_config_l3cr,
	mpc7450_enable_nap,
	mpc7450_enable_sleep,
	mpc7450_corefreq,
	0
};

void
init_cpu_vector()
{
	switch (mfpvr() >> 16) {

	case E300C1:
	case E300C2:
	case E300C3:
	case E300C4:
		ppc_vector = &e300;
		break;

	case MPC603:
	case MPC603e:
	case MPC603ev:
		ppc_vector = &mpc603;
		break;

	case MPC604:
	case MPC604e:
	case MPC604r:
		ppc_vector = &mpc604;
		break;

	case MPC750:
		/* XXX */
		if (ppc750fx_probe())
			ppc_vector = &ppc750fx;
		else
			ppc_vector = &mpc750;
		break;

	case PPC750FX:
	case PPC750GX:
		ppc_vector = &ppc750fx;
		break;

	case MPC7400:
	case MPC7410:
		ppc_vector = &mpc7400;
		break;

	case MPC7450:
	case MPC7455:
	case MPC7457:
	case MPC7447A:
	case MPC7448:
	default:	/* XXX Catch unknown CPU here. */
		ppc_vector = &mpc7450;
		break;
	}
}

void
powerpc_config_hid()
{
	if (ppc_vector->config_hid)
		(*ppc_vector->config_hid)();
}

void
powerpc_print_hid(name)
	char *name;
{
	if (ppc_vector->print_hid)
		(*ppc_vector->print_hid)(name);
}

void
powerpc_enable_nap()
{
	if (ppc_vector->enable_nap)
		(*ppc_vector->enable_nap)();
}

void
powerpc_enable_sleep()
{
	if (ppc_vector->enable_sleep)
		(*ppc_vector->enable_sleep)();
}

/*
 * Configure L2 and L3 caches.
 */
void
powerpc_config_cache(name)
	char *name;
{
	if (ppc_vector->config_cache)
		(*ppc_vector->config_cache)(name);
}

u_int
powerpc_corefreq(busfreq)
	u_int busfreq;
{
	if (ppc_vector->corefreq)
		return (*ppc_vector->corefreq)(busfreq);
	else
		return 0;
}

void
loadpir(id)
	int id;
{
	switch (mfpvr() >> 16) {
	case E300C1:
	case E300C2:
	case E300C3:
	case E300C4:
	case MPC603:
	case MPC603e:
	case MPC603ev:
	case MPC750:
	case PPC750FX:
	case PPC750GX:
		break;
	default:
		asm volatile ("mtspr PIR,%0" :: "r"(id));
	}
}

void
initbat()
{
	struct mem_region *allmem, *availmem, *mp;

	/*
	 * Initialize BAT registers to unmapped to not generate
	 * overlapping mappings below.
	 */
	asm volatile ("mtibatu 0,%0" :: "r"(0));
	asm volatile ("mtibatu 1,%0" :: "r"(0));
	asm volatile ("mtibatu 2,%0" :: "r"(0));
	asm volatile ("mtibatu 3,%0" :: "r"(0));
	asm volatile ("mtdbatu 0,%0" :: "r"(0));
	asm volatile ("mtdbatu 1,%0" :: "r"(0));
	asm volatile ("mtdbatu 2,%0" :: "r"(0));
	asm volatile ("mtdbatu 3,%0" :: "r"(0));
	switch (mfpvr() >> 16) {
	case E300C1:
	case E300C2:
	case E300C3:
	case E300C4:
		e300_clearbat();
		break;
	case PPC750FX:
	case PPC750GX:
		ppc750fx_clearbat();
		break;
	}

	/*
	 * Set up initial BAT table to only map the lowest 256 MB area.
	 */
	battable[0x0].batl = BATL(0x00000000, BATL_M);
	battable[0x0].batu = BATU(0x00000000, BATU_256M);

	/*
	 * Now setup fixed bat registers
	 *
	 * Note that we still run in real mode, and the BAT
	 * registers were cleared above.
	 */
	/* BAT0 used for initial 256 MB segment */
	asm volatile ("mtibatl 0,%0; mtibatu 0,%1;"
		      "mtdbatl 0,%0; mtdbatu 0,%1;"
		      :: "r"(battable[0].batl), "r"(battable[0].batu));

	/*
	 * Set up battable to map all RAM regions.
	 * XXX mem_regions() call may need BAT0 set up.
	 */
	mem_regions(&allmem, &availmem);

	switch (mfpvr() >> 16) {

	default:
		for (mp = allmem; mp->size; mp++) {
			paddr_t pa = mp->start & 0xf0000000;
			paddr_t end = mp->start + mp->size;

			do {
				u_int n = pa >> 28;

				battable[n].batl = BATL(pa, BATL_M);
				battable[n].batu = BATU(pa, BATU_256M);
				pa += 0x10000000;
			} while (pa < end);
		}

		/* BAT1 used for secondary 256 MB segment (if exists) */
		asm volatile ("mtibatl 1,%0; mtibatu 1,%1;"
			      "mtdbatl 1,%0; mtdbatu 1,%1;"
		    :: "r"(battable[1].batl), "r"(battable[1].batu));
		break;

	case MPC7455:
	case MPC7457:
	case MPC7447A:
	case MPC7448:
		/*
		 * Use extended BAT size on 7455.
		 */
		for (mp = allmem; mp->size; mp++) {
			paddr_t pa = mp->start & 0xc0000000;
			paddr_t end = mp->start + mp->size;

			do {
				u_int n = pa >> 28;

				battable[n+0].batl = BATL(pa, BATL_M);
				battable[n+0].batu = BATU(pa, BATU_1G);
				battable[n+1].batl = BATL(pa, BATL_M);
				battable[n+1].batu = BATU(pa, BATU_1G);
				battable[n+2].batl = BATL(pa, BATL_M);
				battable[n+2].batu = BATU(pa, BATU_1G);
				battable[n+3].batl = BATL(pa, BATL_M);
				battable[n+3].batu = BATU(pa, BATU_1G);
				pa += 0x40000000;
			} while (pa < end);
		}

		mpc7455_xbsen();
		asm volatile ("mtibatl 0,%0; mtibatu 0,%1;"
			      "mtdbatl 0,%0; mtdbatu 0,%1;"
		    :: "r"(battable[0].batl), "r"(battable[0].batu));
		asm volatile ("mtibatl 1,%0; mtibatu 1,%1;"
			      "mtdbatl 1,%0; mtdbatu 1,%1;"
		    :: "r"(battable[4].batl), "r"(battable[4].batu));
		break;
	}
}

extern u_int32_t idle_hook[];
extern char idle_powersave[], idle_powersave_altivec[];

void
install_idle_hook()
{
	u_int vers;
	u_int nop = 0x60000000;

	vers = mfpvr() >> 16;

	switch (vers) {
	case MPC603:
	case MPC603e:
	case MPC603ev:
	case MPC750:
	case PPC750FX:
	case PPC750GX:
	case E300C1:
	case E300C2:
	case E300C3:
	case E300C4:
		idle_hook[0] = nop;		/* blr    */
		idle_hook[1] = nop;		/* dssall */
		idle_hook[2] = nop;		/* isync  */
		__syncicache(idle_hook, 12);
		break;

	case MPC7400:
	case MPC7410:
	case MPC7450:
	case MPC7455:
	case MPC7457:
	case MPC7447A:
	case MPC7448:
		idle_hook[0] = nop;		/* blr    */
		__syncicache(idle_hook, 4);
		break;
	}
}

#ifdef MULTIPROCESSOR

void
powerpc_initialize(h)
	volatile struct cpu_hatch_data *h;
{
	int i;

	/* Initialize timebase. */
	asm ("mttbl %0; mttbu %0; mttbl %0" :: "r"(0));

	switch (mfpvr() >> 16) {

	case MPC604:
	case MPC604e:
	case MPC604r:
		/* Enable the L1 cache. */
		mpc60x_enable_L1cache();
		break;

	case MPC7450:
	case MPC7455:
	case MPC7457:
	case MPC7447A:
	case MPC7448:
		/* Enable caches, timebase, ... */
		mpc7450_initialize(h);
		break;
	}

	/* Initialize MMU. */
	asm volatile ("mtibatu 0,%0" :: "r"(0));
	asm volatile ("mtibatu 1,%0" :: "r"(0));
	asm volatile ("mtibatu 2,%0" :: "r"(0));
	asm volatile ("mtibatu 3,%0" :: "r"(0));
	asm volatile ("mtdbatu 0,%0" :: "r"(0));
	asm volatile ("mtdbatu 1,%0" :: "r"(0));
	asm volatile ("mtdbatu 2,%0" :: "r"(0));
	asm volatile ("mtdbatu 3,%0" :: "r"(0));

	asm volatile ("mtspr HID0,%0" :: "r"(h->hid0));

	asm volatile ("mtibatl 0,%0; mtibatu 0,%1;"
		      "mtdbatl 0,%0; mtdbatu 0,%1;"
		      :: "r"(battable[0].batl), "r"(battable[0].batu));

	if ((battable[0].batu & BATU_BL) == BATU_1G) {
		/* Use extended BAT size on 745[57]. */
		asm volatile ("mtibatl 1,%0; mtibatu 1,%1;"
			      "mtdbatl 1,%0; mtdbatu 1,%1;"
		    :: "r"(battable[4].batl), "r"(battable[4].batu));
	} else {
		asm volatile ("mtibatl 1,%0; mtibatu 1,%1;"
			      "mtdbatl 1,%0; mtdbatu 1,%1;"
		    :: "r"(battable[1].batl), "r"(battable[1].batu));
	}

	for (i = 0; i < 16; i++)
		asm volatile ("mtsrin %0,%1"
		    :: "r"(h->sr[i]), "r"(i << ADDR_SR_SHFT));
	asm volatile ("mtsdr1 %0" :: "r"(h->sdr1));

	asm volatile ("isync");
}

#endif
