/*	$Id: cpu_e300.c,v 1.9 2009/02/07 05:52:28 tsubai Exp $	*/

/*-
 * Copyright (c) 2008 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/systm.h>

#include <machine/cpufunc.h>
#include <machine/spr.h>

/* Hardware Implementation Dependent registers for the e300 */

#define HID0_EMCP   0x80000000  /* Enable MCP */
#define HID0_ECPE   0x40000000  /* Enable cache parity errors */
#define HID0_EBA    0x20000000  /* Enable 60x bus address parity checking */
#define HID0_EBD    0x10000000  /* Enable 60x bus data parity checking */
#define HID0_SBCLK  0x08000000  /* Select bus clock for test clock pin */
#define HID0_ECLK   0x02000000  /* CLK_OUT clock type selection */
#define HID0_PAR    0x01000000  /* Disable precharge of ARTRY */
#define HID0_DOZE   0x00800000  /* Enable doze mode */
#define HID0_NAP    0x00400000  /* Enable nap mode */
#define HID0_SLEEP  0x00200000  /* Enable sleep mode */
#define HID0_DPM    0x00100000  /* Enable dynamic power management */
#define HID0_ICE    0x00008000  /* Enable i-cache */
#define HID0_DCE    0x00004000  /* Enable d-cache */
#define HID0_ILOCK  0x00002000  /* i-cache lock */
#define HID0_DLOCK  0x00001000  /* d-cache lock */
#define HID0_ICFI   0x00000800  /* i-cache flush invalidate */
#define HID0_DCFI   0x00000400  /* d-cache flush invalidate */
#define HID0_IFEM   0x00000080  /* Enable M-bit for I-fetch */
#define HID0_DECAR  0x00000040  /* Decrerenter auto reload (!e300c1) */
#define HID0_FBIOB  0x00000010  /* Force branch indirect on bus */
#define HID0_ABE    0x00000008  /* Address broadcast enable */
#define HID0_NOOPTI 0x00000001  /* No-op the dcbt(st) */

#define HID0_BITMASK "\020" \
    "\040EMCP\037ECPE\036EBA\035EBD\034SBCLK\032ECLK\031PAR" \
    "\030DOZE\027NAP\026SLEEP\025DPM" \
    "\020ICE\017DCE\016ILOCK\015DLOCK\014ICFI\013DCFI" \
    "\010IFEM\007DECAR\005FBIOB\004ABE\001NOOPTI"

#define HID1_BITMASK "\020" \
    "\040PC0\037PC1\036PC2\035PC3\034PC4\033PC5\032PC6"

#define HID2_LET    0x08000000	/* True little-endian */
#define HID2_IFEB   0x04000000	/* Instruction fetch burst extension */
#define HID2_MESI   0x01000000	/* MESI state enable */
#define HID2_IFEC   0x00800000	/* Instruction fetch cancel extension */
#define HID2_EBQS   0x00400000	/* Enable BIU queue sharing */
#define HID2_EBPX   0x00200000	/* Enable BIU pipeline extension */
#define HID2_ELRW   0x00100000	/* Enable weighted LRU (!e300c1) */
#define HID2_NOKS   0x00080000	/* No kill for snoop (!e300c1) */
#define HID2_HBE    0x00040000	/* High BAT enable */
#define HID2_IWLCK  0x0000e000	/* Instruction cache way-lock */
#define HID2_ICWP   0x00001000	/* Instruction cache way protection */
#define HID2_DWLCK  0x000000e0	/* Data cache way-lock */

#define HID2_BITMASK "\020" \
    "\034LET\033IFEB\031MESI\030IFEC\027EBQS\026EBPX\025ELRW\024NOKS\023HBE" \
    "\015ICWP"

asm ("IBAT4U = 560");
asm ("IBAT5U = 562");
asm ("IBAT6U = 564");
asm ("IBAT7U = 566");
asm ("DBAT4U = 568");
asm ("DBAT5U = 570");
asm ("DBAT6U = 572");
asm ("DBAT7U = 574");

void
e300_config_hid()
{
	u_int hid0, hid2;

	asm ("mfspr %0,HID0" : "=r"(hid0));
	asm ("mfspr %0,HID2" : "=r"(hid2));

	/* XXX for now */
	hid0 &= ~(HID0_DOZE | HID0_NAP | HID0_SLEEP);
	hid0 |= HID0_DOZE | HID0_DPM;
	hid0 |= HID0_EMCP;

	hid2 |= HID2_IFEC;
	hid2 |= HID2_EBQS;
	hid2 |= HID2_EBPX;
	hid2 |= HID2_HBE;

	asm volatile ("mtspr HID0,%0" :: "r"(hid0));
	asm volatile ("mtspr HID2,%0" :: "r"(hid2));
}

void
e300_print_hid(name)
	const char *name;
{
	u_int hid0, hid1, hid2;
	char buf[160];

	asm ("mfspr %0,HID0" : "=r"(hid0));
	snprintb(buf, sizeof buf, HID0_BITMASK, hid0);
	printf("%s: HID0 %s\n", name, buf);

	asm ("mfspr %0,HID1" : "=r"(hid1));
	snprintb(buf, sizeof buf, HID1_BITMASK, hid1);
	printf("%s: HID1 %s\n", name, buf);

	asm ("mfspr %0,HID2" : "=r"(hid2));
	snprintb(buf, sizeof buf, HID2_BITMASK, hid2);
	printf("%s: HID2 %s\n", name, buf);
}

void
e300_enable_L1cache()
{
	u_int hid0;

	/* Enable L1 caches. */
	asm volatile ("mfspr %0,HID0" : "=r"(hid0));
	hid0 &= ~HID0_ILOCK;
	hid0 &= ~HID0_ICFI;
	asm volatile ("isync");
	asm volatile ("mtspr HID0,%0" :: "r"(hid0 | HID0_ICFI));
	asm volatile ("mtspr HID0,%0" :: "r"(hid0));
	asm volatile ("mtspr HID0,%0" :: "r"(hid0 | HID0_ICE));
	asm volatile ("isync");

	asm volatile ("mfspr %0,HID0" : "=r"(hid0));
	hid0 &= ~HID0_DLOCK;
	hid0 &= ~HID0_DCFI;
	asm volatile ("sync");
	asm volatile ("mtspr HID0,%0" :: "r"(hid0 | HID0_DCFI));
	asm volatile ("mtspr HID0,%0" :: "r"(hid0));
	asm volatile ("mtspr HID0,%0" :: "r"(hid0 | HID0_DCE));
}

/*
 * Invalidate bat0 - bat7.
 */
void
e300_clearbat()
{

	asm volatile ("mtibatu 0,%0" :: "r"(0));
	asm volatile ("mtibatu 1,%0" :: "r"(0));
	asm volatile ("mtibatu 2,%0" :: "r"(0));
	asm volatile ("mtibatu 3,%0" :: "r"(0));
#if 0 /* XXX Avoid assembler error. */
	asm volatile ("mtibatu 4,%0" :: "r"(0));
	asm volatile ("mtibatu 5,%0" :: "r"(0));
	asm volatile ("mtibatu 6,%0" :: "r"(0));
	asm volatile ("mtibatu 7,%0" :: "r"(0));
#else
	asm volatile ("mtspr IBAT4U,%0" :: "r"(0));
	asm volatile ("mtspr IBAT5U,%0" :: "r"(0));
	asm volatile ("mtspr IBAT6U,%0" :: "r"(0));
	asm volatile ("mtspr IBAT7U,%0" :: "r"(0));
#endif

	asm volatile ("mtdbatu 0,%0" :: "r"(0));
	asm volatile ("mtdbatu 1,%0" :: "r"(0));
	asm volatile ("mtdbatu 2,%0" :: "r"(0));
	asm volatile ("mtdbatu 3,%0" :: "r"(0));
#if 0 /* XXX Avoid assembler error. */
	asm volatile ("mtdbatu 4,%0" :: "r"(0));
	asm volatile ("mtdbatu 5,%0" :: "r"(0));
	asm volatile ("mtdbatu 6,%0" :: "r"(0));
	asm volatile ("mtdbatu 7,%0" :: "r"(0));
#else
	asm volatile ("mtspr DBAT4U,%0" :: "r"(0));
	asm volatile ("mtspr DBAT5U,%0" :: "r"(0));
	asm volatile ("mtspr DBAT6U,%0" :: "r"(0));
	asm volatile ("mtspr DBAT7U,%0" :: "r"(0));
#endif
}
