/*	$Id: powerpc_machdep.c,v 1.42 2012/06/06 13:57:11 tsubai Exp $	*/

/*
 * Copyright (C) 1995, 1996 Wolfgang Solfrank.
 * Copyright (C) 1995, 1996 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "opt_altivec.h"
#include "opt_compat_netbsd.h"

#include <sys/param.h>
#include <sys/conf.h>
#include <sys/cpu.h>
#include <sys/disklabel.h>
#include <sys/exec.h>
#include <sys/mount.h>
#include <sys/proc.h>
#include <sys/syscallargs.h>
#include <sys/systm.h>
#include <uvm/uvm_extern.h>
#ifdef ALTIVEC
#include <machine/altivec.h>
#endif
#include <machine/fpu.h>
#include <machine/pcb.h>

/*
 * Set up registers on exec.
 */
void
setregs(l, pack, stack)
	struct lwp *l;
	struct exec_package *pack;
	u_long stack;
{
	struct proc *p = l->l_proc;
	struct trapframe *tf = trapframe(l);
	struct ps_strings arginfo;

	bzero(tf, sizeof *tf);
	tf->fixreg[1] = -roundup(-stack + 8, 16);

	/*
	 * XXX Machine-independent code has already copied arguments and
	 * XXX environment to userland.  Get them back here.
	 */
	copyin_psstrings(p, &arginfo);

	/*
	 * Set up arguments for _start():
	 *	_start(argc, argv, envp, obj, cleanup, ps_strings);
	 *
	 * Notes:
	 *	- obj and cleanup are the auxiliary and termination
	 *	  vectors.  They are fixed up by ld.elf_so.
	 *	- ps_strings is a NetBSD extension, and will be
	 *	  ignored by executables which are strictly
	 *	  compliant with the SVR4 ABI.
	 *
	 * XXX We have to set both regs and retval here due to different
	 * XXX calling convention in trap.c and init_main.c.
	 */
	tf->fixreg[3] = arginfo.ps_nargvstr;
	tf->fixreg[4] = (register_t)arginfo.ps_argvstr;
	tf->fixreg[5] = (register_t)arginfo.ps_envstr;
	tf->fixreg[6] = 0;			/* auxiliary vector */
	tf->fixreg[7] = 0;			/* termination vector */
	tf->fixreg[8] = p->p_psstrp;		/* NetBSD extension */

	tf->srr0 = pack->ep_entry;
	tf->srr1 = PSL_MBO | PSL_USERSET | PSL_FE_DFLT;
	((struct pcb *)l->l_addr)->pcb_flags = 0;
}

struct sigframe {
	register_t sf_sp;
	register_t sf_lr;
	siginfo_t sf_si;
	ucontext_t sf_uc;
};

/*
 * Send a signal to process.
 */
void
sendsig_siginfo(ksi, mask)
	const ksiginfo_t *ksi;
	const sigset_t *mask;
{
	int sig = ksi->ksi_signo;
	struct lwp *l = curlwp;
	struct proc *p = l->l_proc;
	struct sigacts *ps = p->p_sigacts;
	struct trapframe *tf;
	struct sigframe *fp, frame;
	int onstack, error;
	sig_t catcher = SIGACTION(p, sig).sa_handler;

	tf = trapframe(l);

	/* Do we need to jump onto the signal stack? */
	onstack =
	    (l->l_sigstk.ss_flags & (SS_DISABLE | SS_ONSTACK)) == 0 &&
	    (SIGACTION(p, sig).sa_flags & SA_ONSTACK) != 0;

	/* Allocate space for the signal handler context. */
	if (onstack)
		fp = (struct sigframe *)((caddr_t)l->l_sigstk.ss_sp +
						l->l_sigstk.ss_size);
	else
		fp = (struct sigframe *)tf->fixreg[1];
	fp = (struct sigframe *)((long)(fp - 1) & ~0xf);

	frame.sf_si._info = ksi->ksi_info;
	frame.sf_uc.uc_flags = _UC_SIGMASK;
#ifdef _UC_SETSTACK
	frame.sf_uc.uc_flags |= (l->l_sigstk.ss_flags & SS_ONSTACK)
	    ? _UC_SETSTACK : _UC_CLRSTACK;
#endif
	frame.sf_uc.uc_sigmask = *mask;
	frame.sf_uc.uc_link = l->l_ctxlink;
	bzero(&frame.sf_uc.uc_stack, sizeof frame.sf_uc.uc_stack);
	cpu_getmcontext(l, &frame.sf_uc.uc_mcontext, &frame.sf_uc.uc_flags);

	sendsig_reset(l, sig);

	mutex_exit(p->p_lock);
	error = copyout(&frame, fp, sizeof frame);
	mutex_enter(p->p_lock);

	if (error != 0) {
		/*
		 * Process has trashed its stack; give it an illegal
		 * instructoin to halt it in its tracks.
		 */
		sigexit(l, SIGILL);
		/* NOTREACHED */
	}

	/*
	 * Build context to run handler in.  Note the trampoline version
	 * numbers are coordinated with machine-dependent code in libc.
	 */
	switch (ps->sa_sigdesc[sig].sd_vers) {
	case 2:
		tf->fixreg[1] = (register_t)fp;
		tf->srr0 = (register_t)catcher;
		tf->fixreg[3] = (register_t)sig;
		tf->fixreg[4] = (register_t)&fp->sf_si;
		tf->fixreg[5] = (register_t)&fp->sf_uc;
		tf->fixreg[31] = tf->fixreg[5];
		tf->lr = (register_t)ps->sa_sigdesc[sig].sd_tramp;
		break;

	default:
		/* Don't know what trampoline version; kill it. */
		sigexit(l, SIGILL);
	}

	/* Remember that we're now on the signal stack. */
	if (onstack)
		l->l_sigstk.ss_flags |= SS_ONSTACK;
}

void
cpu_getmcontext(l, mcp, flagp)
	struct lwp *l;
	mcontext_t *mcp;
	unsigned int *flagp;
{
	const struct trapframe *tf = trapframe(l);
	__gregset_t *gr = &mcp->__gregs;
	struct pcb *pcb = l->l_addr;

	/* Save GPR context. */
	memcpy(gr->__r, tf->fixreg, sizeof gr->__r);	/* r0-31 */
	gr->__r_cr  = tf->cr;
	gr->__r_lr  = tf->lr;
	gr->__r_pc  = tf->srr0;
	gr->__r_msr = tf->srr1;				/* XXX */
	gr->__r_ctr = tf->ctr;
	gr->__r_xer = tf->xer;
	gr->__r_mq  = 0;				/* For now. */
	*flagp |= _UC_CPU;

	/* Save FPR context, if any. */
	if (pcb->pcb_flags & PCB_FPU) {
#ifdef PPC_HAVE_FPU
		/* If the FPU has valid data, dump its context to the PCB. */
		if (fpulwp)
			save_fpu(fpulwp);
#ifdef MULTIPROCESSOR
		if (pcb->pcb_fpcpu)
			save_fpu_lwp(l);
#endif
#endif
		memcpy(mcp->__fpregs.__fpu_regs, pcb->pcb_fpu.fpreg,
		    sizeof mcp->__fpregs.__fpu_regs);
		mcp->__fpregs.__fpu_fpscr = *(u_int64_t *)&pcb->pcb_fpu.fpscr;
		mcp->__fpregs.__fpu_valid = 1;
		*flagp |= _UC_FPU;
	} else
		mcp->__fpregs.__fpu_valid = 0;

#ifdef ALTIVEC
	/* Save VEC context, if any. */
	if (pcb->pcb_flags & PCB_VEC) {
		/* If the vector unit has valid data, dump its context to the PCB first. */
		if (veclwp)
			save_vec(veclwp);
#ifdef MULTIPROCESSOR
		if (pcb->pcb_veccpu)
			save_vec_lwp(l);
#endif
		memcpy(mcp->__vrf.__vr, pcb->pcb_vec.vreg,
		    sizeof mcp->__vrf.__vr);
		mcp->__vrf.__vscr = pcb->pcb_vec.vscr[3];
		mcp->__vrf.__vrsave = pcb->pcb_vec.vrsave;
		*flagp |= _UC_POWERPC_VEC;
	} else
#endif
		mcp->__vrf.__vrsave = 0;
}

int
cpu_setmcontext(l, mcp, flags)
	struct lwp *l;
	const mcontext_t *mcp;
	unsigned int flags;
{
	struct trapframe *tf = trapframe(l);
	const __gregset_t *gr = &mcp->__gregs;
	struct pcb *pcb = l->l_addr;

	/* Restore GPR context, if any. */
	if (flags & _UC_CPU) {
		memcpy(&tf->fixreg, gr, 32 * 4);
		tf->cr   = gr->__r_cr;
		tf->lr   = gr->__r_lr;
		tf->srr0 = gr->__r_pc;
		tf->srr1 &= ~(PSL_FE0|PSL_FE1);
		tf->srr1 |= gr->__r_msr & (PSL_FE0|PSL_FE1);
		tf->ctr  = gr->__r_ctr;
		tf->xer  = gr->__r_xer;
		/* unused = gr->__r_mq; */
	}

	/* Restore FPR context, if any. */
	if ((flags & _UC_FPU) && mcp->__fpregs.__fpu_valid) {
#ifdef PPC_HAVE_FPU
		/* XXX we don't need to save the state, just to drop it */
		if (fpulwp)
			save_fpu(fpulwp);
#ifdef MULTIPROCESSOR
		if (pcb->pcb_fpcpu)
			save_fpu_lwp(l);
#endif
#endif
		memcpy(pcb->pcb_fpu.fpreg, mcp->__fpregs.__fpu_regs,
		    sizeof pcb->pcb_fpu.fpreg);
		*(u_int64_t *)&pcb->pcb_fpu.fpscr = mcp->__fpregs.__fpu_fpscr;
		pcb->pcb_flags |= PCB_FPU;
		tf->srr1 &= ~PSL_FP;
	}

#ifdef ALTIVEC
	/* Restore VEC context, if any. */
	if ((flags & _UC_POWERPC_VEC) && mcp->__vrf.__vrsave) {
		/* XXX we don't need to save the state, just to drop it */
		if (veclwp)
			save_vec(veclwp);
#ifdef MULTIPROCESSOR
		if (pcb->pcb_veccpu)
			save_vec_lwp(l);
#endif
		memcpy(pcb->pcb_vec.vreg, mcp->__vrf.__vr,
		    sizeof pcb->pcb_vec.vreg);
		pcb->pcb_vec.vscr[3] = mcp->__vrf.__vscr;
		pcb->pcb_vec.vrsave = mcp->__vrf.__vrsave;
		pcb->pcb_flags |= PCB_VEC;
		tf->srr1 &= ~PSL_VEC;
	}
#endif

#ifdef _UC_SETSTACK
	if (flags & _UC_SETSTACK)
		l->l_sigstk.ss_flags |= SS_ONSTACK;
	if (flags & _UC_CLRSTACK)
		l->l_sigstk.ss_flags &= ~SS_ONSTACK;
#endif
	return 0;
}

int
cpu_mcontext_validate(l, mcp)
	struct lwp *l;
	const mcontext_t *mcp;
{
	return 0;
}

int
cpu_lwp_setprivate(l, p)
	struct lwp *l;
	void *p;
{
	struct trapframe *tf = trapframe(l);

	tf->fixreg[2] = (register_t)p;
	return 0;
}

/*
 * Crash dump handling.
 */
u_int32_t dumpmag = 0x8fca0101;		/* magic number */
int dumpsize = 0;			/* size of dump in pages */
long dumplo = -1;			/* blocks */

/*
 * This is called by main to set dumplo and dumpsize.
 */
void
cpu_dumpconf()
{
	const struct bdevsw *bdev;
	int nblks;		/* size of dump device */
	int skip;

	if (dumpdev == NODEV)
		return;
	bdev = bdevsw_lookup(dumpdev);
	if (bdev == NULL) {
		dumpdev = NODEV;
		return;
	}
	if (bdev->d_psize == NULL)
		return;
	nblks = (*bdev->d_psize)(dumpdev);
	if (nblks <= ctod(1))
		return;

	dumpsize = physmem;

	/* Skip enough blocks at start of disk to preserve an eventual disklabel. */
	skip = LABELSECTOR + 1;
	skip += ctod(1) - 1;
	skip = ctod(dtoc(skip));
	if (dumplo < skip)
		dumplo = skip;

	/* Put dump at end of partition */
	if (dumpsize > dtoc(nblks - dumplo))
		dumpsize = dtoc(nblks - dumplo);
	if (dumplo < nblks - ctod(dumpsize))
		dumplo = nblks - ctod(dumpsize);
}

/*
 * Process the tail end of a posix_spawn() for the child.
 */
void
cpu_spawn_return(l)
	struct lwp *l;
{
	struct trapframe *frame = trapframe(l);

	userret(l, frame);
}

void
__syncicache(from, len)
	void *from;
	int len;
{
	int l, off;
	char *p;

	off = (u_int)from & (CACHELINESIZE - 1);
	len += off;

	p = from;
	l = len;
	do {
		asm volatile ("dcbst 0,%0" :: "r"(p));
		p += CACHELINESIZE;
	} while ((l -= CACHELINESIZE) > 0);
	asm volatile ("sync");

	p = from;
	l = len;
	do {
		asm volatile ("icbi 0,%0" :: "r"(p));
		p += CACHELINESIZE;
	} while ((l -= CACHELINESIZE) > 0);
	asm volatile ("sync");	/* required on 745x */
	asm volatile ("isync");
}
