/*	$Id: trap.c,v 1.56 2012/03/01 08:48:35 tsubai Exp $	*/

/*
 * Copyright (C) 1995, 1996 Wolfgang Solfrank.
 * Copyright (C) 1995, 1996 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "opt_altivec.h"
#include "opt_ddb.h"
#include "opt_multiprocessor.h"

#include <sys/param.h>
#include <sys/kauth.h>
#include <sys/kmem.h>
#include <sys/pool.h>
#include <sys/proc.h>
#include <sys/systm.h>

#include <uvm/uvm_extern.h>

#include <machine/altivec.h>
#include <machine/cpu.h>
#include <machine/db_machdep.h>
#include <machine/fpu.h>
#include <machine/frame.h>
#include <machine/pcb.h>
#include <machine/pmap.h>
#include <machine/psl.h>
#include <machine/trap.h>

static int fix_unaligned __P((struct lwp *, struct trapframe *));
static inline void setusr __P((int));

int badaddr_read __P((void *, size_t, int *));

void
trap(frame)
	struct trapframe *frame;
{
	struct lwp *l = curlwp;
	struct proc *p = l ? l->l_proc : NULL;
	struct pcb *pcb = curpcb;
	vaddr_t va = frame->dar;
	int type = frame->exc;
	int ftype, rv;
	ksiginfo_t ksi;

	curcpu()->ci_data.cpu_ntrap++;

	if (frame->srr1 & PSL_PR) {
		type |= EXC_USER;
		LWP_CACHE_CREDS(l, p);
	}

	switch (type) {
	case EXC_TRC|EXC_USER:
		frame->srr1 &= ~PSL_SE;
		KSI_INIT_TRAP(&ksi);
		ksi.ksi_signo = SIGTRAP;
		ksi.ksi_addr = (void *)frame->srr0;
		ksi.ksi_trap = EXC_TRC;
		ksi.ksi_code = TRAP_TRACE;
		trapsignal(l, &ksi);
		break;
	case EXC_DSI:
	{
		faultbuf *fb;

		if (frame->dsisr & DSISR_DABR) {	/* DABR match */
			printf("Data address breakpoint\n");
			goto brain_damage;
		}

#ifdef DIAGNOSTIC
		/*
		 * Only query UVM if no interrupts are active (this applies
		 * "on-fault" as well.
		 */
		if (intr_depth != -1)
			rv = EFAULT;
		else
#endif
		{
			struct vm_map *map;

			map = kernel_map;
			if ((va >> ADDR_SR_SHFT) == USER_SR) {
				sr_t user_sr;

				asm ("mfsr %0, %1"
				     : "=r"(user_sr) : "K"(USER_SR));
				va &= ADDR_PIDX | ADDR_POFF;
				va |= user_sr << ADDR_SR_SHFT;
				map = &p->p_vmspace->vm_map;
			}
			if (frame->dsisr & DSISR_STORE)
				ftype = VM_PROT_WRITE;
			else
				ftype = VM_PROT_READ;
			rv = uvm_fault(map, trunc_page(va), ftype);
			if (rv == 0)
				return;
			if (rv == EACCES)
				rv = EFAULT;
		}
		if ((fb = pcb->pcb_onfault) != NULL) {
			frame->srr0 = (*fb)[0];
			frame->fixreg[1] = (*fb)[1];
			frame->fixreg[2] = (*fb)[2];
			frame->fixreg[3] = rv;
			frame->cr = (*fb)[3];
			bcopy(&(*fb)[4], &frame->fixreg[13],
			    19 * sizeof(register_t));
			return;
		}
		goto brain_damage;
	}
	case EXC_DSI|EXC_USER:
		if (frame->dsisr & DSISR_STORE)
			ftype = VM_PROT_WRITE;
		else
			ftype = VM_PROT_READ;
		rv = uvm_fault(&p->p_vmspace->vm_map, trunc_page(va), ftype);
		if (rv == 0) {
			uvm_grow(p, va);
			break;
		}
		KSI_INIT_TRAP(&ksi);
		ksi.ksi_addr = (void *)va;
		ksi.ksi_trap = EXC_DSI;
		if (rv == EACCES)
			ksi.ksi_code = SEGV_ACCERR;
		else
			ksi.ksi_code = SEGV_MAPERR;
		if (rv == ENOMEM) {
			printf("UVM: pid %d (%s) lid %d, uid %d killed: "
			       "out of swap\n",
			       p->p_pid, p->p_comm, l->l_lid,
			       l->l_cred ? kauth_cred_geteuid(l->l_cred) : -1);
			ksi.ksi_signo = SIGKILL;
			trapsignal(l, &ksi);
		} else {
			ksi.ksi_signo = SIGSEGV;
			trapsignal(l, &ksi);
		}
		break;
	case EXC_ISI|EXC_USER:
		ftype = VM_PROT_READ | VM_PROT_EXECUTE;
		rv = uvm_fault(&p->p_vmspace->vm_map, trunc_page(frame->srr0),
		    ftype);
		if (rv == 0)
			break;
		KSI_INIT_TRAP(&ksi);
		ksi.ksi_signo = SIGSEGV;
		ksi.ksi_addr = (void *)frame->srr0;
		ksi.ksi_trap = EXC_ISI;
		ksi.ksi_code = SEGV_MAPERR;
		trapsignal(l, &ksi);
		break;
	case EXC_FPU|EXC_USER:
		if (fpulwp)
			save_fpu(fpulwp);
#ifdef MULTIPROCESSOR
		if (pcb->pcb_fpcpu)
			save_fpu_lwp(l);
#endif
		fpulwp = l;
		pcb->pcb_fpcpu = curcpu();
		enable_fpu(l);
		break;

	case EXC_VEC|EXC_USER:
#ifdef ALTIVEC
		if (veclwp)
			save_vec(veclwp);
#ifdef MULTIPROCESSOR
		if (pcb->pcb_veccpu)
			save_vec_lwp(l);
#endif
		veclwp = l;
		pcb->pcb_veccpu = curcpu();
		enable_vec(l);
#else
		KSI_INIT_TRAP(&ksi);
		ksi.ksi_signo = SIGILL;
		ksi.ksi_addr = (void *)frame->srr0;
		ksi.ksi_trap = EXC_VEC;
		ksi.ksi_code = ILL_ILLOPC;
		trapsignal(l, &ksi);
#endif /* ALTIVEC */
		break;

	case EXC_VECA|EXC_USER:		/* XXX not yet */
		KSI_INIT_TRAP(&ksi);
		ksi.ksi_signo = SIGFPE;
		ksi.ksi_addr = (void *)frame->srr0;
		ksi.ksi_trap = EXC_VECA;
		ksi.ksi_code = FPE_FLTINV;
		trapsignal(l, &ksi);
		break;

	case EXC_AST|EXC_USER:
		astpending = 0;		/* we are about to do it */
		curcpu()->ci_data.cpu_nsoft++;
		if (l->l_pflag & LP_OWEUPC) {
			l->l_pflag &= ~LP_OWEUPC;
			ADDUPROF(l);
		}
		/* Check whether we are being preempted. */
		if (want_resched)
			preempt();
		break;

	case EXC_ALI|EXC_USER:
		if (fix_unaligned(l, frame) != 0) {
			KSI_INIT_TRAP(&ksi);
			ksi.ksi_signo = SIGBUS;
			ksi.ksi_addr = (void *)frame->srr0;
			ksi.ksi_trap = EXC_ALI;
			ksi.ksi_code = BUS_ADRALN;
			trapsignal(l, &ksi);
		} else
			frame->srr0 += 4;
		break;

	case EXC_PGM|EXC_USER:
		KSI_INIT_TRAP(&ksi);
		ksi.ksi_addr = (void *)frame->srr0;
		ksi.ksi_trap = EXC_PGM;
		if (frame->srr1 & 0x00020000) {
			ksi.ksi_signo = SIGTRAP;
			ksi.ksi_code = TRAP_BRKPT;
			trapsignal(l, &ksi);
		} else if (frame->srr1 & 0x00100000) {
			/* XXX This doesn't happen unless PSL_FEn are set. */
			save_fpu(l);
			ksi.ksi_signo = SIGFPE;
			ksi.ksi_code = FPE_FLTRES;	/* XXX recheck */
			trapsignal(l, &ksi);
		} else {
			ksi.ksi_signo = SIGILL;
			ksi.ksi_code = ILL_ILLOPC;
			trapsignal(l, &ksi);
		}
		break;

	case EXC_MCHK:
		if (pcb->pcb_onfault) {
			faultbuf *fb = pcb->pcb_onfault;

			frame->srr0 = (*fb)[0];
			frame->fixreg[1] = (*fb)[1];
			frame->fixreg[2] = (*fb)[2];
			frame->fixreg[3] = EFAULT;
			frame->cr = (*fb)[3];
			bcopy(&(*fb)[4], &frame->fixreg[13],
			      19 * sizeof(register_t));
			return;
		}
		goto brain_damage;

	default:
brain_damage:
		printf("trap type %x at %lx\n", type, frame->srr0);
#ifdef DDB
		if (kdb_trap(type, frame))
			return;
#endif
#ifdef TRAP_PANICWAIT
		printf("Press a key to panic.\n");
		cngetc();
#endif
		panic("trap");
	}

	userret(l, frame);
}

static inline void
setusr(content)
	int content;
{
	asm volatile ("isync; mtsr %0,%1; isync"
		      :: "n"(USER_SR), "r"(content));
}

int
copyin(udaddr, kaddr, len)
	const void *udaddr;
	void *kaddr;
	size_t len;
{
	const char *up = udaddr;
	char *kp = kaddr;
	char *p;
	int rv;
	size_t l;
	faultbuf env;

	if ((rv = setfault(env)) != 0) {
		curpcb->pcb_onfault = 0;
		return rv;
	}
	while (len > 0) {
		p = (char *)USER_ADDR + ((u_int)up & ~SEGMENT_MASK);
		l = ((char *)USER_ADDR + SEGMENT_LENGTH) - p;
		if (l > len)
			l = len;
		setusr(curpcb->pcb_pm->pm_sr[(u_int)up >> ADDR_SR_SHFT]);
		bcopy(p, kp, l);
		up += l;
		kp += l;
		len -= l;
	}
	asm volatile ("sync");
	curpcb->pcb_onfault = 0;
	return 0;
}

int
copyout(kaddr, udaddr, len)
	const void *kaddr;
	void *udaddr;
	size_t len;
{
	const char *kp = kaddr;
	char *up = udaddr;
	char *p;
	int rv;
	size_t l;
	faultbuf env;

	if ((rv = setfault(env)) != 0) {
		curpcb->pcb_onfault = 0;
		return rv;
	}
	while (len > 0) {
		p = (char *)USER_ADDR + ((u_int)up & ~SEGMENT_MASK);
		l = ((char *)USER_ADDR + SEGMENT_LENGTH) - p;
		if (l > len)
			l = len;
		setusr(curpcb->pcb_pm->pm_sr[(u_int)up >> ADDR_SR_SHFT]);
		bcopy(kp, p, l);
		up += l;
		kp += l;
		len -= l;
	}
	asm volatile ("sync");
	curpcb->pcb_onfault = 0;
	return 0;
}

/*
 * kcopy(const void *src, void *dst, size_t len);
 *
 * Copy len bytes from src to dst, aborting if we encounter a fatal
 * page fault.
 *
 * kcopy() _must_ save and restore the old fault handler since it is
 * called by uiomove(), which may be in the path of servicing a non-fatal
 * page fault.
 */
int
kcopy(src, dst, len)
	const void *src;
	void *dst;
	size_t len;
{
	faultbuf env, *oldfault;
	int rv;

	oldfault = curpcb->pcb_onfault;
	if ((rv = setfault(env)) != 0) {
		curpcb->pcb_onfault = oldfault;
		return rv;
	}

	bcopy(src, dst, len);

	asm volatile ("sync");
	curpcb->pcb_onfault = oldfault;
	return 0;
}

int
badaddr(addr, size)
	void *addr;
	size_t size;
{
	return badaddr_read(addr, size, NULL);
}

int
badaddr_read(addr, size, rptr)
	void *addr;
	size_t size;
	int *rptr;
{
	faultbuf env;
	int x;

	/* Get rid of any stale machine checks that have been waiting.  */
	asm volatile ("sync; isync");

	if (setfault(env)) {
		curpcb->pcb_onfault = 0;
		asm volatile ("sync");
		return 1;
	}

	asm volatile ("sync");

	switch (size) {
	case 1:
		x = *(volatile int8_t *)addr;
		break;
	case 2:
		x = *(volatile int16_t *)addr;
		break;
	case 4:
		x = *(volatile int32_t *)addr;
		break;
	default:
		panic("badaddr: invalid size (%d)", size);
	}

	/* Make sure we took the machine check, if we caused one. */
	asm volatile ("sync; isync");

	curpcb->pcb_onfault = 0;
	asm volatile ("sync");	/* To be sure. */

	/* Use the value to avoid reorder. */
	if (rptr)
		*rptr = x;

	return 0;
}

/*
 * For now, this only deals with the particular unaligned access case
 * that gcc tends to generate.  Eventually it should handle all of the
 * possibilities that can happen on a 32-bit PowerPC in big-endian mode.
 */

static int
fix_unaligned(l, frame)
	struct lwp *l;
	struct trapframe *frame;
{
	int indicator = EXC_ALI_OPCODE_INDICATOR(frame->dsisr);

	switch (indicator) {
	case EXC_ALI_LFD:
	case EXC_ALI_STFD:
		{
			int reg = EXC_ALI_RST(frame->dsisr);
			double *fpr = &curpcb->pcb_fpu.fpreg[reg];

			/* Juggle the FPU to ensure that we've initialized
			 * the FPRs, and that their current state is in
			 * the PCB.
			 */
			if (fpulwp != l) {
				if (fpulwp)
					save_fpu(fpulwp);
				enable_fpu(l);
			}
			save_fpu(l);

			if (indicator == EXC_ALI_LFD) {
				if (copyin((void *)frame->dar, fpr,
				    sizeof(double)) != 0)
					return -1;
				enable_fpu(l);
			} else {
				if (copyout(fpr, (void *)frame->dar,
				    sizeof(double)) != 0)
					return -1;
			}
			return 0;
		}
		break;
	}

	return -1;
}

/*
 * Start a new LWP.
 */
void
startlwp(arg)
	void *arg;
{
	int err;
	ucontext_t *uc = arg;
	struct lwp *l = curlwp;
	struct trapframe *frame = trapframe(l);

	err = cpu_setmcontext(l, &uc->uc_mcontext, uc->uc_flags);
#ifdef DIAGNOSTIC
	if (err)
		printf("Error %d from cpu_setmcontext.", err);
#endif
	kmem_free(uc, sizeof *uc);
	userret(l, frame);
}

#include <sys/userret.h>

void
userret(l, frame)
	struct lwp *l;
	struct trapframe *frame;
{
	mi_userret(l);

	/*
	 * If someone stole the fp or vector unit while we were away,
	 * disable it
	 */
	if (l != fpulwp || ((struct pcb *)l->l_addr)->pcb_fpcpu != curcpu())
		frame->srr1 &= ~PSL_FP;
#ifdef ALTIVEC
	if (l != veclwp || ((struct pcb *)l->l_addr)->pcb_veccpu != curcpu())
		frame->srr1 &= ~PSL_VEC;
#endif
}
