/*	$Id: obio.c,v 1.27 2015/06/23 19:26:53 tsubai Exp $	*/

/*-
 * Copyright (c) 2004 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "locators.h"

#include <sys/param.h>
#include <sys/device.h>
#include <sys/systm.h>

#include <machine/autoconf.h>
#include <machine/mpc860dev.h>

#include <ts3000/board.h>

struct obioreg {
	u_char reset1;
	u_char clock_enable1;
	u_char clock_source1;
	u_char misc;
	u_char intr_enable1;
	u_char intr_enable2;
	u_char intr_enable3;
	u_char intr_enable4;
	u_char intr_status1;
	u_char intr_status2;
	u_char intr_status3;
	u_char intr_status4;
	u_char fpga_id;
	u_char reserved1[3];
	u_char reset2;
	u_char clock_enable2;
	u_char clock_source2;
	u_char reserved2;
	u_char intr_enable5;
	u_char intr_enable6;
	u_char reserved3[2];
	u_char intr_status5;
	u_char intr_status6;
	u_char reserved4[6];
	u_char bsmux;
	u_char reserved5[6];
	u_char cpld_id;
};

#define CF_RESET	0x02
#define CF_INTR_EN	0x04

struct obio_softc {
	struct device sc_dev;
};

int obio_match(struct device *, struct cfdata *, void *);
void obio_attach(struct device *, struct device *, void *);
int obio_print(void *, const char *);

void obio_com_reset(int);
void obio_gpio_init(void);
int cpldintr1(void *);
int cpldintr2(void *);
int comintr(void *);

extern time_t lastcomintr;
static volatile struct obioreg *cpld;
static void *coms[48];

CFDECL(obio, sizeof(struct obio_softc), obio_match, obio_attach);

int
obio_match(parent, cf, aux)
	struct device *parent;
	struct cfdata *cf;
	void *aux;
{
	struct confargs *ca = aux;

	if (strcmp(ca->ca_name, "obio") == 0)
		return 1;

	return 0;
}

void
obio_attach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
{
	volatile struct mpc860dev *immr = get_immr();
	struct device *com;
	struct confargs ca;
	paddr_t flash_base[2], wdc_base, cpld_base, uart_base;
	int locs[OBIOCF_NLOCS], wdc_vec, nports, i;

	printf("\n");

	switch (board_type) {
	case BOARD_TS3000:
		flash_base[0] = immr->siu_br0 & 0xffff8000;
		flash_base[1] = 0;
		wdc_vec = 6;	/* IRQ 3 */
		wdc_base = immr->siu_br5 & 0xffff8000;
		cpld_base = immr->siu_br6 & 0xffff8000;
		uart_base = immr->siu_br7 & 0xffff8000;
		break;
	case BOARD_ACS16:
	case BOARD_ACS32:
	case BOARD_ACS48:
		flash_base[0] = immr->siu_br0 & 0xffff8000;
		flash_base[1] = 0;
		wdc_vec = 10;	/* IRQ 5 */
		wdc_base = immr->siu_br2 & 0xffff8000;
		cpld_base = immr->siu_br3 & 0xffff8000;
		uart_base = immr->siu_br4 & 0xffff8000;
		break;
	default:
		flash_base[0] = immr->siu_br0 & 0xffff8000;
		flash_base[1] = immr->siu_br1 & 0xffff8000;
		wdc_base = 0;
		cpld_base = immr->siu_br3 & 0xffff8000;
		uart_base = immr->siu_br4 & 0xffff8000;
	}

	cpld = (void *)cpld_base;

	if (wdc_base) {
		/* XXX LED? */
		cpld->misc |= 0x08;

#if 0		/* Reset IDE. */
		cpld->misc |= CF_RESET;
		delay(100);
		cpld->misc &= ~CF_RESET;
		delay(1000000);
#endif

		/* Enable IDE interrupts. */
		cpld->misc |= CF_INTR_EN;

		ca.ca_name = "wdc";
		ca.ca_addr = wdc_base;
		ca.ca_vec[0] = wdc_vec;
		config_found(self, &ca, obio_print);
	}

	for (i = 0; i < 2; i++) {
		if (flash_base[i] == 0)
			break;
		ca.ca_name = "flash";
		ca.ca_addr = flash_base[i];
		ca.ca_vec[0] = -1;
		config_found(self, &ca, obio_print);
	}

	switch (board_type) {
	case BOARD_TS3000: nports = 48; break;
	case BOARD_TS2000: nports = 32; break;
	case BOARD_TS1000: nports = 16; break;
	case BOARD_TS800:  nports = 8; break;
	case BOARD_TS400:  nports = 4; break;
	case BOARD_ACS48:  nports = 48; break;
	case BOARD_ACS32:  nports = 32; break;
	case BOARD_ACS16:  nports = 16; break;
	case BOARD_CS4008: nports = 8; break;
	default:
		panic("unknown board");
	}

	obio_gpio_init();
	obio_com_reset(nports);

	switch (nports) {
	case 48:
		cpld->intr_enable6 = 0xff;	/* port 41-48 */
		cpld->intr_enable5 = 0xff;	/* port 33-40 */
	case 32:
		cpld->intr_enable4 = 0xff;	/* port 25-32 */
		cpld->intr_enable3 = 0xff;	/* port 17-24 */
	case 16:
		cpld->intr_enable2 = 0xff;	/* port  9-16 */
	case 8:
		cpld->intr_enable1 = 0xff;	/* port  1-8 */
		break;
	case 4:
		cpld->intr_enable1 = 0x0f;	/* port  1-4 */
	}

	for (i = 0; i < nports; i++) {
		locs[OBIOCF_VEC] = OBIOCF_VEC_DEFAULT;
		locs[OBIOCF_ADDR] = uart_base + i * 8;
		ca.ca_name = "com";
		ca.ca_addr = uart_base + i * 8;
		if (i < 32)
			ca.ca_vec[0] = 2;	/* IRQ1 */
		else
			ca.ca_vec[0] = 4;	/* IRQ2 */
		com = config_found_sm_loc(self, "obio", locs, &ca, obio_print,
		    config_stdsubmatch);
		if (com)
			coms[i] = device_private(com);
	}
	intr_establish(2, IST_LEVEL, IPL_SERIAL, cpldintr1, self);
	if (nports < 32)
		return;
	intr_establish(4, IST_LEVEL, IPL_SERIAL, cpldintr2, self);
}

int
cpldintr1(v)
	void *v;
{
	u_int status;
	int n;

	status = cpld->intr_status1 | cpld->intr_status2 << 8 |
	    cpld->intr_status3 << 16 | cpld->intr_status4 << 24;
	while (status) {
		n = ffs(status) - 1;
		status &= ~(1 << n);
		if (coms[n])
			comintr(coms[n]);
	}
	lastcomintr = time_second;
	return 1;
}

int
cpldintr2(v)
	void *v;
{
	u_int status;
	int n;

	status = cpld->intr_status5 | cpld->intr_status6 << 8;
	while (status) {
		n = ffs(status) - 1;
		status &= ~(1 << n);
		if (coms[32 + n])
			comintr(coms[32 + n]);
	}
	lastcomintr = time_second;
	return 1;
}

int
obio_print(aux, bus)
	void *aux;
	const char *bus;
{
	struct confargs *ca = aux;

	if (bus)
		printf("%s at %s", ca->ca_name, bus);
	if (ca->ca_addr)
		printf(" addr 0x%x", ca->ca_addr);
	if (ca->ca_vec[0] > 0)
		printf(" vec %d", ca->ca_vec[0]);
	return UNCONF;
}

void
obio_com_reset(nports)
	int nports;
{
	u_int mask;

	cpld->intr_enable1 = 0;
	cpld->intr_enable2 = 0;
	cpld->intr_enable3 = 0;
	cpld->intr_enable4 = 0;
	cpld->intr_enable5 = 0;
	cpld->intr_enable6 = 0;

	if (nports > 32) {
		cpld->reset1 = 0xff;
		cpld->reset2 = 0x0f;
		cpld->clock_source1 = 0;
		cpld->clock_source2 = 0;
		cpld->clock_enable1 = 0xff;
		cpld->clock_enable2 = 0x0f;
		cpld->misc |= 0x01;
		delay(10000);
		cpld->reset1 = 0;
		cpld->reset2 = 0;
	} else {
		mask = (1 << (nports / 4)) - 1;

		cpld->reset1 = mask;
		cpld->clock_source1 = 0;
		cpld->clock_enable1 = mask;
		cpld->misc |= 0x01;
		delay(10000);
		cpld->reset1 = 0;
	}
}

void
obio_gpio_init()
{
	volatile struct mpc860dev *immr = get_immr();

	switch (board_type) {
	case BOARD_TS3000:
		immr->cpm_padir |= 0xafff;
		immr->cpm_papar &= ~0xefff;
		immr->cpm_paodr &= ~0xefff;
		immr->cpm_padat &= ~0xefff;

		immr->cpm_pbdir &= ~0x1000;
		immr->cpm_pbodr &= ~0x1000;
		immr->cpm_pbpar &= ~0x1000;

		immr->cpm_pcdir &= ~0x0700;
		immr->cpm_pcpar &= ~0x0700;
		immr->cpm_pcdat &= ~0x0700;
		break;
	case BOARD_ACS16:
	case BOARD_ACS32:
	case BOARD_ACS48:
		immr->cpm_padir |= 0xafff;
		immr->cpm_papar &= ~0xefff;
		immr->cpm_paodr &= ~0xefff;
		immr->cpm_padat &= ~0xefff;

		immr->cpm_pbdir &= ~0x11f00;	/* XXX CHECK */
		immr->cpm_pbodr &= ~0x11f00;
		immr->cpm_pbpar &= ~0x11f00;

		immr->cpm_pcdir &= ~0x07f8;	/* XXX CHECK */
		immr->cpm_pcpar &= ~0x07f8;
		immr->cpm_pcdat &= ~0x07f8;
		break;
	default:
		immr->cpm_padir |= 0xafff;
		immr->cpm_papar &= ~0xafff;
		immr->cpm_paodr &= ~0xafff;
		immr->cpm_padat &= ~0xafff;

		immr->cpm_pbdir &= ~0x1c3e;
		immr->cpm_pbodr &= ~0x1c3e;
		immr->cpm_pbpar &= ~0x1c3e;

		immr->cpm_pcdir &= ~0x0728;
		immr->cpm_pcpar &= ~0x0728;
		immr->cpm_pcdat &= ~0x0728;
	}
}
