/*	$Id: rtc.c,v 1.3 2015/06/25 22:34:46 tsubai Exp $	*/

/*-
 * Copyright (c) 2009 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <dev/clock_subr.h>
#include <machine/autoconf.h>
#include <machine/mpc860dev.h>
#include <ts3000/board.h>

static int rtc_gettime(struct todr_chip_handle *, struct timeval *);
static int rtc_settime(struct todr_chip_handle *, struct timeval *);
static int m4t_init(void);
static int m4t_gettime(struct todr_chip_handle *, struct clock_ymdhms *);
static int m4t_settime(struct todr_chip_handle *, struct clock_ymdhms *);

void
rtc_attach()
{
	static struct todr_chip_handle rtc;

	switch (board_type) {
	default:
		rtc.todr_gettime = rtc_gettime;
		rtc.todr_settime = rtc_settime;
		break;
	case BOARD_ACS16:
	case BOARD_ACS32:
	case BOARD_ACS48:
	case BOARD_CS4008:
		m4t_init();
		rtc.todr_gettime_ymdhms = m4t_gettime;
		rtc.todr_settime_ymdhms = m4t_settime;
		break;
	}
	todr_attach(&rtc);
}

int
rtc_gettime(ch, t)
	struct todr_chip_handle *ch;
	struct timeval *t;
{
	volatile struct mpc860dev *immr = get_immr();
	time_t delta;

	t->tv_sec = immr->siu_rtc;
	immr->siu_rtck = 0;		/* lock */

	if (ch->base_time) {
		delta = t->tv_sec - ch->base_time;
		if (delta < 0)
			delta = -delta;
		if (delta > SECYR*5)
			return EINVAL;
	}
	return 0;
}

int
rtc_settime(ch, t)
	struct todr_chip_handle *ch;
	struct timeval *t;
{
	volatile struct mpc860dev *immr = get_immr();

	immr->siu_rtck = 0x55ccaa33;	/* unlock */
	immr->siu_rtc = t->tv_sec;
	immr->siu_rtck = 0;		/* lock again */
	return 0;
}

int
m4t_init()
{
	volatile struct mpc860dev *immr = get_immr();

	/* Initialize PB31. */
	immr->cpm_pbodr &= ~0x00001;
	immr->cpm_pbpar &= ~0x00001;
	immr->cpm_pbdat |=  0x00001;
	immr->cpm_pbdir |=  0x00001;
	return 0;
}

int
m4t_gettime(ch, dt)
	struct todr_chip_handle *ch;
	struct clock_ymdhms *dt;
{
	volatile struct mpc860dev *immr = get_immr();
	int error = 0;
	u_char bcd[8];

	/* RTC chip enable */
	immr->cpm_pbdat &= ~0x00001;
	delay(2000);

	bcd[0] = 0x00 | 1;		/* READ | register */
	if (spi_xfer(bcd, 8)) {
		error = EIO;
		goto out;
	}

	dt->dt_sec  = FROMBCD(bcd[1]);
	dt->dt_min  = FROMBCD(bcd[2]);
	dt->dt_hour = FROMBCD(bcd[3] & 0x3f);
	dt->dt_day  = FROMBCD(bcd[5]);
	dt->dt_mon  = FROMBCD(bcd[6]);
	dt->dt_year = FROMBCD(bcd[7]) + 2000;

out:
	immr->cpm_pbdat |= 0x00001;
	return error;
}

int
m4t_settime(ch, dt)
	struct todr_chip_handle *ch;
	struct clock_ymdhms *dt;
{
	volatile struct mpc860dev *immr = get_immr();
	int error = 0;
	u_char bcd[8];

	/* XXX äη夬꤬Ⱥ */
	bcd[0] = 0x80 | 1;		/* WRITE | register */
	bcd[1] = TOBCD(dt->dt_sec);
	bcd[2] = TOBCD(dt->dt_min);
	bcd[3] = TOBCD(dt->dt_hour);
	bcd[4] = dt->dt_wday;
	bcd[5] = TOBCD(dt->dt_day);
	bcd[6] = TOBCD(dt->dt_mon);
	bcd[7] = TOBCD(dt->dt_year - 2000);

	/* RTC chip enable */
	immr->cpm_pbdat &= ~0x00001;
	delay(2000);

	if (spi_xfer(bcd, 8))
		error = EIO;

	immr->cpm_pbdat |= 0x00001;
	return error;
}
