/*	$Id: installboot.c,v 1.3 2011/01/07 03:50:10 tsubai Exp $	*/

/*-
 * Copyright (c) 1998 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by Paul Kranenburg.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *        This product includes software developed by the NetBSD
 *        Foundation, Inc. and its contributors.
 * 4. Neither the name of The NetBSD Foundation nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/disklabel.h>
#include <sys/exec_elf.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/statvfs.h>
#include <sys/sysctl.h>

#include <ufs/ufs/dinode.h>
#include <ufs/ufs/dir.h>
#include <ufs/ffs/fs.h>

#ifdef BOOT_AOUT
#include <a.out.h>
#endif
#include <err.h>
#include <fcntl.h>
#include <nlist.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "extern.h"

int	verbose, nowrite;
char	*boot, *proto, *dev;

#define BOOTSECTOR_OFFSET 0

struct nlist nl[] = {
#define X_BLOCKTABLE	0
	{"_block_table"},
#define X_BLOCKSIZE	1
	{"_block_size"},
	{NULL}
};

u_int	*block_table;		/* block number array in prototype image */
u_int	*block_size_p;		/* filesystem block size */
u_int	max_block_count = 16;

int 	main __P((int, char *[]));
char	*loadprotoblocks __P((char *, long *));
int	loadblocknums __P((char *, int));
int	loaddisklabel(char *, int, int);
void usage(void);
void devread(int, void *, int, size_t, char *);

void
usage()
{
	fprintf(stderr, "\
usage: installboot [-n] [-v] <boot> <proto> <device>\n\
       installboot -c <boot>\n");
	exit(1);
}

int
main(argc, argv)
	int argc;
	char *argv[];
{
	int	c;
	int	devfd;
	char	*protostore;
	long	protosize;
	int	crconly = 0;

	while ((c = getopt(argc, argv, "cnv")) != -1) {
		switch (c) {

		case 'n':
			/* Do not actually write the bootblock to disk */
			nowrite = 1;
			break;
		case 'v':
			verbose = 1;
			break;
		case 'c':
			/* Fill CRC */
			crconly = 1;
			break;
		default:
			usage();
		}
	}

	argc -= optind;
	argv += optind;

	if (crconly) {
		if (argc < 1)
			usage();
		addcrc(argv[0]);
		exit(0);
	}

	if (argc < 3)
		usage();

	boot = argv[0];
	proto = argv[1];
	dev = argv[2];

	if (verbose) {
		printf("boot: %s\n", boot);
		printf("proto: %s\n", proto);
		printf("device: %s\n", dev);
	}

	/* Load proto blocks into core */
	if ((protostore = loadprotoblocks(proto, &protosize)) == NULL)
		exit(1);

	/* Open and check raw disk device */
	if ((devfd = open(dev, O_RDONLY, 0)) < 0)
		err(1, "open: %s", dev);

	/* Extract and load block numbers */
	if (loadblocknums(boot, devfd) != 0)
		exit(1);

	/* Preserve disklabel, fill magic and CRC */
	if (loaddisklabel(protostore, protosize, devfd))
		exit(1);

	(void)close(devfd);

	if (nowrite)
		return 0;

	/* Write patched proto bootblocks into the superblock */
	if (protosize > SBLOCKSIZE - DEV_BSIZE)
		errx(1, "proto bootblocks too big");

	if ((devfd = open(dev, O_RDWR, 0)) < 0)
		err(1, "open: %s", dev);

	if (lseek(devfd, BOOTSECTOR_OFFSET, SEEK_SET) != BOOTSECTOR_OFFSET)
		err(1, "lseek bootstrap");

	/* Sync filesystems (to clean in-memory superblock?) */
	sync();

	if (write(devfd, protostore, protosize) != protosize)
		err(1, "write bootstrap");
	(void)close(devfd);
	return 0;
}

char *
loadprotoblocks(fname, size)
	char *fname;
	long *size;
{
	int	fd, sz;
	char	*bp;
	struct stat statbuf;
#ifdef BOOT_AOUT
	struct exec *hp;
#endif
	long	off;
	Elf32_Ehdr *eh;
	Elf32_Phdr *ph;

	/* Locate block number array in proto file */
	if (nlist(fname, nl) != 0) {
		warnx("nlist: %s: symbols not found", fname);
		return NULL;
	}
#ifdef BOOT_AOUT
	if (nl[X_BLOCKTABLE].n_type != N_DATA + N_EXT) {
		warnx("nlist: %s: wrong type", nl[X_BLOCKTABLE].n_un.n_name);
		return NULL;
	}
	if (nl[X_BLOCKCOUNT].n_type != N_DATA + N_EXT) {
		warnx("nlist: %s: wrong type", nl[X_BLOCKCOUNT].n_un.n_name);
		return NULL;
	}
	if (nl[X_BLOCKSIZE].n_type != N_DATA + N_EXT) {
		warnx("nlist: %s: wrong type", nl[X_BLOCKSIZE].n_un.n_name);
		return NULL;
	}
#endif

	if ((fd = open(fname, O_RDONLY)) < 0) {
		warn("open: %s", fname);
		return NULL;
	}
	if (fstat(fd, &statbuf) != 0) {
		warn("fstat: %s", fname);
		close(fd);
		return NULL;
	}
	if ((bp = calloc(roundup(statbuf.st_size, DEV_BSIZE), 1)) == NULL) {
		warnx("malloc: %s: no memory", fname);
		close(fd);
		return NULL;
	}
	if (read(fd, bp, statbuf.st_size) != statbuf.st_size) {
		warn("read: %s", fname);
		free(bp);
		close(fd);
		return NULL;
	}
	close(fd);

#ifdef BOOT_AOUT
	hp = (struct exec *)bp;
#endif
	eh = (Elf32_Ehdr *)bp;
	ph = (Elf32_Phdr *)(bp + eh->e_phoff);
	sz = roundup(ph->p_filesz, 512);
	bzero(bp + ph->p_offset + ph->p_filesz, sz - ph->p_filesz);

	/* Calculate the symbols' location within the proto file */
	off = ph->p_offset - ph->p_vaddr;
	block_table = (u_int *)(bp + nl[X_BLOCKTABLE].n_value + off);
	block_size_p = (u_int *)(bp + nl[X_BLOCKSIZE].n_value + off);

	if ((int)block_table & 3) {
		warnx("%s: invalid address: block_table = %p",
		     fname, block_table);
		free(bp);
		close(fd);
		return NULL;
	}
	if ((int)block_size_p & 3) {
		warnx("%s: invalid address: block_size_p = %p",
		     fname, block_size_p);
		free(bp);
		close(fd);
		return NULL;
	}

	if (verbose)
		printf("proto bootblock size: %d\n", sz);

	/*
	 * We convert the a.out header in-vitro into something that
	 * Sun PROMs understand.
	 * Old-style (sun4) ROMs do not expect a header at all, so
	 * we turn the first two words into code that gets us past
	 * the 32-byte header where the actual code begins. In assembly
	 * speak:
	 *	.word	MAGIC		! a NOP
	 *	ba,a	start		!
	 *	.skip	24		! pad
	 * start:
	 */

	*size = sz;
	return bp + ph->p_offset;
}

void
devread(fd, buf, blk, size, msg)
	int	fd;
	void	*buf;
	int	blk;
	size_t	size;
	char	*msg;
{
	if (lseek(fd, dbtob(blk), SEEK_SET) != dbtob(blk))
		err(1, "%s: devread: lseek", msg);

	if (read(fd, buf, size) != size)
		err(1, "%s: devread: read", msg);
}

static char sblock[SBLOCKSIZE];

int
loadblocknums(boot, devfd)
char	*boot;
int	devfd;
{
	int		i, fd, blk, ndb;
	struct stat	statbuf;
	struct statvfs	statvfsbuf;
	struct fs	*fs;
	char		*buf;
	int32_t		*ap1;
	int64_t		*ap2;
	struct ufs1_dinode *ip1;
	struct ufs2_dinode *ip2;
	off_t		sboff;
	static off_t	sbsearch[] = SBLOCKSEARCH;
#define ufs1 (fs->fs_magic == FS_UFS1_MAGIC)

	/*
	 * Open 2nd-level boot program and record the block numbers
	 * it occupies on the filesystem represented by `devfd'.
	 */
	if ((fd = open(boot, O_RDONLY)) < 0)
		err(1, "open: %s", boot);

	if (fstatvfs(fd, &statvfsbuf) != 0)
		err(1, "statvfs: %s", boot);

	if (strcmp(statvfsbuf.f_fstypename, "ffs") != 0 &&
	    strcmp(statvfsbuf.f_fstypename, "ufs") != 0) {
		errx(1, "%s: must be on an FFS filesystem", boot);
	}

	if (fsync(fd) != 0)
		err(1, "fsync: %s", boot);

	if (fstat(fd, &statbuf) != 0)
		err(1, "fstat: %s", boot);

	close(fd);

	/* Read superblock */
	for (i = 0; sbsearch[i] != -1; i++) {
		sboff = sbsearch[i];
		devread(devfd, sblock, btodb(sboff), SBLOCKSIZE, "superblock");
		fs = (struct fs *)sblock;
		if (fs->fs_magic == FS_UFS1_MAGIC ||
		    fs->fs_magic == FS_UFS2_MAGIC)
			goto found;
	}
	errx(1, "No superblock found");

found:
	/* Read inode */
	if ((buf = malloc(fs->fs_bsize)) == NULL)
		errx(1, "No memory for filesystem block");

	blk = fsbtodb(fs, ino_to_fsba(fs, statbuf.st_ino));
	devread(devfd, buf, blk, fs->fs_bsize, "inode");
	ip1 = (struct ufs1_dinode *)buf + ino_to_fsbo(fs, statbuf.st_ino);
	ip2 = (struct ufs2_dinode *)buf + ino_to_fsbo(fs, statbuf.st_ino);

	/*
	 * Register filesystem block size.
	 */
	*block_size_p = fs->fs_bsize;

	/*
	 * Get the block numbers; we don't handle fragments
	 */
	if (ufs1)
		ndb = howmany(ip1->di_size, fs->fs_bsize);
	else
		ndb = howmany(ip2->di_size, fs->fs_bsize);
	if (ndb > max_block_count)
		errx(1, "%s: Too many blocks", boot);

	if (verbose)
		printf("%s: block numbers: ", boot);
	ap1 = ip1->di_db;
	ap2 = ip2->di_db;
	if (ufs1) {
		for (i = 0; i < NDADDR && *ap1 && ndb; i++, ap1++, ndb--) {
			blk = fsbtodb(fs, *ap1);
			block_table[i] = blk;
			if (verbose)
				printf("%d ", blk);
		}
	} else {
		for (i = 0; i < NDADDR && *ap2 && ndb; i++, ap2++, ndb--) {
			blk = fsbtodb(fs, *ap2);
			block_table[i] = blk;
			if (verbose)
				printf("%d ", blk);
		}
	}
	if (verbose)
		printf("\n");

	if (ndb == 0)
		return 0;

	/*
	 * Just one level of indirections; there isn't much room
	 * for more in the 1st-level bootblocks anyway.
	 */
	if (verbose)
		printf("%s: block numbers (indirect): ", boot);
	if (ufs1)
		blk = ip1->di_ib[0];
	else
		blk = ip2->di_ib[0];
	devread(devfd, buf, blk, fs->fs_bsize, "indirect block");
	ap1 = (int32_t *)buf;
	ap2 = (int64_t *)buf;
	if (ufs1)
		for (; i < NINDIR(fs) && *ap1 && ndb; i++, ap1++, ndb--) {
			blk = fsbtodb(fs, *ap1);
			block_table[i] = blk;
			if (verbose)
				printf("%d ", blk);
		}
	else
		for (; i < NINDIR(fs) && *ap2 && ndb; i++, ap2++, ndb--) {
			blk = fsbtodb(fs, *ap2);
			block_table[i] = blk;
			if (verbose)
				printf("%d ", blk);
		}
	if (verbose)
		printf("\n");

	if (ndb)
		errx(1, "%s: Too many blocks", boot);
	return 0;
}

int
loaddisklabel(protostore, protosize, devfd)
	char *protostore;
	int protosize, devfd;
{
	int n;
	u_int16_t *src, *dst;
	char labelblock[512];

	devread(devfd, labelblock, LABELSECTOR, 512, "disklabel");

	n = sizeof(struct disklabel);
	src = (void *)(labelblock + LABELOFFSET);
	dst = (void *)(protostore + LABELOFFSET);
	while (n > 0) {
		*dst++ = bswap16(*src++);
		n -= 2;
	}

	fillmagic(protostore, protosize);

	n = protosize;
	src = (void *)protostore;
	while (n > 0) {
		*src = bswap16(*src);
		src++;
		n -= 2;
	}

	return 0;
}
