/*	$Id: autoconf.c,v 1.24 2015/06/23 19:27:32 tsubai Exp $	*/

/*-
 * Copyright (c) 2001,2002 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/conf.h>
#include <sys/device.h>
#include <sys/kernel.h>
#include <sys/systm.h>

#include <machine/autoconf.h>
#include <machine/cpufunc.h>
#include <machine/mpc860dev.h>

#include <ts3000/board.h>

static u_int calc_crc(u_char *, u_int);
static void *search_record(void *, int, int);
static void getenaddr(void);
void board_init(void);

static struct device *alldevs;
int board_type, board_ram;
char fec_enaddr[6];

/*
 * Determine device configuration for a machine.
 */
void
cpu_configure()
{
	intr_calculatemasks();
	getenaddr();
	timebase_init();
	sswt();

	/* Reserve some DPRAM regions. */
	dpram_reserve(0x800, 0x50);	/* SMC1 */

	alldevs = config_rootfound("mainbus", NULL);
	if (alldevs == NULL)
		panic("configure: mainbus not configured");

	spl0();

	/* Enable hardware interrupts. */
	mtmsr(mfmsr() | PSL_EE | PSL_RI);
}

void
cpu_rootconf()
{
	struct device *dv;
	char *bootdev = "";

	if (strncmp(bootpath, "disk:", 5) == 0)
		bootdev = "wd0";
	if (strncmp(bootpath, "net:", 4) == 0)
		bootdev = "fec0";

	for (dv = alldevs; dv; dv = TAILQ_NEXT(dv, dv_list))
		if (strcmp(dv->dv_xname, bootdev) == 0)
			break;

	booted_device = dv;
	booted_partition = 0;

	printf("boot device: %s\n",
	    booted_device ? booted_device->dv_xname : "<unknown>");

	setroot(booted_device, booted_partition);
}

u_int
calc_crc(p, len)
	u_char *p;
	u_int len;
{
	u_int crc = 0;
	int i;

	while (len--) {
		crc ^= *p++ << 24;
		for (i = 0; i < 8; i++) {
			if (crc & 0x80000000)
				crc ^= 0x88108000;
			crc <<= 1;
		}
	}
	return crc >> 16;
}

struct nvh {
	u_short type;
	u_short size;
	u_short crc;
	u_char data[2];	/* variable length */
};

void *
search_record(addr, type, tries)
	void *addr;
	int type, tries;
{
	struct nvh *p = addr;

	while (tries-- > 0) {
		if (p->type == 9999)
			break;
		if ((p->type > 200 && p->type != 9000) || p->size > 10000)
			break;
		if (p->crc != calc_crc(p->data, p->size))
			break;
		if (p->type == type)
			return p->data;

		p = (void *)(p->data + p->size);
	}
	return NULL;
}

struct nvconfig {
	u_char boot_option;
	u_char filename[40];
	u_char watchdog;
	u_char check_crc;
	u_char memtest;
	u_char res[4];
	u_short maxint;
	u_char enaddr[6];
	u_char myaddr[4];
	u_char mynetmask[4];
	u_char console_speed;
	u_char lcd_reboot;
	u_char cpu_utilization;
	u_char mii_operation;
};

void
getenaddr()
{
	volatile struct mpc860dev *immr = get_immr();
	struct nvconfig *nv;
	paddr_t base;
	char *flash;
	u_int nvoffset;

	switch (board_type) {
	case BOARD_TS3000:
	case BOARD_ACS16:
	case BOARD_ACS32:	/* untested */
	case BOARD_ACS48:	/* untested */
	case BOARD_CS4008:
		nvoffset = 0x30000;
		break;
	default:
		nvoffset = 0x20000;

		/* Move FLASH, UART, and CPLD address (to fit one TLB). */
		asm volatile ("sync");
		immr->siu_br0 &= ~1;	/* disable */
		immr->siu_br1 &= ~1;
		immr->siu_br3 &= ~1;
		immr->siu_br4 &= ~1;
		asm volatile ("eieio");
		immr->siu_br0 &= ~0xffff8000;
		immr->siu_br1 &= ~0xffff8000;
		immr->siu_br3 &= ~0xffff8000;
		immr->siu_br4 &= ~0xffff8000;
		immr->siu_br0 |= 0x40000000;
		immr->siu_br1 |= 0x40200000;
		immr->siu_br3 |= 0x40400000;
		immr->siu_br4 |= 0x40408000;
		immr->siu_br0 |= 1;	/* enable again */
		immr->siu_br1 |= 1;
		immr->siu_br3 |= 1;
		immr->siu_br4 |= 1;
		asm volatile ("eieio");
	}

	base = immr->siu_br0 & 0xffff8000;
	flash = (char *)base + nvoffset;

	nv = search_record(flash, 68, 1000);
	if (nv == NULL) {
		printf("Warning: failed to find FEC ethernet address\n");
		return;
	}

	bcopy(nv->enaddr, fec_enaddr, 6);
}

struct machinfo {
	u_int memory_tested;
	u_int memory_detected;
	u_char altboot;
	u_char error_code;
	u_char pad1[1];
	u_char board_type;
	u_int flash_size;
	u_char clock_freq;
	u_char pad2[3];
	u_char board_version;
	u_char pad3[6];
	u_char watchdog;
	u_char pad4[20];
	u_int CF_size;
	u_char serial[14+1];
};

char rtc_value[20];

void
board_init()
{
	struct machinfo *mp = (void *)0x2000;

	board_type = mp->board_type;
	board_ram  = mp->memory_detected;

	bcopy(mp->pad4, rtc_value, 20);
}
