/*	$Id: machdep.c,v 1.43 2013/08/08 01:19:32 tsubai Exp $	*/

/*
 * Copyright (C) 1995, 1996 Wolfgang Solfrank.
 * Copyright (C) 1995, 1996 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "opt_compat_netbsd.h"
#include "opt_ddb.h"
#include "opt_ipkdb.h"
#include "opt_modular.h"
#include "opt_multiprocessor.h"
#include "ksyms.h"

#include <sys/param.h>
#include <sys/boot_flag.h>
#include <sys/buf.h>
#include <sys/kernel.h>
#include <sys/ksyms.h>
#include <sys/mbuf.h>
#include <sys/mount.h>
#include <sys/msgbuf.h>
#include <sys/reboot.h>
#include <sys/sysctl.h>
#include <sys/systm.h>

#include <machine/powerpc.h>
#include <machine/cpufunc.h>
#include <machine/mpc860dev.h>
#include <machine/pcb.h>
#include <machine/spr.h>
#include <machine/trap.h>

#include <dev/cons.h>

struct vm_map *mb_map;
struct vm_map *phys_map;

/*
 * Global variables used here and there
 */
struct cpu_info cpu_info[1];
paddr_t avail_start, avail_end;
paddr_t msgbuf_paddr;
char bootpath[256];
time_t lastcomintr;

extern struct consdev consdev_smc;
extern struct pcb *proc0paddr;
extern int board_ram;

extern char _edata[], _end[];
extern char trapcode[], alitrap[], dsitrap[], isitrap[];
extern char decrint[], syscalltrap[], itlbmiss[], dtlbmiss[];
extern char extint[];
#ifdef DDB
extern char ddblow[];
#endif

void board_init(void);
void ext_intr(void);
void identifycpu(char *);
void install_extint(void (*)(void));
void setplprcr(u_int);
static void inittlb(void);
static void enable_cache();
#define syncicache(addr, len) __syncicache((void *)(addr), len)

void
initppc(args, kend)
	char *args;
	paddr_t kend;
{
	volatile struct mpc860dev *immr = get_immr();
	int exc;

	inittlb();

	/* Enable address translations. */
	mtmsr(PSL_DR|PSL_IR);

	/* Clear BSS. */
	bzero(_edata, _end - _edata);

	/* Allocate interrupt (and DDB) stack. */
	kend += 8192;
	cpu_info[0].ci_intstk = (void *)kend;
#ifdef DDB
	kend += 8192;
	cpu_info[0].ci_ddbstk = (void *)kend;
#endif

	asm volatile ("mtsprg 0,%0" :: "r"(&cpu_info[0]));

	/* IMMR = 0xff000000; */

	/* Disable CPM interrupts (before consinit). */
	immr->cpm_cimr = 0;

	consinit();

	board_init();
	strcpy(bootpath, args);

#if NKSYMS || defined(DDB) || defined(MODULAR)
	paddr_t ssym = 0;
	const char *cp;

	if (pmatch(bootpath, "* ssym=", &cp) > 0)
		ssym = strtoul(cp, NULL, 16);
	printf("ssym = %lx, kend = %lx\n", ssym, kend);
	if (ssym)
		ksyms_addsyms_elf(1, (void *)ssym, NULL);
#endif

	args = strchr(bootpath, ' ');
	if (args) {
		*args++ = 0;
		args = strchr(args, '-');
		if (args) {
			while (*args) {
				if (*args == ' ')
					break;
				BOOT_FLAG(*args++, boothowto);
			}
		}
	}
#ifdef DDB
	if (boothowto & RB_KDB)
		Debugger();
#endif

	lwp0.l_cpu = curcpu();
	lwp0.l_addr = proc0paddr;
	bzero(lwp0.l_addr, sizeof(struct pcb));

	curlwp = &lwp0;
	curpcb = proc0paddr;

	curpm = curpcb->pcb_pm = pmap_kernel();

	/*
	 * Set up trap vectors
	 */
	for (exc = 0x200; exc < 0x2000; exc += 0x100)
		bcopy(trapcode, (void *)exc, 0x100);

	bcopy(alitrap, (void *)EXC_ALI, 0x100);
	bcopy(decrint, (void *)EXC_DECR, 0x100);
	bcopy(itlbmiss, (char *)EXC_ITLBMISS, 0x100);
	bcopy(dtlbmiss, (char *)EXC_DTLBMISS, 0x100);
	bcopy(isitrap, (void *)EXC_ITLBERR, 0x100);
	bcopy(dsitrap, (void *)EXC_DTLBERR, 0x100);
	bcopy(syscalltrap, (void *)EXC_SC, 0x100);
#ifdef DDB
	bcopy(ddblow, (void *)EXC_PGM, 0x100);
	bcopy(ddblow, (void *)EXC_TRC, 0x100);
	bcopy(ddblow, (void *)EXC_BPT, 0x100);
#endif
	bcopy(extint, (void *)EXC_EXI, 0x100);

	/*
	 * external interrupt handler install
	 */
	install_extint(ext_intr);
	intr_depth = -1;

	syncicache(0, 0x2000);

	avail_start = roundup(kend, NBPG);
	avail_end = board_ram;

	physmem = btoc(avail_end);
	avail_end -= round_page(MSGBUFSIZE);
	msgbuf_paddr = avail_end;

	uvm_setpagesize();
	uvm_page_physload(atop(avail_start), atop(avail_end),
	    atop(avail_start), atop(avail_end), VM_FREELIST_DEFAULT);

	pmap_bootstrap();

	u_int *pte = pmap_kernel()->pm_pdir;
	paddr_t pa;

	/* Map all RAM with 8MB page. */
	for (pa = 0; pa < board_ram; pa += 4096*1024)
		*pte++ = pa | TWC_8M | TWC_V;

	enable_cache();

	/* Initialize power-saving mode. */
	immr->siu_sccr |= SCCR_PRQEN | SCCR_CRQEN;
	immr->siu_plprcrk = 0x55ccaa33;	/* unlock */
}

/*
 * This should probably be in autoconf!				XXX
 */
char cpu_model[80];
char machine[] = MACHINE;		/* from <machine/param.h> */
char machine_arch[] = MACHINE_ARCH;	/* from <machine/param.h> */

void
install_extint(handler)
	void (*handler) __P((void));
{
	extern u_int extint_call[1];
	long offset = (u_long)handler - (u_long)extint_call;
	int msr;

#ifdef DIAGNOSTIC
	if (abs(offset) > 0x1ffffff)
		panic("install_extint: too far away");
#endif
	offset &= 0x03fffffc;
	msr = mfmsr();
	mtmsr(msr & ~PSL_EE);
	*extint_call = (*extint_call & 0xfc000003) | offset;
	syncicache(extint_call, sizeof extint_call);
	mtmsr(msr);
}

/*
 * Machine dependent startup code.
 */
void
cpu_startup()
{
	vaddr_t minaddr, maxaddr;
	char pbuf[9];

	initmsgbuf((caddr_t)msgbuf_paddr, round_page(MSGBUFSIZE));

	printf("%s", version);
	identifycpu(cpu_model);

	format_bytes(pbuf, sizeof(pbuf), ctob(physmem));
	printf("total memory = %s\n", pbuf);

	minaddr = 0;

	/*
	 * Allocate a submap for physio
	 */
	phys_map = uvm_km_suballoc(kernel_map, &minaddr, &maxaddr,
	    VM_PHYS_SIZE, 0, FALSE, NULL);

	/*
	 * No need to allocate an mbuf cluster submap.  Mbuf clusters
	 * are allocated via the pool allocator, and we use direct-mapped
	 * pool pages.
	 */

	format_bytes(pbuf, sizeof(pbuf), ptoa(uvmexp.free));
	printf("avail memory = %s\n", pbuf);
}

/*
 * consinit
 * Initialize system console.
 */
void
consinit()
{
	struct consdev *cp = &consdev_smc;
	static int initted;

	if (initted)
		return;
	initted = 1;

	(*cp->cn_probe)(cp);
	(*cp->cn_init)(cp);
	cn_tab = cp;
}

/*
 * Machine dependent system variables.
 */
static int
sysctl_machdep_booted_kernel(SYSCTLFN_ARGS)
{
	struct sysctlnode node = *rnode;
	char *cp;

	cp = strrchr(bootpath, '/');
	if (cp == NULL) {
		cp = strchr(bootpath, ':');
		if (cp == NULL)
			return ENOENT;
	}
	cp++;
	node.sysctl_data = cp;
	node.sysctl_size = strlen(cp) + 1;
	return sysctl_lookup(SYSCTLFN_CALL(&node));
}

SYSCTL_SETUP(sysctl_machdep_setup, "sysctl machdep subtree setup")
{
	sysctl_createv(clog, 0, NULL, NULL, CTLFLAG_PERMANENT,
	    CTLTYPE_NODE, "machdep", NULL, NULL, 0, NULL, 0,
	    CTL_MACHDEP, CTL_EOL);

	sysctl_createv(clog, 0, NULL, NULL, CTLFLAG_PERMANENT|CTLFLAG_IMMEDIATE,
	    CTLTYPE_INT, "cachelinesize", NULL, NULL, CACHELINESIZE, NULL, 0,
	    CTL_MACHDEP, CPU_CACHELINE, CTL_EOL);

	sysctl_createv(clog, 0, NULL, NULL, CTLFLAG_PERMANENT,
	    CTLTYPE_STRING, "booted_kernel", NULL,
	    sysctl_machdep_booted_kernel, 0, NULL, 0,
	    CTL_MACHDEP, CTL_CREATE, CTL_EOL);
}

/*
 * Crash dump handling.
 */
void
dumpsys()
{
	printf("dumpsys: TBD\n");
}

/*
 * Halt or reboot the machine after syncing/dumping according to howto.
 */
void
cpu_reboot(howto, what)
	int howto;
	char *what;
{
	static int syncing;
	static char str[256];
	char *ap = str, *ap1 = ap;

	boothowto = howto;
	if (!cold && !(howto & RB_NOSYNC) && !syncing) {
		syncing = 1;
		vfs_shutdown();		/* sync */
		resettodr();		/* set wall clock */
	}

	splhigh();

	if (!cold && (howto & RB_DUMP))
		dumpsys();

	doshutdownhooks();

	if (howto & RB_HALT) {
		printf("halted\n\n");

		/* flush cache for msgbuf */
		syncicache(msgbuf_paddr, round_page(MSGBUFSIZE));

		ppc_exit();
	}

	printf("rebooting\n\n");
	if (what && *what) {
		if (strlen(what) > sizeof str - 5)
			printf("boot string too large, ignored\n");
		else {
			strcpy(str, what);
			ap1 = ap = str + strlen(str);
			*ap++ = ' ';
		}
	}
	*ap++ = '-';
	if (howto & RB_SINGLE)
		*ap++ = 's';
	if (howto & RB_KDB)
		*ap++ = 'd';
	*ap++ = 0;
	if (ap[-2] == '-')
		*ap1 = 0;

	/* flush cache for msgbuf */
	syncicache(msgbuf_paddr, round_page(MSGBUFSIZE));

	ppc_boot(str);
}

void
ppc_exit()
{
	printf("\n");
	printf("The operating system has halted.\n");
	printf("Please type any character to reboot.\n\n");
	cnpollc(1);
	cngetc();
	cnpollc(0);

	ppc_boot(NULL);
}

void
ppc_boot(str)
	char *str;
{
	volatile struct mpc860dev *immr = get_immr();

	immr->cpm_padat &= ~0x1000;
	immr->cpm_padir |= 0x1000;
	delay(1000);
	immr->cpm_padat |= 0x1000;

	/* Cause watchdog reset. */
	for (;;);
}

/*
 * Enter low power mode.
 */
void
cpu_idle()
{
	volatile struct mpc860dev *immr = get_immr();
	int msr, dec, plprcr;

	if (time_second - lastcomintr < 5)
		return;

	msr = mfmsr();
	mtmsr(msr | PSL_POW);

	plprcr = immr->siu_plprcr;
	setplprcr(plprcr | PLPRCR_TMIST);
	setplprcr(plprcr | PLPRCR_LPM_DOZE);

	mtmsr(msr & ~PSL_EE);
	dec = mfdec();
	if (dec < 0) {
		mtdec(0x7fff0000);
		mtdec(dec);
	}
	mtmsr(msr);
}


static inline void
eieio()
{
	asm volatile ("eieio; sync");
}

/* Service software watchdog timer. */
void
sswt()
{
	volatile struct mpc860dev *immr = get_immr();

	immr->siu_swsr = 0x556c;
	eieio();
	immr->siu_swsr = 0xaa39;
	eieio();
}

#define SETTLB(tlb, pa, va, size, mode) do {				\
	u_int rpn;							\
									\
	mtspr(MD_CTR, CTR_TLB_INDX(tlb) | CTR_PPCS | CTR_TWAM);		\
	mtspr(MI_CTR, CTR_TLB_INDX(tlb) | CTR_PPCS);			\
	mtspr(MD_EPN, (va) | EPN_EV);					\
	mtspr(MI_EPN, (va) | EPN_EV);					\
	mtspr(MD_TWC, __CONCAT(TWC_, size) | TWC_V);			\
	mtspr(MI_TWC, __CONCAT(TWC_, size) | TWC_V);			\
	rpn = (pa) | __CONCAT(RPN_, size) | (mode) | RPN_SH | RPN_V;	\
	rpn |= RPN_KHIT;						\
	mtspr(MD_RPN, rpn);						\
	mtspr(MI_RPN, rpn);						\
} while (0)

void
inittlb()
{
	/* Invalidate all TLBs. */
	mtspr(MD_CTR, CTR_PPCS | CTR_TWAM);
	mtspr(MI_CTR, CTR_PPCS);
	asm volatile ("tlbia");

	mtspr(M_CASID, 0);
	mtspr(MD_AP, 0x55555555);
	mtspr(MI_AP, 0x55555555);

	/* Map 0x00000000-0x007fffff. */
	SETTLB(28, 0, 0, 8M, RPN_C);

	/* Map 0x00800000-0x00ffffff. */
	SETTLB(29, 0x00800000, 0x00800000, 8M, RPN_C);

	/* Map 0x40000000-0x407fffff (obio). */
	SETTLB(30, 0x40000000, 0x40000000, 8M, RPN_C | RPN_CI);

	/* Map 0xff000000-0xff07ffff (IMMR). */
	SETTLB(31, 0xff000000, 0xff000000, 512K, RPN_C | RPN_CI);

	/* Reserve last four TLBs. */
	mtspr(MD_CTR, CTR_PPCS | CTR_RSV4 | CTR_TWAM);
	mtspr(MI_CTR, CTR_PPCS | CTR_RSV4);

	asm volatile ("tlbia; sync");
}

#define CST_EN		0x80000000	/* Enabled (RO) */
#define CST_FWT		0x40000000	/* Force write thru (D)(RO) */
#define CST_LES		0x20000000	/* Little endian swap (D)(RO) */
#define CST_CMD		0x0f000000
#define CST_CMD_FWT	0x01000000	/* Force write thru enable (D) */
#define CST_CMD_CE	0x02000000	/* Cache enable */
#define CST_CMD_FWTDIS	0x03000000	/* Force write thru disable (D) */
#define CST_CMD_CD	0x04000000	/* Cache disable */
#define CST_CMD_LES	0x05000000	/* Little endian swap enable (D) */
#define CST_CMD_LLCB	0x06000000	/* Load and lock cache block */
#define CST_CMD_LESDIS	0x07000000	/* Little endian swap disable (D) */
#define CST_CMD_ULCB	0x08000000	/* Unlock cache block */
#define CST_CMD_ULALL	0x0a000000	/* Unlock all */
#define CST_CMD_INVALL	0x0c000000	/* Invalidate all */
#define CST_CMD_FCB	0x0e000000	/* Flush cache block (D) */
#define CST_CCER1	0x00200000	/* Cache error type 1 (RO) */
#define CST_CCER2	0x00100000	/* Cache error type 2 (RO) */
#define CST_CCER3	0x00080000	/* Reserved (RO) */

void
enable_cache()
{
	mtspr(IC_CST, CST_CMD_ULALL);
	mtspr(IC_CST, CST_CMD_INVALL);
	mtspr(IC_CST, CST_CMD_CE);

	mtspr(DC_CST, CST_CMD_ULALL);
	mtspr(DC_CST, CST_CMD_INVALL);
	mtspr(DC_CST, CST_CMD_CE);
}

void
flushcache(from, len)
	void *from;
	int len;
{
	int l, off;
	char *p;

	off = (u_int)from & (CACHELINESIZE - 1);
	len += off;

	p = from;
	l = len;
	do {
		asm volatile ("dcbf 0,%0" :: "r"(p));
		p += CACHELINESIZE;
	} while ((l -= CACHELINESIZE) > 0);
	asm volatile ("sync");
}

void
invalidatecache(from, len)
	void *from;
	int len;
{
	int l, off;
	char *p;

	off = (u_int)from & (CACHELINESIZE - 1);
	len += off;

	p = from;
	l = len;
	do {
		asm volatile ("dcbi 0,%0" :: "r"(p));
		p += CACHELINESIZE;
	} while ((l -= CACHELINESIZE) > 0);
	asm volatile ("sync");
}
